local addonName, addon = ...

addon.SP = addon.SP or {}; local SP = addon.SP

function SP.getSpell(name)
	local s = name:upper():gsub("[ :%-%(%)'\"]","")
	return SP.spells[s] and s
end

function SP.isSpell(name)
	return SP.getSpell(name) ~= nil
end

function SP.getSpellId(name)
	local s = SP.spells[name]
	return s and s.id
end

function SP.getSpellById(id)
	if not SP.spellsById then
		SP.spellsById = {}
		for key, s in pairs(SP.spells) do
			SP.spellsById[s.id] = key
		end
	end
	return SP.spellsById[id]
end

function SP.getLocalizedName(name)
	local id = ((type(name) == "number") and name) or (SP.spells[name] and SP.spells[name].id)
	return id and (GetSpellInfo(id))
end

local function isTradeSkillKnown(name, localizedName)
	if GuidelimeDataChar.tradeSkills and GuidelimeDataChar.tradeSkills[name] then return true end
	if SP.getTradeSkillIndex(localizedName) then
		if not GuidelimeDataChar.tradeSkills then GuidelimeDataChar.tradeSkills = {} end
		GuidelimeDataChar.tradeSkills[name] = true 
		return true
	end
	return false
end

function SP.getSpellRank(name)
	local s = SP.spells[name]
	if not s then return 0 end
	local localizedName = GetSpellInfo(s.id)
	local id = select(7, GetSpellInfo(localizedName))
	if not id or not IsSpellKnown(id) then return (isTradeSkillKnown(name, localizedName) and 1) or 0 end
	local rank = GetSpellSubtext(id)
	if rank:sub(1, RANK:len()) == RANK then	return tonumber(rank:sub(RANK:len() + 2)) end
	return 1 
end

function SP.getTradeSkillIndex(localizedName)
	for i = 1, GetNumTradeSkills() do
		local skillName = GetTradeSkillInfo(i)
		if skillName == localizedName then return i end
	end
end

function SP.isRequiredSpell(name, spellMin, spellMax)
	local value = SP.getSpellRank(name)
	if spellMin ~= nil and value < spellMin then return false end
	if spellMax ~= nil and value >= spellMax then return false end
	return true
end

SP.spells = {
		["FLYINGCARPET"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 20000,
			["id"] = 60969,
			["icon"] = 136249,
			["name"] = "Flying Carpet",
		},
		["MINORMANAPOTION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Minor Mana Potion",
			["icon"] = 136243,
			["id"] = 2331,
		},
		["FLASKOFSTONEBLOOD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 53758,
			["icon"] = 236879,
			["name"] = "Flask of Stoneblood",
		},
		["GLYPHOFSCORCH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Scorch",
			["icon"] = 136243,
			["id"] = 56371,
		},
		["EXPERTJEWELCRAFTER"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Expert Jewelcrafter",
			["icon"] = 134073,
			["id"] = 28896,
		},
		["SAPPHIRESIGNET"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["name"] = "Sapphire Signet",
			["icon"] = 136243,
			["id"] = 26903,
		},
		["TYPHOON"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Typhoon",
			["icon"] = 236170,
			["id"] = 50516,
		},
		["ENDLESSHEALINGPOTION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 58871,
			["icon"] = 136240,
			["name"] = "Endless Healing Potion",
		},
		["GLYPHOFDEATHGRIP"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Glyph of Death Grip",
			["icon"] = 132918,
			["id"] = 57213,
		},
		["ARTISANFIRSTAID"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Artisan First Aid",
			["icon"] = 135966,
			["id"] = 10847,
		},
		["COBALTGAUNTLETS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55835,
			["icon"] = 136241,
			["name"] = "Cobalt Gauntlets",
		},
		["RUNICLEATHERGAUNTLETS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Runic Leather Gauntlets",
			["icon"] = 136243,
			["id"] = 19055,
		},
		["MOONFIRE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2822,
			["name"] = "Moonfire",
			["icon"] = 136243,
			["id"] = 563,
		},
		["FROSTSAVAGEBELT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 59582,
			["icon"] = 136249,
			["name"] = "Frostsavage Belt",
		},
		["MITHRILCASING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Mithril Casing",
			["icon"] = 136243,
			["id"] = 12599,
		},
		["RUNECLOTHSHOULDERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Runecloth Shoulders",
			["icon"] = 132149,
			["id"] = 18449,
		},
		["WHITELINENSHIRT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "White Linen Shirt",
			["icon"] = 132149,
			["id"] = 2393,
		},
		["ARTISANSCRIBE"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Artisan Scribe",
			["icon"] = 237171,
			["id"] = 45378,
		},
		["ICYTOUCH"] = {
			["maxRange"] = 20,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 45477,
			["icon"] = 237526,
			["name"] = "Icy Touch",
		},
		["WORMDELIGHT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 45551,
			["icon"] = 133971,
			["name"] = "Worm Delight",
		},
		["FIGURINEJADEOWL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Figurine - Jade Owl",
			["icon"] = 136243,
			["id"] = 26872,
		},
		["CORPSEEXPLOSION"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 1411,
			["id"] = 17616,
			["icon"] = 136133,
			["name"] = "Corpse Explosion",
		},
		["GLYPHOFFRENZIEDREGENERATION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 54810,
			["icon"] = 136243,
			["name"] = "Glyph of Frenzied Regeneration",
		},
		["INKOFTHESEA"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 57715,
			["icon"] = 132918,
			["name"] = "Ink of the Sea",
		},
		["SAVAGESARONITELEGPLATES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55310,
			["icon"] = 136241,
			["name"] = "Savage Saronite Legplates",
		},
		["HELMOFCOMMAND"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55302,
			["icon"] = 136241,
			["name"] = "Helm of Command",
		},
		["STORMHERALD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 50000,
			["id"] = 36263,
			["icon"] = 136241,
			["name"] = "Stormherald",
		},
		["FROSTWEAVEBAG"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 56007,
			["icon"] = 136249,
			["name"] = "Frostweave Bag",
		},
		["HEROICPRESENCE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 6562,
			["icon"] = 133123,
			["name"] = "Heroic Presence",
		},
		["THICKSUNCRYSTAL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53855,
			["icon"] = 134071,
			["name"] = "Thick Sun Crystal",
		},
		["WHIRLWIND"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Whirlwind",
			["icon"] = 132369,
			["id"] = 1680,
		},
		["ROCKBITERWEAPON"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 8017,
			["icon"] = 136086,
			["name"] = "Rockbiter Weapon",
		},
		["SILVERPLATEDSHOTGUN"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Silver-plated Shotgun",
			["icon"] = 136243,
			["id"] = 3949,
		},
		["COARSEBLASTINGPOWDER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Coarse Blasting Powder",
			["icon"] = 136243,
			["id"] = 3929,
		},
		["SCROLLOFSTAMINAV"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Scroll of Stamina V",
			["icon"] = 132918,
			["id"] = 50617,
		},
		["EARTHENVEST"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Earthen Vest",
			["icon"] = 132149,
			["id"] = 8764,
		},
		["SILKYICESHARDBOOTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 56019,
			["icon"] = 136249,
			["name"] = "Silky Iceshard Boots",
		},
		["RUNESTRIKE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 56815,
			["icon"] = 237518,
			["name"] = "Rune Strike",
		},
		["HEAVYMITHRILGAUNTLET"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Heavy Mithril Gauntlet",
			["icon"] = 132961, --136192,
			["id"] = 9928,
		},
		["SMOOTHSUNCRYSTAL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53853,
			["icon"] = 134071,
			["name"] = "Smooth Sun Crystal",
		},
		["GREENSILKENSHOULDERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Green Silken Shoulders",
			["icon"] = 132149,
			["id"] = 8774,
		},
		["PERSISTENTEARTHSIEGEDIAMOND"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 55402,
			["icon"] = 134071,
			["name"] = "Persistent Earthsiege Diamond",
		},
		["JADEDAGGERPENDANT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 56195,
			["icon"] = 134071,
			["name"] = "Jade Dagger Pendant",
		},
		["FORCEFULDARKJADE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53920,
			["icon"] = 134071,
			["name"] = "Forceful Dark Jade",
		},
		["PYROBLAST"] = {
			["maxRange"] = 35,
			["minRange"] = 0,
			["castTime"] = 4704,
			["id"] = 11366,
			["icon"] = 135808,
			["name"] = "Pyroblast",
		},
		["DAMPENMAGIC"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 604,
			["icon"] = 136006,
			["name"] = "Dampen Magic",
		},
		["FROSTWARD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 6143,
			["icon"] = 135850,
			["name"] = "Frost Ward",
		},
		["SARONITEBULWARK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55014,
			["icon"] = 136241,
			["name"] = "Saronite Bulwark",
		},
		["EXPERTCOOK"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 18816,
			["name"] = "Expert Cook",
			["icon"] = 133971,
			["id"] = 19886,
		},
		["LIFEBLOOM"] = {
			["maxRange"] = 40,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Lifebloom",
			["icon"] = 134206,
			["id"] = 33763,
		},
		["RUNEDCOPPERPANTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Runed Copper Pants",
			["icon"] = 134583, --136192,
			["id"] = 3324,
		},
		["DARKARCTICLEGGINGS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 51569,
			["icon"] = 136247,
			["name"] = "Dark Arctic Leggings",
		},
		["VIRULENTSPAULDERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 60651,
			["icon"] = 136247,
			["name"] = "Virulent Spaulders",
		},
		["BOLDBLOODSTONE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53831,
			["icon"] = 134071,
			["name"] = "Bold Bloodstone",
		},
		["COLDSNAP"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 11958,
			["icon"] = 135865,
			["name"] = "Cold Snap",
		},
		["BRILLIANTTITANSTEELHELM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55374,
			["icon"] = 136241,
			["name"] = "Brilliant Titansteel Helm",
		},
		["LAVABURST"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 21158,
			["icon"] = 135830,
			["name"] = "Lava Burst",
		},
		["SPARKLINGAZUREMOONSTONE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 28953,
			["icon"] = 136243,
			["name"] = "Sparkling Azure Moonstone",
		},
		["FLAMESTRIKE"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 1882,
			["id"] = 2120,
			["icon"] = 135826,
			["name"] = "Flamestrike",
		},
		["TWILIGHTTOME"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 64053,
			["icon"] = 132918,
			["name"] = "Twilight Tome",
		},
		["ARCTICSHOULDERPADS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 50946,
			["icon"] = 136247,
			["name"] = "Arctic Shoulderpads",
		},
		["QUICKDRAWQUIVER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Quickdraw Quiver",
			["icon"] = 136247,
			["id"] = 14930,
		},
		["EXPERTFISHING"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 18816,
			["name"] = "Expert Fishing",
			["icon"] = 136245,
			["id"] = 19889,
		},
		["PRISTINEHUGECITRINE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53887,
			["icon"] = 134071,
			["name"] = "Pristine Huge Citrine",
		},
		["CHALLENGINGSHOUT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Challenging Shout",
			["icon"] = 132091,
			["id"] = 1161,
		},
		["SPELLPOWERELIXIR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 33721,
			["icon"] = 236885,
			["name"] = "Spellpower Elixir",
		},
		["GLOBALTHERMALSAPPERCHARGE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 56488,
			["icon"] = 135826,
			["name"] = "Global Thermal Sapper Charge",
		},
		["ANTIVENOM"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Anti-Venom",
			["icon"] = 136068,
			["id"] = 7932,
		},
		["GLYPHOFLESSERHEALINGWAVE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Lesser Healing Wave",
			["icon"] = 136243,
			["id"] = 55438,
		},
		["ARCANERESISTANCE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 20592,
			["icon"] = 136116,
			["name"] = "Arcane Resistance",
		},
		["MASTERSINSCRIPTIONOFTHECRAG"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 4000,
			["id"] = 61118,
			["icon"] = 237171,
			["name"] = "Master's Inscription of the Crag",
		},
		["PRISTINEMONARCHTOPAZ"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 53989,
			["icon"] = 134071,
			["name"] = "Pristine Monarch Topaz",
		},
		["HANDOFFREEDOM"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Hand of Freedom",
			["icon"] = 135968,
			["id"] = 1044,
		},
		["DREAMWEAVECIRCLET"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["name"] = "Dreamweave Circlet",
			["icon"] = 132149,
			["id"] = 12092,
		},
		["HOLYLIGHT"] = {
			["maxRange"] = 40,
			["minRange"] = 0,
			["castTime"] = 2352,
			["name"] = "Holy Light",
			["icon"] = 135920,
			["id"] = 635,
		},
		["GLYPHOFHOLYNOVA"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 55683,
			["icon"] = 136243,
			["name"] = "Glyph of Holy Nova",
		},
		["HANDOFSALVATION"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Hand of Salvation",
			["icon"] = 135967,
			["id"] = 1038,
		},
		["OVERCHARGEDCAPACITOR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 56464,
			["icon"] = 136243,
			["name"] = "Overcharged Capacitor",
		},
		["APPRENTICEHERBALIST"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Apprentice Herbalist",
			["icon"] = 136246,
			["id"] = 2372,
		},
		["NIMBLELEATHERGLOVES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Nimble Leather Gloves",
			["icon"] = 136247,
			["id"] = 9074,
		},
		["DUSKWEAVELEGGINGS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 55901,
			["icon"] = 136249,
			["name"] = "Duskweave Leggings",
		},
		["SUMMONWARHORSE"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Summon Warhorse",
			["icon"] = 136103,
			["id"] = 13820,
		},
		["SCROLLOFRECALL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 9408,
			["name"] = "Scroll of Recall",
			["icon"] = 134940,
			["id"] = 48129,
		},
		["APPRENTICEFIRSTAID"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Apprentice First Aid",
			["icon"] = 135966,
			["id"] = 3279,
		},
		["COBALTTENDERIZER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55201,
			["icon"] = 136241,
			["name"] = "Cobalt Tenderizer",
		},
		["MAIM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Maim",
			["icon"] = 132134,
			["id"] = 22570,
		},
		["THEPLANAREDGE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 50000,
			["id"] = 34541,
			["icon"] = 136241,
			["name"] = "The Planar Edge",
		},
		["RUBYHARE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 56121,
			["icon"] = 237182,
			["name"] = "Ruby Hare",
		},
		["EMBOSSEDLEATHERPANTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Embossed Leather Pants",
			["icon"] = 136247,
			["id"] = 3759,
		},
		["REDMAGEWEAVEBAG"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Red Mageweave Bag",
			["icon"] = 132149,
			["id"] = 12079,
		},
		["NERUBIANCHESTGUARD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 50956,
			["icon"] = 136247,
			["name"] = "Nerubian Chestguard",
		},
		["RUNESCROLLOFFORTITUDE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 69385,
			["icon"] = 132918,
			["name"] = "Runescroll of Fortitude",
		},
		["MYSTICFROSTWOVENROBE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 55911,
			["icon"] = 136249,
			["name"] = "Mystic Frostwoven Robe",
		},
		["AMULETOFTRUESIGHT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["name"] = "Amulet of Truesight",
			["icon"] = 134072,
			["id"] = 63743,
		},
		["ASPECTOFTHEMONKEY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 13163,
			["icon"] = 132159,
			["name"] = "Aspect of the Monkey",
		},
		["SMELTCOBALT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 1500,
			["id"] = 49252,
			["icon"] = 135811,
			["name"] = "Smelt Cobalt",
		},
		["GLYPHOFSEARINGPAIN"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Searing Pain",
			["icon"] = 136243,
			["id"] = 56226,
		},
		["DIAMONDCUTREFRACTORSCOPE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2822,
			["id"] = 61468,
			["icon"] = 136243,
			["name"] = "Diamond-cut Refractor Scope",
		},
		["ICEBLOCK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 27619,
			["icon"] = 135841,
			["name"] = "Ice Block",
		},
		["SCROLLOFSTAMINAIV"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Scroll of Stamina IV",
			["icon"] = 132918,
			["id"] = 50616,
		},
		["NERUBIANGLOVES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 50959,
			["icon"] = 136247,
			["name"] = "Nerubian Gloves",
		},
		["ELIXIROFMIGHTYSTRENGTH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 54218,
			["icon"] = 136240,
			["name"] = "Elixir of Mighty Strength",
		},
		["HEAVYSILVERRING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Heavy Silver Ring",
			["icon"] = 136243,
			["id"] = 25305,
		},
		["CHAMPIONSHUGECITRINE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53874,
			["icon"] = 134071,
			["name"] = "Champion's Huge Citrine",
		},
		["HEAVYSHARPENINGSTONE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Heavy Sharpening Stone",
			["icon"] = 135250, --136192,
			["id"] = 2674,
		},
		["GLYPHOFFLASHOFLIGHT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 54936,
			["icon"] = 136243,
			["name"] = "Glyph of Flash of Light",
		},
		["TRUESILVERCHAMPION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["id"] = 10015,
			["icon"] = 135317, --136192,
			["name"] = "Truesilver Champion",
		},
		["SHOOT"] = {
			["maxRange"] = 30,
			["minRange"] = 5,
			["castTime"] = 1500,
			["id"] = 3018,
			["icon"] = 132222,
			["name"] = "Shoot",
		},
		["FROSTFEVER"] = {
			["maxRange"] = 50000,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 55095,
			["icon"] = 237522,
			["name"] = "Frost Fever",
		},
		["RAVAGE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Ravage",
			["icon"] = 132152,
			["id"] = 3242,
		},
		["LAYONHANDS"] = {
			["maxRange"] = 40,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Lay on Hands",
			["icon"] = 135928,
			["id"] = 633,
		},
		["SMOKEDROCKFIN"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 45560,
			["icon"] = 133971,
			["name"] = "Smoked Rockfin",
		},
		["PICKPOCKET"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 921,
			["icon"] = 133644,
			["name"] = "Pick Pocket",
		},
		["MANGLECAT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Mangle (Cat)",
			["icon"] = 132135,
			["id"] = 33876,
		},
		["GOBLINMORTAR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 12716,
			["icon"] = 134535,
			["name"] = "Goblin Mortar",
		},
		["RIGHTEOUSDEFENSE"] = {
			["maxRange"] = 40,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Righteous Defense",
			["icon"] = 135068,
			["id"] = 31789,
		},
		["EARTHSHOCK"] = {
			["maxRange"] = 25,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 8042,
			["icon"] = 136026,
			["name"] = "Earth Shock",
		},
		["LIONHEARTCHAMPION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 50000,
			["id"] = 34540,
			["icon"] = 136241,
			["name"] = "Lionheart Champion",
		},
		["WEAPONVELLUMII"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Weapon Vellum II",
			["icon"] = 132918,
			["id"] = 59488,
		},
		["MINORREJUVENATIONPOTION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Minor Rejuvenation Potion",
			["icon"] = 136243,
			["id"] = 2332,
		},
		["WILDSCALEBREASTPLATE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 60669,
			["icon"] = 136247,
			["name"] = "Wildscale Breastplate",
		},
		["COPPERSHORTSWORD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Copper Shortsword",
			["icon"] = 135327, --136192,
			["id"] = 2739,
		},
		["BRILLIANTGOLDENDRAENITE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 28938,
			["icon"] = 136243,
			["name"] = "Brilliant Golden Draenite",
		},
		["SWIPECAT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Swipe (Cat)",
			["icon"] = 134296,
			["id"] = 62078,
		},
		["COARSEDYNAMITE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Coarse Dynamite",
			["icon"] = 136243,
			["id"] = 3931,
		},
		["FLAMETONGUEWEAPON"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 8024,
			["icon"] = 135814,
			["name"] = "Flametongue Weapon",
		},
		["JOURNEYMANBLACKSMITH"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Journeyman Blacksmith",
			["icon"] = 136241,
			["id"] = 2021,
		},
		["SMALLSILKPACK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Small Silk Pack",
			["icon"] = 136249,
			["id"] = 3813,
		},
		["OILOFIMMOLATION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Oil of Immolation",
			["icon"] = 136243,
			["id"] = 11451,
		},
		["DEATHGRIP"] = {
			["maxRange"] = 50000,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 49560,
			["icon"] = 237532,
			["name"] = "Death Grip",
		},
		["CRAFTEDHEAVYSHOT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Crafted Heavy Shot",
			["icon"] = 136243,
			["id"] = 3930,
		},
		["MASTERJEWELCRAFTER"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 28901,
			["icon"] = 134073,
			["name"] = "Master Jewelcrafter",
		},
		["APPRENTICEBLACKSMITH"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Apprentice Blacksmith",
			["icon"] = 136241,
			["id"] = 2020,
		},
		["SAPPHIREPENDANTOFWINTERNIGHT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["name"] = "Sapphire Pendant of Winter Night",
			["icon"] = 136243,
			["id"] = 26908,
		},
		["INSCRIBEDHUGECITRINE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53872,
			["icon"] = 134071,
			["name"] = "Inscribed Huge Citrine",
		},
		["GLOWINGSHADOWDRAENITE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 28925,
			["icon"] = 136243,
			["name"] = "Glowing Shadow Draenite",
		},
		["GLYPHOFLAVALASH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Lava Lash",
			["icon"] = 136243,
			["id"] = 55444,
		},
		["FIRSTAID"] = {
			["maxRange"] = 15,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 746,
			["icon"] = 135915,
			["name"] = "First Aid",
		},
		["ENCHANTEDTHORIUM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Enchanted Thorium",
			["icon"] = 133229,
			["id"] = 17180,
		},
		["HEROISM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 23682,
			["icon"] = 135953,
			["name"] = "Heroism",
		},
		["STEELPLATEHELM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Steel Plate Helm",
			["icon"] = 133071, --136192,
			["id"] = 9935,
		},
		["ELIXIROFMIGHTYAGILITY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53840,
			["icon"] = 136240,
			["name"] = "Elixir of Mighty Agility",
		},
		["LESSERINVISIBILITYPOTION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Lesser Invisibility Potion",
			["icon"] = 136243,
			["id"] = 3448,
		},
		["BALANCEDTWILIGHTOPAL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 53969,
			["icon"] = 134071,
			["name"] = "Balanced Twilight Opal",
		},
		["STRANGULATE"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 47476,
			["icon"] = 136214,
			["name"] = "Strangulate",
		},
		["RUNEOFSPELLBREAKING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 54447,
			["icon"] = 136120,
			["name"] = "Rune of Spellbreaking",
		},
		["DARKNERUBIANCHESTPIECE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 60629,
			["icon"] = 136247,
			["name"] = "Dark Nerubian Chestpiece",
		},
		["LINENBOOTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Linen Boots",
			["icon"] = 132149,
			["id"] = 2386,
		},
		["MOLLE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 54710,
			["icon"] = 133871,
			["name"] = "MOLL-E",
		},
		["TURTLESCALEBRACERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Turtle Scale Bracers",
			["icon"] = 136247,
			["id"] = 10518,
		},
		["SAVAGESARONITEHAUBERK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55311,
			["icon"] = 136241,
			["name"] = "Savage Saronite Hauberk",
		},
		["GLYPHOFBLOODSTRIKE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 59332,
			["icon"] = 136243,
			["name"] = "Glyph of Blood Strike",
		},
		["SOLIDBRONZERING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Solid Bronze Ring",
			["icon"] = 136243,
			["id"] = 25490,
		},
		["CUREDHEAVYHIDE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Cured Heavy Hide",
			["icon"] = 136247,
			["id"] = 3818,
		},
		["SKINNING"] = {
			["maxRange"] = 5,
			["minRange"] = 0,
			["castTime"] = 1500,
			["id"] = 8613,
			["icon"] = 134366,
			["name"] = "Skinning",
		},
		["ELIXIROFTHESAGES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Elixir of the Sages",
			["icon"] = 134809,
			["id"] = 17535,
		},
		["EMERALDCHOKER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 64725,
			["icon"] = 134071,
			["name"] = "Emerald Choker",
		},
		["SALTSHAKER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Salt Shaker",
			["icon"] = 134459,
			["id"] = 19566,
		},
		["FROSTLEATHERCLOAK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Frost Leather Cloak",
			["icon"] = 136247,
			["id"] = 9198,
		},
		["RUNICLEATHERSHOULDERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Runic Leather Shoulders",
			["icon"] = 136243,
			["id"] = 19103,
		},
		["GLYPHOFMAGEARMOR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 56383,
			["icon"] = 136243,
			["name"] = "Glyph of Mage Armor",
		},
		["GLYPHOFLIGHTNINGSHIELD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Lightning Shield",
			["icon"] = 136243,
			["id"] = 55448,
		},
		["FELIRONHATCHET"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 29557,
			["icon"] = 132402, --136192,
			["name"] = "Fel Iron Hatchet",
		},
		["HEAVYCOPPERMAUL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Heavy Copper Maul",
			["icon"] = 133055, --136192,
			["id"] = 7408,
		},
		["TRUESILVERCOMMANDERSRING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["name"] = "Truesilver Commander's Ring",
			["icon"] = 134072,
			["id"] = 34959,
		},
		["MYSTERIOUSTAROT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Mysterious Tarot",
			["icon"] = 132918,
			["id"] = 48247,
		},
		["ARCTICLEGGINGS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 50945,
			["icon"] = 136247,
			["name"] = "Arctic Leggings",
		},
		["GLYPHOFFLASHHEAL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Flash Heal",
			["icon"] = 136243,
			["id"] = 55679,
		},
		["CALLOFTHEELEMENTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 66842,
			["icon"] = 310730,
			["name"] = "Call of the Elements",
		},
		["THICKBRONZEDARTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Thick Bronze Darts",
			["icon"] = 136241,
			["id"] = 34979,
		},
		["GREENIRONHAUBERK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["name"] = "Green Iron Hauberk",
			["icon"] = 132624, --136192,
			["id"] = 3508,
		},
		["HEARTSEEKERSCOPE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2822,
			["id"] = 55135,
			["icon"] = 136243,
			["name"] = "Heartseeker Scope",
		},
		["GNOMISHDEATHRAY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 12759,
			["icon"] = 136243,
			["name"] = "Gnomish Death Ray",
		},
		["BARBARICIRONCOLLAR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Barbaric Iron Collar",
			["icon"] = 136243,
			["id"] = 25498,
		},
		["FORMALWHITESHIRT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Formal White Shirt",
			["icon"] = 132149,
			["id"] = 3871,
		},
		["BLIGHT"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 9796,
			["icon"] = 136066,
			["name"] = "Blight",
		},
		["STARKHUGECITRINE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53889,
			["icon"] = 134071,
			["name"] = "Stark Huge Citrine",
		},
		["SCROLLOFSTRENGTHVI"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Scroll of Strength VI",
			["icon"] = 132918,
			["id"] = 58489,
		},
		["CONJUREREFRESHMENT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2822,
			["id"] = 42955,
			["icon"] = 236212,
			["name"] = "Conjure Refreshment",
		},
		["INVISIBILITY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 66,
			["icon"] = 132220,
			["name"] = "Invisibility",
		},
		["DRAKEFISTHAMMER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 50000,
			["id"] = 34545,
			["icon"] = 136241,
			["name"] = "Drakefist Hammer",
		},
		["HAMMERPICK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 56459,
			["icon"] = 136243,
			["name"] = "Hammer Pick",
		},
		["FROSTWEAVEGLOVES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Frostweave Gloves",
			["icon"] = 132149,
			["id"] = 18411,
		},
		["MANAPOTION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Mana Potion",
			["icon"] = 136243,
			["id"] = 3452,
		},
		["STONESCALEOIL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Stonescale Oil",
			["icon"] = 136243,
			["id"] = 17551,
		},
		["GHOSTWEAVEBELT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Ghostweave Belt",
			["icon"] = 132149,
			["id"] = 18410,
		},
		["COARSESHARPENINGSTONE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Coarse Sharpening Stone",
			["icon"] = 135249, --136192,
			["id"] = 2665,
		},
		["MANGLEBEAR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Mangle (Bear)",
			["icon"] = 132135,
			["id"] = 33878,
		},
		["ENDURINGDARKJADE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53918,
			["icon"] = 134071,
			["name"] = "Enduring Dark Jade",
		},
		["ICEBANEGIRDLE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 61009,
			["icon"] = 136241,
			["name"] = "Icebane Girdle",
		},
		["GLYPHOFPOWERWORDSHIELD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Power Word: Shield",
			["icon"] = 136243,
			["id"] = 55672,
		},
		["GLYPHOFINCINERATE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 56242,
			["icon"] = 136243,
			["name"] = "Glyph of Incinerate",
		},
		["RUNEOFLICHBANE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 53331,
			["icon"] = 135914,
			["name"] = "Rune of Lichbane",
		},
		["BRILLIANTAUTUMNSGLOW"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 53956,
			["icon"] = 134071,
			["name"] = "Brilliant Autumn's Glow",
		},
		["PRISMATICBLACKDIAMOND"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 62941,
			["icon"] = 136243,
			["name"] = "Prismatic Black Diamond",
		},
		["SHADOWRESISTANCE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 20579,
			["icon"] = 136152,
			["name"] = "Shadow Resistance",
		},
		["CONEOFCOLD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 120,
			["icon"] = 135852,
			["name"] = "Cone of Cold",
		},
		["LESSERHEALINGWAVE"] = {
			["maxRange"] = 40,
			["minRange"] = 0,
			["castTime"] = 1411,
			["id"] = 8004,
			["icon"] = 136043,
			["name"] = "Lesser Healing Wave",
		},
		["STONEFORGEDCLAYMORE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["id"] = 36133,
			["icon"] = 135347, --136192,
			["name"] = "Stoneforged Claymore",
		},
		["LESSERHEAL"] = {
			["maxRange"] = 40,
			["minRange"] = 0,
			["castTime"] = 1411,
			["id"] = 2050,
			["icon"] = 135929,
			["name"] = "Lesser Heal",
		},
		["HEAVYWOOLENPANTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Heavy Woolen Pants",
			["icon"] = 132149,
			["id"] = 3850,
		},
		["EXPERTLEATHERWORKER"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Expert Leatherworker",
			["icon"] = 133611,
			["id"] = 3812,
		},
		["FELSTEELWHISPERKNIVES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 34983,
			["icon"] = 136241,
			["name"] = "Felsteel Whisper Knives",
		},
		["DUSKWEAVEWRISTWRAPS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 55920,
			["icon"] = 136249,
			["name"] = "Duskweave Wristwraps",
		},
		["FLIGHTFORM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Flight Form",
			["icon"] = 132128,
			["id"] = 33943,
		},
		["ORNATETHORIUMHANDAXE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Ornate Thorium Handaxe",
			["icon"] = 132403, --136192,
			["id"] = 16969,
		},
		["GLYPHOFTHEGHOUL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 57222,
			["icon"] = 132918,
			["name"] = "Glyph of the Ghoul",
		},
		["SOLIDSHARPENINGSTONE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Solid Sharpening Stone",
			["icon"] = 135251, --136192,
			["id"] = 9918,
		},
		["STRENGTHOFEARTHTOTEM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 8075,
			["icon"] = 136023,
			["name"] = "Strength of Earth Totem",
		},
		["BARBARICHARNESS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Barbaric Harness",
			["icon"] = 136247,
			["id"] = 6661,
		},
		["FROSTSAVAGEGLOVES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 59586,
			["icon"] = 136249,
			["name"] = "Frostsavage Gloves",
		},
		["NETHERWEAVEBRACERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 26764,
			["icon"] = 132149,
			["name"] = "Netherweave Bracers",
		},
		["RUNICHEALINGPOTION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53836,
			["icon"] = 136240,
			["name"] = "Runic Healing Potion",
		},
		["GLYPHOFPLAGUESTRIKE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 57221,
			["icon"] = 132918,
			["name"] = "Glyph of Plague Strike",
		},
		["HEAVYJADERING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Heavy Jade Ring",
			["icon"] = 136243,
			["id"] = 36524,
		},
		["MACESPECIALIZATION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 5530,
			["icon"] = 133476,
			["name"] = "Mace Specialization",
		},
		["ELIXIROFWATERBREATHING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Elixir of Water Breathing",
			["icon"] = 136243,
			["id"] = 7179,
		},
		["SCROLLOFSTAMINAVII"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 50619,
			["icon"] = 132918,
			["name"] = "Scroll of Stamina VII",
		},
		["GREENWORKMANSSHIRT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 56000,
			["icon"] = 136249,
			["name"] = "Green Workman's Shirt",
		},
		["STALWARTHUGECITRINE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53890,
			["icon"] = 134071,
			["name"] = "Stalwart Huge Citrine",
		},
		["STEELBREASTPLATE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Steel Breastplate",
			["icon"] = 132740, --136192,
			["id"] = 9916,
		},
		["DRAGONSBREATH"] = {
			["maxRange"] = 50,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 29964,
			["icon"] = 134153,
			["name"] = "Dragon's Breath",
		},
		["SCROLLOFSTAMINAVIII"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 50620,
			["icon"] = 132918,
			["name"] = "Scroll of Stamina VIII",
		},
		["DIVINEPROTECTION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Divine Protection",
			["icon"] = 135954,
			["id"] = 498,
		},
		["NETHERWEAVEGLOVES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 26770,
			["icon"] = 132149,
			["name"] = "Netherweave Gloves",
		},
		["MYSTICFROSTWOVENSHOULDERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 55910,
			["icon"] = 136249,
			["name"] = "Mystic Frostwoven Shoulders",
		},
		["DENSESHARPENINGSTONE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Dense Sharpening Stone",
			["icon"] = 135252, --136192,
			["id"] = 16641,
		},
		["VISAGELIQUIFICATIONGOGGLES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 50000,
			["id"] = 56484,
			["icon"] = 136243,
			["name"] = "Visage Liquification Goggles",
		},
		["SCROLLOFAGILITYII"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Scroll of Agility II",
			["icon"] = 132918,
			["id"] = 58473,
		},
		["NETHERWEAVEBOOTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 26772,
			["icon"] = 132149,
			["name"] = "Netherweave Boots",
		},
		["SAPPHIREOWL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 56186,
			["icon"] = 237178,
			["name"] = "Sapphire Owl",
		},
		["EMERALDBOAR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 56188,
			["icon"] = 237186,
			["name"] = "Emerald Boar",
		},
		["WINDFORGEDRAPIER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["id"] = 36131,
			["icon"] = 135340, --136192,
			["name"] = "Windforged Rapier",
		},
		["THORIUMWIDGET"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Thorium Widget",
			["icon"] = 136243,
			["id"] = 19791,
		},
		["LIONSINK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Lion's Ink",
			["icon"] = 132918,
			["id"] = 57704,
		},
		["SUBTLEBLOODSTONE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53843,
			["icon"] = 134071,
			["name"] = "Subtle Bloodstone",
		},
		["WIZARDWEAVELEGGINGS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Wizardweave Leggings",
			["icon"] = 132149,
			["id"] = 18421,
		},
		["TRANSMUTEEARTHSIEGEDIAMOND"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 57427,
			["icon"] = 134085,
			["name"] = "Transmute: Earthsiege Diamond",
		},
		["SCROLLOFINTELLECTII"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Scroll of Intellect II",
			["icon"] = 132918,
			["id"] = 50598,
		},
		["SPARKLINGCHALCEDONY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53940,
			["icon"] = 134071,
			["name"] = "Sparkling Chalcedony",
		},
		["FROSTSAVAGESHOULDERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 59584,
			["icon"] = 136249,
			["name"] = "Frostsavage Shoulders",
		},
		["STEALTH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 1784,
			["icon"] = 132320,
			["name"] = "Stealth",
		},
		["GLYPHOFFROSTSHOCK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Frost Shock",
			["icon"] = 136243,
			["id"] = 55443,
		},
		["FELIRONSHELLS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 30346,
			["icon"] = 136243,
			["name"] = "Fel Iron Shells",
		},
		["ARCTICHELM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 51572,
			["icon"] = 136247,
			["name"] = "Arctic Helm",
		},
		["FIRERESISTANCETOTEM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 8184,
			["icon"] = 135832,
			["name"] = "Fire Resistance Totem",
		},
		["GLYPHOFRIGHTEOUSDEFENSE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Righteous Defense",
			["icon"] = 136243,
			["id"] = 54929,
		},
		["NATURESGRASP"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Nature's Grasp",
			["icon"] = 136063,
			["id"] = 16689,
		},
		["ELIXIROFBRUTEFORCE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Elixir of Brute Force",
			["icon"] = 134839,
			["id"] = 17537,
		},
		["SHIFTINGSHADOWCRYSTAL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53860,
			["icon"] = 134071,
			["name"] = "Shifting Shadow Crystal",
		},
		["THORIUMBELT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Thorium Belt",
			["icon"] = 132519, --136192,
			["id"] = 16643,
		},
		["RIGIDGOLDENDRAENITE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 28948,
			["icon"] = 136243,
			["name"] = "Rigid Golden Draenite",
		},
		["SHIELDWALL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Shield Wall",
			["icon"] = 132362,
			["id"] = 871,
		},
		["GOBLINROCKETFUELRECIPE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 12715,
			["icon"] = 136243,
			["name"] = "Goblin Rocket Fuel Recipe",
		},
		["SEALOFLIGHT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Seal of Light",
			["icon"] = 135917,
			["id"] = 20165,
		},
		["GLYPHOFSMITE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Smite",
			["icon"] = 136243,
			["id"] = 55692,
		},
		["SCROLLOFSPIRITII"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Scroll of Spirit II",
			["icon"] = 132918,
			["id"] = 50605,
		},
		["FACESOFDOOM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 59498,
			["icon"] = 132918,
			["name"] = "Faces of Doom",
		},
		["BRILLIANTSARONITEPAULDRONS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 59440,
			["icon"] = 136241,
			["name"] = "Brilliant Saronite Pauldrons",
		},
		["ICYBLASTINGPRIMERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 39971,
			["icon"] = 136243,
			["name"] = "Icy Blasting Primers",
		},
		["TOUGHENEDLEATHERARMOR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Toughened Leather Armor",
			["icon"] = 136247,
			["id"] = 2166,
		},
		["MINDAMPLIFICATIONDISH"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 470,
			["id"] = 67799,
			["icon"] = 135995,
			["name"] = "Mind Amplification Dish",
		},
		["COMPACTHARVESTREAPERKIT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Compact Harvest Reaper Kit",
			["icon"] = 136243,
			["id"] = 3963,
		},
		["SILVERSKELETONKEY"] = {
			["maxRange"] = 5,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Silver Skeleton Key",
			["icon"] = 136243,
			["id"] = 19646,
		},
		["FIREEATERSGUIDE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Fire Eater's Guide",
			["icon"] = 132918,
			["id"] = 59489,
		},
		["SPIDERSILKBOOTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Spidersilk Boots",
			["icon"] = 132149,
			["id"] = 3855,
		},
		["ARMORPLATEDCOMBATSHOTGUN"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 50000,
			["id"] = 56479,
			["icon"] = 136243,
			["name"] = "Armor Plated Combat Shotgun",
		},
		["EXPANSIVEMIND"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 20591,
			["icon"] = 132864,
			["name"] = "Expansive Mind",
		},
		["CORRODEDSARONITEWOUNDBRINGER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55184,
			["icon"] = 136241,
			["name"] = "Corroded Saronite Woundbringer",
		},
		["TITANSTEELBONECRUSHER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55370,
			["icon"] = 136241,
			["name"] = "Titansteel Bonecrusher",
		},
		["GLYPHOFMULTISHOT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Multi-Shot",
			["icon"] = 136243,
			["id"] = 56836,
		},
		["BRONZEFRAMEWORK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Bronze Framework",
			["icon"] = 136243,
			["id"] = 3953,
		},
		["CRUDESCOPE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2822,
			["name"] = "Crude Scope",
			["icon"] = 136243,
			["id"] = 3974,
		},
		["DARKMOONCARD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Darkmoon Card",
			["icon"] = 132918,
			["id"] = 59502,
		},
		["FIREGOGGLES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["name"] = "Fire Goggles",
			["icon"] = 136243,
			["id"] = 12594,
		},
		["PYGMYOIL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 53805,
			["icon"] = 134718,
			["name"] = "Pygmy Oil",
		},
		["GLYPHOFRENDING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Glyph of Rending",
			["icon"] = 132918,
			["id"] = 57163,
		},
		["GLYPHOFRAPIDFIRE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 56828,
			["icon"] = 136243,
			["name"] = "Glyph of Rapid Fire",
		},
		["BLACKJELLY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 64358,
			["icon"] = 252178,
			["name"] = "Black Jelly",
		},
		["BLACKMOUTHOIL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 1500,
			["name"] = "Blackmouth Oil",
			["icon"] = 136243,
			["id"] = 7836,
		},
		["BRILLIANTSUNCRYSTAL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53852,
			["icon"] = 134071,
			["name"] = "Brilliant Sun Crystal",
		},
		["JOURNEYMANRIDING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 33391,
			["icon"] = 136103,
			["name"] = "Journeyman Riding",
		},
		["GREATERDARKMOONCARD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 59503,
			["icon"] = 132918,
			["name"] = "Greater Darkmoon Card",
		},
		["WATERSHIELD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 23575,
			["icon"] = 132315,
			["name"] = "Water Shield",
		},
		["GOBLINBOMBDISPENSER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 12755,
			["icon"] = 136243,
			["name"] = "Goblin Bomb Dispenser",
		},
		["FELSTEELSTABILIZER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 30309,
			["icon"] = 136243,
			["name"] = "Felsteel Stabilizer",
		},
		["ARCANEBLAST"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 10833,
			["icon"] = 136146,
			["name"] = "Arcane Blast",
		},
		["SPARKLINGSKYSAPPHIRE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 53953,
			["icon"] = 134071,
			["name"] = "Sparkling Sky Sapphire",
		},
		["EARTHSHATTER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 35750,
			["icon"] = 132838,
			["name"] = "Earth Shatter",
		},
		["ELIXIROFWISDOM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Elixir of Wisdom",
			["icon"] = 136243,
			["id"] = 3171,
		},
		["ICEBORNEBOOTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 50942,
			["icon"] = 136247,
			["name"] = "Iceborne Boots",
		},
		["ROSECOLOREDGOGGLES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["name"] = "Rose Colored Goggles",
			["icon"] = 136243,
			["id"] = 12618,
		},
		["MEDIUMARMORKIT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Medium Armor Kit",
			["icon"] = 136247,
			["id"] = 2165,
		},
		["RUNEFORGING"] = {
			["maxRange"] = 50000,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 53424,
			["icon"] = 237523,
			["name"] = "Runeforging",
		},
		["COPPERBATTLEAXE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Copper Battle Axe",
			["icon"] = 135420, --136192,
			["id"] = 3293,
		},
		["WATERBREATHING"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 131,
			["icon"] = 136148,
			["name"] = "Water Breathing",
		},
		["SPIKEDCOBALTHELM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 54917,
			["icon"] = 136241,
			["name"] = "Spiked Cobalt Helm",
		},
		["LIVINGBOMB"] = {
			["maxRange"] = 45,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 20475,
			["icon"] = 132863,
			["name"] = "Living Bomb",
		},
		["PORTALIRONFORGE"] = {
			["maxRange"] = 10,
			["minRange"] = 0,
			["castTime"] = 9408,
			["id"] = 11416,
			["icon"] = 135743,
			["name"] = "Portal: Ironforge",
		},
		["FROSTHIDELEGARMOR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 50965,
			["icon"] = 136247,
			["name"] = "Frosthide Leg Armor",
		},
		["BRILLIANTSARONITEBRACERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 59438,
			["icon"] = 136241,
			["name"] = "Brilliant Saronite Bracers",
		},
		["FLEXWEAVEUNDERLAY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 55002,
			["icon"] = 136243,
			["name"] = "Flexweave Underlay",
		},
		["NERUBIANLEGARMOR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3763,
			["id"] = 50902,
			["icon"] = 136247,
			["name"] = "Nerubian Leg Armor",
		},
		["TIGERSEYEBAND"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Tigerseye Band",
			["icon"] = 136243,
			["id"] = 32179,
		},
		["SUMMONIMP"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 9408,
			["id"] = 688,
			["icon"] = 136218,
			["name"] = "Summon Imp",
		},
		["SCROLLOFAGILITYV"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Scroll of Agility V",
			["icon"] = 132918,
			["id"] = 58480,
		},
		["HUGETHORIUMBATTLEAXE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Huge Thorium Battleaxe",
			["icon"] = 135581, --136192,
			["id"] = 16971,
		},
		["WORMHOLEGENERATORNORTHREND"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 67920,
			["icon"] = 136243,
			["name"] = "Wormhole Generator: Northrend",
		},
		["MAIL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 8737,
			["icon"] = 132627,
			["name"] = "Mail",
		},
		["SAVAGECOBALTSLICER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55177,
			["icon"] = 136241,
			["name"] = "Savage Cobalt Slicer",
		},
		["ARTISANFISHING"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Artisan Fishing",
			["icon"] = 136245,
			["id"] = 18249,
		},
		["HOLYSHOCK"] = {
			["maxRange"] = 20,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Holy Shock",
			["icon"] = 135972,
			["id"] = 20473,
		},
		["THUNDER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 50000,
			["id"] = 34547,
			["icon"] = 136241,
			["name"] = "Thunder",
		},
		["GLYPHOFEXECUTION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Glyph of Execution",
			["icon"] = 132918,
			["id"] = 57156,
		},
		["CHAINHEAL"] = {
			["maxRange"] = 40,
			["minRange"] = 0,
			["castTime"] = 2352,
			["id"] = 1064,
			["icon"] = 136042,
			["name"] = "Chain Heal",
		},
		["CALLOFTHEANCESTORS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 66843,
			["icon"] = 310731,
			["name"] = "Call of the Ancestors",
		},
		["ELEMENTALBLASTINGPOWDER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 30303,
			["icon"] = 136243,
			["name"] = "Elemental Blasting Powder",
		},
		["JOURNEYMANTAILOR"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Journeyman Tailor",
			["icon"] = 136249,
			["id"] = 3912,
		},
		["BEASTMASTERY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 53270,
			["icon"] = 236175,
			["name"] = "Beast Mastery",
		},
		["ORNATESARONITEWAISTGUARD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 56551,
			["icon"] = 136241,
			["name"] = "Ornate Saronite Waistguard",
		},
		["HIGHPOWEREDFLASHLIGHT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["name"] = "High-powered Flashlight",
			["icon"] = 136243,
			["id"] = 63750,
		},
		["HANDMOUNTEDPYROROCKET"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 54998,
			["icon"] = 136243,
			["name"] = "Hand-Mounted Pyro Rocket",
		},
		["FIREBALL"] = {
			["maxRange"] = 35,
			["minRange"] = 0,
			["castTime"] = 1411,
			["id"] = 133,
			["icon"] = 135812,
			["name"] = "Fireball",
		},
		["FELIRONCHAINCOIF"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 29551,
			["icon"] = 133137, --136192,
			["name"] = "Fel Iron Chain Coif",
		},
		["HAMMEROFWRATH"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Hammer of Wrath",
			["icon"] = 132326,
			["id"] = 24239,
		},
		["GLYPHOFVAMPIRICBLOOD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 57227,
			["icon"] = 132918,
			["name"] = "Glyph of Vampiric Blood",
		},
		["ELIXIROFDETECTUNDEAD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Elixir of Detect Undead",
			["icon"] = 136243,
			["id"] = 11460,
		},
		["PUMMEL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Pummel",
			["icon"] = 132938,
			["id"] = 6552,
		},
		["RESOLUTEHUGECITRINE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53893,
			["icon"] = 134071,
			["name"] = "Resolute Huge Citrine",
		},
		["SUPERIORMANAPOTION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Superior Mana Potion",
			["icon"] = 136243,
			["id"] = 17553,
		},
		["EMPOWERRUNEWEAPON"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 47568,
			["icon"] = 135372,
			["name"] = "Empower Rune Weapon",
		},
		["APPRENTICEFISHING"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Apprentice Fishing",
			["icon"] = 136245,
			["id"] = 7733,
		},
		["SCROLLOFAGILITYIII"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Scroll of Agility III",
			["icon"] = 132918,
			["id"] = 58476,
		},
		["GIFTOFTHEWILD"] = {
			["maxRange"] = 40,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Gift of the Wild",
			["icon"] = 136038,
			["id"] = 21849,
		},
		["HIEXPLOSIVEBOMB"] = {
			["maxRange"] = 15,
			["minRange"] = 0,
			["castTime"] = 941,
			["name"] = "Hi-Explosive Bomb",
			["icon"] = 135826,
			["id"] = 12543,
		},
		["SNOWFALLINK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 57716,
			["icon"] = 132918,
			["name"] = "Snowfall Ink",
		},
		["BIGBRONZEBOMB"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Big Bronze Bomb",
			["icon"] = 136243,
			["id"] = 3950,
		},
		["DRYPORKRIBS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Dry Pork Ribs",
			["icon"] = 134004, --136192,
			["id"] = 2546,
		},
		["MALACHITEPENDANT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Malachite Pendant",
			["icon"] = 136243,
			["id"] = 32178,
		},
		["BRONZETORC"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Bronze Torc",
			["icon"] = 136243,
			["id"] = 38175,
		},
		["COWER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Cower",
			["icon"] = 132118,
			["id"] = 1742,
		},
		["GOUGE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 1776,
			["icon"] = 132155,
			["name"] = "Gouge",
		},
		["THUNDERCLAP"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Thunder Clap",
			["icon"] = 136105,
			["id"] = 6343,
		},
		["CLOAKOFTHEMOON"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 56014,
			["icon"] = 136249,
			["name"] = "Cloak of the Moon",
		},
		["HANDFULOFFELIRONBOLTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 30305,
			["icon"] = 136243,
			["name"] = "Handful of Fel Iron Bolts",
		},
		["TOUGHNESS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 12299,
			["icon"] = 135892,
			["name"] = "Toughness",
		},
		["ROBEOFPOWER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["name"] = "Robe of Power",
			["icon"] = 132149,
			["id"] = 8770,
		},
		["SHADOWJADEFOCUSINGLENS"] = {
			["maxRange"] = 60,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 56191,
			["icon"] = 134072,
			["name"] = "Shadow Jade Focusing Lens",
		},
		["ARCTICGLOVES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 50947,
			["icon"] = 136247,
			["name"] = "Arctic Gloves",
		},
		["THEBLACKPEARL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["id"] = 41415,
			["icon"] = 134071,
			["name"] = "The Black Pearl",
		},
		["JOURNEYMANMINER"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Journeyman Miner",
			["icon"] = 136248,
			["id"] = 2582,
		},
		["EVISCERATE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 2098,
			["icon"] = 132292,
			["name"] = "Eviscerate",
		},
		["GLYPHOFLIGHTNINGBOLT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Lightning Bolt",
			["icon"] = 136243,
			["id"] = 55453,
		},
		["TEMPEREDSARONITELEGPLATES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 54554,
			["icon"] = 136241,
			["name"] = "Tempered Saronite Legplates",
		},
		["GLYPHOFFREEZINGTRAP"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Freezing Trap",
			["icon"] = 136243,
			["id"] = 56845,
		},
		["SHOVELTUSKSTEAK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 45550,
			["icon"] = 133971,
			["name"] = "Shoveltusk Steak",
		},
		["SMELTIRON"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 1500,
			["name"] = "Smelt Iron",
			["icon"] = 136243,
			["id"] = 3307,
		},
		["LARGECOPPERBOMB"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Large Copper Bomb",
			["icon"] = 136243,
			["id"] = 3937,
		},
		["PESTILENCE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 50842,
			["icon"] = 136182,
			["name"] = "Pestilence",
		},
		["EMBOSSEDLEATHERGLOVES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Embossed Leather Gloves",
			["icon"] = 136247,
			["id"] = 3756,
		},
		["RUNEDCOPPERGAUNTLETS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Runed Copper Gauntlets",
			["icon"] = 132938, --136192,
			["id"] = 3323,
		},
		["BLACKMAGEWEAVEROBE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Black Mageweave Robe",
			["icon"] = 132149,
			["id"] = 12050,
		},
		["ARTISANENGINEER"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Artisan Engineer",
			["icon"] = 136243,
			["id"] = 12657,
		},
		["SOCKETBRACER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 55628,
			["icon"] = 133273,
			["name"] = "Socket Bracer",
		},
		["DIVINEINTERVENTION"] = {
			["maxRange"] = 40,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Divine Intervention",
			["icon"] = 136106,
			["id"] = 19752,
		},
		["SUPERIORHEALINGPOTION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Superior Healing Potion",
			["icon"] = 136243,
			["id"] = 11457,
		},
		["RIGHTEOUSGREAVES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55304,
			["icon"] = 136241,
			["name"] = "Righteous Greaves",
		},
		["TWISTINGNETHERCHAINSHIRT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 50000,
			["id"] = 34530,
			["icon"] = 136241,
			["name"] = "Twisting Nether Chain Shirt",
		},
		["DUSKWEAVEROBE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 55921,
			["icon"] = 136249,
			["name"] = "Duskweave Robe",
		},
		["WICKEDEDGEOFTHEPLANES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 50000,
			["id"] = 36260,
			["icon"] = 136241,
			["name"] = "Wicked Edge of the Planes",
		},
		["MASTERLEATHERWORKER"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 32550,
			["icon"] = 133611,
			["name"] = "Master Leatherworker",
		},
		["TOMEOFKINGS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Tome of Kings",
			["icon"] = 132918,
			["id"] = 59484,
		},
		["GLYPHOFTHEHAWK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 56856,
			["icon"] = 136243,
			["name"] = "Glyph of the Hawk",
		},
		["MOONSHROUD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 56001,
			["icon"] = 132149,
			["name"] = "Moonshroud",
		},
		["IRONGRENADE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Iron Grenade",
			["icon"] = 136243,
			["id"] = 3962,
		},
		["GOBLINSAPPERCHARGE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 12760,
			["icon"] = 136243,
			["name"] = "Goblin Sapper Charge",
		},
		["FROSTWOVENROBE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 55903,
			["icon"] = 136249,
			["name"] = "Frostwoven Robe",
		},
		["GLYPHOFFEINT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 56804,
			["icon"] = 136243,
			["name"] = "Glyph of Feint",
		},
		["HILLMANSCLOAK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Hillman's Cloak",
			["icon"] = 136247,
			["id"] = 3760,
		},
		["TOTEMICRECALL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 36936,
			["icon"] = 310733,
			["name"] = "Totemic Recall",
		},
		["INSCRIBEDFLAMESPESSARITE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 28910,
			["icon"] = 136243,
			["name"] = "Inscribed Flame Spessarite",
		},
		["FURIOUSSARONITEBEATSTICK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55182,
			["icon"] = 136241,
			["name"] = "Furious Saronite Beatstick",
		},
		["PUREHORNSPAULDERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 60671,
			["icon"] = 136247,
			["name"] = "Purehorn Spaulders",
		},
		["GLYPHOFBACKSTAB"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Backstab",
			["icon"] = 136243,
			["id"] = 56800,
		},
		["RUNECLOTHBANDAGE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 1500,
			["name"] = "Runecloth Bandage",
			["icon"] = 133681, --136192,
			["id"] = 18629,
		},
		["COBALTLEGPLATES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 52567,
			["icon"] = 136241,
			["name"] = "Cobalt Legplates",
		},
		["FRACTUREDBLOODSTONE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53845,
			["icon"] = 134071,
			["name"] = "Fractured Bloodstone",
		},
		["THICKADAMANTITENECKLACE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["id"] = 31051,
			["icon"] = 136243,
			["name"] = "Thick Adamantite Necklace",
		},
		["CURETOXINS"] = {
			["maxRange"] = 40,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 526,
			["icon"] = 136067,
			["name"] = "Cure Toxins",
		},
		["THORIUMSHELLS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Thorium Shells",
			["icon"] = 136243,
			["id"] = 19800,
		},
		["ELIXIROFMINORFORTITUDE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Elixir of Minor Fortitude",
			["icon"] = 136243,
			["id"] = 2334,
		},
		["YELLOWLUMBERJACKSHIRT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 55995,
			["icon"] = 136249,
			["name"] = "Yellow Lumberjack Shirt",
		},
		["CORRUPTION"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 172,
			["icon"] = 136118,
			["name"] = "Corruption",
		},
		["ARTISANSKINNER"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Artisan Skinner",
			["icon"] = 134366,
			["id"] = 10769,
		},
		["GLYPHOFEXORCISM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Exorcism",
			["icon"] = 136243,
			["id"] = 54934,
		},
		["WEAPONVELLUMIII"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 59501,
			["icon"] = 132918,
			["name"] = "Weapon Vellum III",
		},
		["COBALTSKELETONKEY"] = {
			["maxRange"] = 5,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 59404,
			["icon"] = 136243,
			["name"] = "Cobalt Skeleton Key",
		},
		["MANUALOFCLOUDS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Manual of Clouds",
			["icon"] = 132918,
			["id"] = 59494,
		},
		["AVENGINGWRATH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Avenging Wrath",
			["icon"] = 135875,
			["id"] = 31884,
		},
		["HANDSTITCHEDLEATHERBELT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Handstitched Leather Belt",
			["icon"] = 136247,
			["id"] = 3753,
		},
		["DAWNSTARINK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Dawnstar Ink",
			["icon"] = 132918,
			["id"] = 57706,
		},
		["HEROICSTRIKE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Heroic Strike",
			["icon"] = 132282,
			["id"] = 78,
		},
		["HIIMPACTMITHRILSLUGS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Hi-Impact Mithril Slugs",
			["icon"] = 136243,
			["id"] = 12596,
		},
		["MOLTENARMOR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 30482,
			["icon"] = 132221,
			["name"] = "Molten Armor",
		},
		["GLYPHOFFEARWARD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Fear Ward",
			["icon"] = 136243,
			["id"] = 55678,
		},
		["PURIFIEDSHADOWPEARL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 41429,
			["icon"] = 134071,
			["name"] = "Purified Shadow Pearl",
		},
		["BACKSTAB"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 53,
			["icon"] = 132090,
			["name"] = "Backstab",
		},
		["EXPERTBLACKSMITH"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Expert Blacksmith",
			["icon"] = 136241,
			["id"] = 3539,
		},
		["SWIMSPEEDPOTION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Swim Speed Potion",
			["icon"] = 136243,
			["id"] = 7841,
		},
		["BLOODPLAGUE"] = {
			["maxRange"] = 50000,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 55078,
			["icon"] = 237514,
			["name"] = "Blood Plague",
		},
		["GHOSTWOLF"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 1882,
			["id"] = 2645,
			["icon"] = 136095,
			["name"] = "Ghost Wolf",
		},
		["GRANDMASTERBLACKSMITH"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 51298,
			["icon"] = 136241,
			["name"] = "Grand Master Blacksmith",
		},
		["GRANDMASTERHERBALIST"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 50301,
			["icon"] = 136246,
			["name"] = "Grand Master Herbalist",
		},
		["BIGBRONZEKNIFE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Big Bronze Knife",
			["icon"] = 135640, --136192,
			["id"] = 3491,
		},
		["DEMORALIZINGROAR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Demoralizing Roar",
			["icon"] = 132121,
			["id"] = 99,
		},
		["GREENLEATHERBELT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Green Leather Belt",
			["icon"] = 136247,
			["id"] = 3774,
		},
		["GLYPHOFCRUSADERSTRIKE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Crusader Strike",
			["icon"] = 136243,
			["id"] = 54927,
		},
		["TOMEOFTHEDAWN"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Tome of the Dawn",
			["icon"] = 132918,
			["id"] = 59475,
		},
		["RUNEOFCINDERGLACIER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 53341,
			["icon"] = 136130,
			["name"] = "Rune of Cinderglacier",
		},
		["TEMPEREDSARONITEBREASTPLATE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 54553,
			["icon"] = 136241,
			["name"] = "Tempered Saronite Breastplate",
		},
		["RESURGENTHEALINGPOTION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53838,
			["icon"] = 136240,
			["name"] = "Resurgent Healing Potion",
		},
		["THICKARMORKIT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Thick Armor Kit",
			["icon"] = 136247,
			["id"] = 10487,
		},
		["RIPTIDE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 22419,
			["icon"] = 132343,
			["name"] = "Riptide",
		},
		["COPPERAXE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Copper Axe",
			["icon"] = 132417, --136192,
			["id"] = 2738,
		},
		["EAGLEBANEBRACERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 60652,
			["icon"] = 136247,
			["name"] = "Eaglebane Bracers",
		},
		["SHADOWWORDPAIN"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 589,
			["icon"] = 136207,
			["name"] = "Shadow Word: Pain",
		},
		["SPIKEDCOBALTBELT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 54946,
			["icon"] = 136241,
			["name"] = "Spiked Cobalt Belt",
		},
		["MAMMOTHMEAL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 45549,
			["icon"] = 133971,
			["name"] = "Mammoth Meal",
		},
		["GNOMISHARMYKNIFE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 56462,
			["icon"] = 136243,
			["name"] = "Gnomish Army Knife",
		},
		["HANDFULOFCOPPERBOLTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Handful of Copper Bolts",
			["icon"] = 136243,
			["id"] = 3922,
		},
		["SEALOFRIGHTEOUSNESS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Seal of Righteousness",
			["icon"] = 132325,
			["id"] = 20154,
		},
		["BLACKDUSKWEAVEWRISTWRAPS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 55943,
			["icon"] = 136249,
			["name"] = "Black Duskweave Wristwraps",
		},
		["JORMSCALEFOOTPADS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 60666,
			["icon"] = 136247,
			["name"] = "Jormscale Footpads",
		},
		["DUSKWEAVESHOULDERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 55923,
			["icon"] = 136249,
			["name"] = "Duskweave Shoulders",
		},
		["UNSTABLETRIGGER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Unstable Trigger",
			["icon"] = 136243,
			["id"] = 12591,
		},
		["FROSTFIREBOLT"] = {
			["maxRange"] = 40,
			["minRange"] = 0,
			["castTime"] = 2822,
			["id"] = 44614,
			["icon"] = 236217,
			["name"] = "Frostfire Bolt",
		},
		["DUSKYLEATHERARMOR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Dusky Leather Armor",
			["icon"] = 136247,
			["id"] = 9196,
		},
		["SARONITEMINDCRUSHER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55185,
			["icon"] = 136241,
			["name"] = "Saronite Mindcrusher",
		},
		["RAZORSTRIKEBREASTPLATE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 60649,
			["icon"] = 136247,
			["name"] = "Razorstrike Breastplate",
		},
		["SWIPEBEAR"] = {
			["maxRange"] = 8,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Swipe (Bear)",
			["icon"] = 134296,
			["id"] = 769,
		},
		["DARKLEATHERPANTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Dark Leather Pants",
			["icon"] = 136247,
			["id"] = 7135,
		},
		["CUREDRUGGEDHIDE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Cured Rugged Hide",
			["icon"] = 136243,
			["id"] = 19047,
		},
		["MAJORHEALINGPOTION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Major Healing Potion",
			["icon"] = 136243,
			["id"] = 17556,
		},
		["SOFTSOLEDLINENBOOTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Soft-soled Linen Boots",
			["icon"] = 132149,
			["id"] = 3845,
		},
		["GLYPHOFHUNTERSMARK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Hunter's Mark",
			["icon"] = 136243,
			["id"] = 56829,
		},
		["JUDGEMENTOFLIGHT"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Judgement of Light",
			["icon"] = 135959,
			["id"] = 20185,
		},
		["SEALOFJUSTICE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Seal of Justice",
			["icon"] = 135971,
			["id"] = 20164,
		},
		["JOURNEYMANSCRIBE"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Journeyman Scribe",
			["icon"] = 237171,
			["id"] = 45376,
		},
		["HEAVYMITHRILBREASTPLATE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Heavy Mithril Breastplate",
			["icon"] = 132745, --136192,
			["id"] = 9959,
		},
		["ELIXIROFDETECTDEMON"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Elixir of Detect Demon",
			["icon"] = 136243,
			["id"] = 11478,
		},
		["FROSTSAVAGEBOOTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 59585,
			["icon"] = 136249,
			["name"] = "Frostsavage Boots",
		},
		["PATHOFFROST"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 3714,
			["icon"] = 237528,
			["name"] = "Path of Frost",
		},
		["HANDSTITCHEDLINENBRITCHES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Handstitched Linen Britches",
			["icon"] = 132149,
			["id"] = 3842,
		},
		["GLYPHOFFIRENOVA"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Fire Nova",
			["icon"] = 136243,
			["id"] = 55450,
		},
		["STORMFORGEDAXE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["id"] = 36134,
			["icon"] = 132432, --136192,
			["name"] = "Stormforged Axe",
		},
		["PARRY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 3124,
			["icon"] = 132269,
			["name"] = "Parry",
		},
		["GLYPHOFRUNETAP"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 59327,
			["icon"] = 136243,
			["name"] = "Glyph of Rune Tap",
		},
		["SPELLPOWERGOGGLESXTREME"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["name"] = "Spellpower Goggles Xtreme",
			["icon"] = 136243,
			["id"] = 12615,
		},
		["RUNEDMANABAND"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 64727,
			["icon"] = 134071,
			["name"] = "Runed Mana Band",
		},
		["ARTISANHERBALIST"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Artisan Herbalist",
			["icon"] = 136246,
			["id"] = 11994,
		},
		["PRECISEBLOODSTONE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 54017,
			["icon"] = 134071,
			["name"] = "Precise Bloodstone",
		},
		["COBALTHELM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 52571,
			["icon"] = 136241,
			["name"] = "Cobalt Helm",
		},
		["GNOMISHSHRINKRAY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 12899,
			["icon"] = 136243,
			["name"] = "Gnomish Shrink Ray",
		},
		["GLYPHOFBLINK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Blink",
			["icon"] = 136243,
			["id"] = 56365,
		},
		["SMELTETERNIUM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 1500,
			["id"] = 29359,
			["icon"] = 136243,
			["name"] = "Smelt Eternium",
		},
		["SMALLBRONZEBOMB"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Small Bronze Bomb",
			["icon"] = 136243,
			["id"] = 3941,
		},
		["ICEARMOR"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 1214,
			["icon"] = 135835,
			["name"] = "Ice Armor",
		},
		["FROSTWOVENBOOTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 55906,
			["icon"] = 136249,
			["name"] = "Frostwoven Boots",
		},
		["ARMORVELLUMIII"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 59500,
			["icon"] = 132918,
			["name"] = "Armor Vellum III",
		},
		["DRAGONMAW"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 50000,
			["id"] = 34546,
			["icon"] = 136241,
			["name"] = "Dragonmaw",
		},
		["NOURISH"] = {
			["maxRange"] = 40,
			["minRange"] = 0,
			["castTime"] = 1411,
			["name"] = "Nourish",
			["icon"] = 236162,
			["id"] = 50464,
		},
		["RUNICLEATHERARMOR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Runic Leather Armor",
			["icon"] = 136243,
			["id"] = 19102,
		},
		["LONGSILKENCLOAK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Long Silken Cloak",
			["icon"] = 132149,
			["id"] = 3861,
		},
		["RUNECLOTHHEADBAND"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Runecloth Headband",
			["icon"] = 132149,
			["id"] = 18444,
		},
		["GLYPHOFARCANEMISSILES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Arcane Missiles",
			["icon"] = 136243,
			["id"] = 56363,
		},
		["GREATERBLESSINGOFKINGS"] = {
			["maxRange"] = 40,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Greater Blessing of Kings",
			["icon"] = 135993,
			["id"] = 25898,
		},
		["SHADOWRESISTANCEAURA"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Shadow Resistance Aura",
			["icon"] = 136192,
			["id"] = 19876,
		},
		["STORMBOUNDTOME"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Stormbound Tome",
			["icon"] = 132918,
			["id"] = 59493,
		},
		["SKYSAPPHIREAMULET"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 64726,
			["icon"] = 134071,
			["name"] = "Sky Sapphire Amulet",
		},
		["HONEDCOBALTCLEAVER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55174,
			["icon"] = 136241,
			["name"] = "Honed Cobalt Cleaver",
		},
		["GLYPHOFCONSECRATION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Consecration",
			["icon"] = 136243,
			["id"] = 54928,
		},
		["DIVINEPLEA"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Divine Plea",
			["icon"] = 237537,
			["id"] = 54428,
		},
		["SPELLPOWERGOGGLESXTREMEPLUS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["name"] = "Spellpower Goggles Xtreme Plus",
			["icon"] = 136243,
			["id"] = 19794,
		},
		["SMELTSILVER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 1500,
			["name"] = "Smelt Silver",
			["icon"] = 136243,
			["id"] = 2658,
		},
		["GNOMISHBATTLEGOGGLES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 50000,
			["id"] = 30575,
			["icon"] = 136243,
			["name"] = "Gnomish Battle Goggles",
		},
		["CLEANSE"] = {
			["maxRange"] = 40,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Cleanse",
			["icon"] = 135953,
			["id"] = 4987,
		},
		["MASTERMINER"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 29355,
			["icon"] = 136248,
			["name"] = "Master Miner",
		},
		["ENCHANTEDPEARL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 56530,
			["icon"] = 134071,
			["name"] = "Enchanted Pearl",
		},
		["SEALOFWISDOM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Seal of Wisdom",
			["icon"] = 135960,
			["id"] = 20166,
		},
		["GREATERBLESSINGOFWISDOM"] = {
			["maxRange"] = 40,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Greater Blessing of Wisdom",
			["icon"] = 135912,
			["id"] = 25894,
		},
		["TRAVELFORM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Travel Form",
			["icon"] = 132144,
			["id"] = 783,
		},
		["BLESSINGOFWISDOM"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Blessing of Wisdom",
			["icon"] = 135970,
			["id"] = 19742,
		},
		["GLYPHOFFROSTTRAP"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 56847,
			["icon"] = 136243,
			["name"] = "Glyph of Frost Trap",
		},
		["FELIRONROD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 32655,
			["icon"] = 134926, --136192,
			["name"] = "Fel Iron Rod",
		},
		["EMBOSSEDLEATHERBOOTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Embossed Leather Boots",
			["icon"] = 136247,
			["id"] = 2161,
		},
		["SCROLLOFSTAMINAII"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Scroll of Stamina II",
			["icon"] = 132918,
			["id"] = 50612,
		},
		["DEFENSE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Defense",
			["icon"] = 132279,
			["id"] = 204,
		},
		["GNOMISHHARMPREVENTIONBELT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["id"] = 12903,
			["icon"] = 136243,
			["name"] = "Gnomish Harm Prevention Belt",
		},
		["BRIGHTBLOODGARNET"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 34590,
			["icon"] = 136243,
			["name"] = "Bright Blood Garnet",
		},
		["TRANQUILITY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Tranquility",
			["icon"] = 136107,
			["id"] = 740,
		},
		["SCROLLOFSPIRITV"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Scroll of Spirit V",
			["icon"] = 132918,
			["id"] = 50608,
		},
		["NIGHTSCAPETUNIC"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Nightscape Tunic",
			["icon"] = 136247,
			["id"] = 10499,
		},
		["DIREBEARFORM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Dire Bear Form",
			["icon"] = 132276,
			["id"] = 9634,
		},
		["BOLTOFFROSTWEAVE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 55899,
			["icon"] = 136249,
			["name"] = "Bolt of Frostweave",
		},
		["MOONCLEAVER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 50000,
			["id"] = 34544,
			["icon"] = 136241,
			["name"] = "Mooncleaver",
		},
		["SOLIDSTONESTATUE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Solid Stone Statue",
			["icon"] = 134230,
			["id"] = 32804,
		},
		["CUREDTHICKHIDE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Cured Thick Hide",
			["icon"] = 136247,
			["id"] = 10482,
		},
		["ROYALGUIDEOFESCAPEROUTES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Royal Guide of Escape Routes",
			["icon"] = 132918,
			["id"] = 59486,
		},
		["BARKSKIN"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Barkskin",
			["icon"] = 136063,
			["id"] = 20655,
		},
		["COBALTBOOTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 52569,
			["icon"] = 136241,
			["name"] = "Cobalt Boots",
		},
		["SMELTKHORIUM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 1500,
			["id"] = 29361,
			["icon"] = 136243,
			["name"] = "Smelt Khorium",
		},
		["DASH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Dash",
			["icon"] = 132120,
			["id"] = 1850,
		},
		["LESSERFLASKOFTOUGHNESS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 53752,
			["icon"] = 236876,
			["name"] = "Lesser Flask of Toughness",
		},
		["BIGBLACKMACE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Big Black Mace",
			["icon"] = 133490, --136192,
			["id"] = 10001,
		},
		["GOLDENROD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Golden Rod",
			["icon"] = 135147, --136192,
			["id"] = 14379,
		},
		["CUREDLIGHTHIDE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Cured Light Hide",
			["icon"] = 136247,
			["id"] = 3816,
		},
		["FROSTSHOCK"] = {
			["maxRange"] = 25,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 8056,
			["icon"] = 135849,
			["name"] = "Frost Shock",
		},
		["SMELTTHORIUM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 1500,
			["name"] = "Smelt Thorium",
			["icon"] = 136243,
			["id"] = 16153,
		},
		["GREATEARTHFORGEDHAMMER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["id"] = 36137,
			["icon"] = 133046, --136192,
			["name"] = "Great Earthforged Hammer",
		},
		["HUNTERSINK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Hunter's Ink",
			["icon"] = 132918,
			["id"] = 57703,
		},
		["FELIRONPLATEGLOVES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 29545,
			["icon"] = 132937, --136192,
			["name"] = "Fel Iron Plate Gloves",
		},
		["RETICULATEDARMORWEBBING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 63770,
			["icon"] = 136243,
			["name"] = "Reticulated Armor Webbing",
		},
		["PEARLCLASPEDCLOAK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Pearl-clasped Cloak",
			["icon"] = 132149,
			["id"] = 6521,
		},
		["BATTLESTANCE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Battle Stance",
			["icon"] = 132349,
			["id"] = 2457,
		},
		["LACERATE"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Lacerate",
			["icon"] = 136231,
			["id"] = 5422,
		},
		["NERUBIANBELT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 50961,
			["icon"] = 136247,
			["name"] = "Nerubian Belt",
		},
		["GRILLEDBONESCALE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 45561,
			["icon"] = 133971,
			["name"] = "Grilled Bonescale",
		},
		["SCROLLOFSTAMINAIII"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Scroll of Stamina III",
			["icon"] = 132918,
			["id"] = 50614,
		},
		["PRACTICELOCK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Practice Lock",
			["icon"] = 136243,
			["id"] = 8334,
		},
		["SIMPLEDRESS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Simple Dress",
			["icon"] = 132149,
			["id"] = 8465,
		},
		["SHADOWBOLT"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 1599,
			["id"] = 686,
			["icon"] = 136197,
			["name"] = "Shadow Bolt",
		},
		["MERCURIALALCHEMISTSTONE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 50000,
			["id"] = 60396,
			["icon"] = 136240,
			["name"] = "Mercurial Alchemist Stone",
		},
		["SLOWFALL"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 130,
			["icon"] = 135992,
			["name"] = "Slow Fall",
		},
		["SERPENTSTING"] = {
			["maxRange"] = 35,
			["minRange"] = 5,
			["castTime"] = -999500,
			["id"] = 1978,
			["icon"] = 132204,
			["name"] = "Serpent Sting",
		},
		["IMMOLATE"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 1882,
			["id"] = 348,
			["icon"] = 135817,
			["name"] = "Immolate",
		},
		["GNOMISHXRAYSPECS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["id"] = 56473,
			["icon"] = 136243,
			["name"] = "Gnomish X-Ray Specs",
		},
		["DARKMOONCARDOFTHENORTH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 59504,
			["icon"] = 132918,
			["name"] = "Darkmoon Card of the North",
		},
		["DAZZLINGDARKJADE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53926,
			["icon"] = 134071,
			["name"] = "Dazzling Dark Jade",
		},
		["CHAINLIGHTNING"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 1882,
			["id"] = 421,
			["icon"] = 136015,
			["name"] = "Chain Lightning",
		},
		["FROSTSCALELEGGINGS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 50951,
			["icon"] = 136247,
			["name"] = "Frostscale Leggings",
		},
		["BULWARKOFTHEANCIENTKINGS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 50000,
			["id"] = 36257,
			["icon"] = 136241,
			["name"] = "Bulwark of the Ancient Kings",
		},
		["BASH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Bash",
			["icon"] = 132114,
			["id"] = 5211,
		},
		["GYROMATICMICROADJUSTOR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Gyromatic Micro-Adjustor",
			["icon"] = 136243,
			["id"] = 12590,
		},
		["SCROLLOFAGILITYVI"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Scroll of Agility VI",
			["icon"] = 132918,
			["id"] = 58481,
		},
		["WINDSHEAR"] = {
			["maxRange"] = 25,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 57994,
			["icon"] = 136018,
			["name"] = "Wind Shear",
		},
		["MINORINSCRIPTIONRESEARCH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Minor Inscription Research",
			["icon"] = 237171,
			["id"] = 61288,
		},
		["HILLMANSSHOULDERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Hillman's Shoulders",
			["icon"] = 136247,
			["id"] = 3768,
		},
		["RUNECLOTHROBE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Runecloth Robe",
			["icon"] = 132149,
			["id"] = 18406,
		},
		["HEALINGTOUCH"] = {
			["maxRange"] = 40,
			["minRange"] = 0,
			["castTime"] = 1411,
			["name"] = "Healing Touch",
			["icon"] = 136041,
			["id"] = 5185,
		},
		["FROSTSTRIKE"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 43568,
			["icon"] = 135846,
			["name"] = "Frost Strike",
		},
		["SHIMMERINGINK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Shimmering Ink",
			["icon"] = 132918,
			["id"] = 57711,
		},
		["LINENBELT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Linen Belt",
			["icon"] = 132149,
			["id"] = 8776,
		},
		["LEGGINGSOFVISCERALSTRIKES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 60660,
			["icon"] = 136247,
			["name"] = "Leggings of Visceral Strikes",
		},
		["LIGHTNINGSHIELD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 324,
			["icon"] = 136051,
			["name"] = "Lightning Shield",
		},
		["CHAINSOFICE"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 1411,
			["id"] = 113,
			["icon"] = 135834,
			["name"] = "Chains of Ice",
		},
		["GRANDMASTERLEATHERWORKER"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 51301,
			["icon"] = 133611,
			["name"] = "Grand Master Leatherworker",
		},
		["TRUESILVERGAUNTLETS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["id"] = 9954,
			["icon"] = 132963, --136192,
			["name"] = "Truesilver Gauntlets",
		},
		["GLYPHOFCLEANSING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Cleansing",
			["icon"] = 136243,
			["id"] = 54935,
		},
		["GLYPHOFEVISCERATE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Eviscerate",
			["icon"] = 136243,
			["id"] = 56802,
		},
		["ROUGHBOOMSTICK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Rough Boomstick",
			["icon"] = 136243,
			["id"] = 3925,
		},
		["KHORIUMPOWERCORE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 30308,
			["icon"] = 136243,
			["name"] = "Khorium Power Core",
		},
		["FROSTSCALECHESTGUARD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 50950,
			["icon"] = 136247,
			["name"] = "Frostscale Chestguard",
		},
		["FURLININGATTACKPOWER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 57683,
			["icon"] = 136247,
			["name"] = "Fur Lining - Attack Power",
		},
		["STANDARDSCOPE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2822,
			["name"] = "Standard Scope",
			["icon"] = 136243,
			["id"] = 3975,
		},
		["PICKLEDFANGTOOTH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 45566,
			["icon"] = 133971,
			["name"] = "Pickled Fangtooth",
		},
		["REJUVENATION"] = {
			["maxRange"] = 40,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Rejuvenation",
			["icon"] = 136081,
			["id"] = 774,
		},
		["STONESKINTOTEM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 8071,
			["icon"] = 136098,
			["name"] = "Stoneskin Totem",
		},
		["FIERCEHUGECITRINE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53876,
			["icon"] = 134071,
			["name"] = "Fierce Huge Citrine",
		},
		["TITANIUMROD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55732,
			["icon"] = 237444, --136192,
			["name"] = "Titanium Rod",
		},
		["GLYPHOFMONSOON"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Monsoon",
			["icon"] = 136243,
			["id"] = 63056,
		},
		["HIBERNATE"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 1411,
			["name"] = "Hibernate",
			["icon"] = 136090,
			["id"] = 2637,
		},
		["ARMORVELLUM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Armor Vellum",
			["icon"] = 132918,
			["id"] = 52739,
		},
		["FELIRONPLATEBOOTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 29548,
			["icon"] = 132554, --136192,
			["name"] = "Fel Iron Plate Boots",
		},
		["RAISEALLY"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 46619,
			["icon"] = 136143,
			["name"] = "Raise Ally",
		},
		["DAUNTINGLEGPLATES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55303,
			["icon"] = 136241,
			["name"] = "Daunting Legplates",
		},
		["ORNATESARONITEPAULDRONS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 56550,
			["icon"] = 136241,
			["name"] = "Ornate Saronite Pauldrons",
		},
		["DEADLYSARONITEDIRK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55206,
			["icon"] = 136241,
			["name"] = "Deadly Saronite Dirk",
		},
		["HEARTSTRIKE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 55050,
			["icon"] = 135675,
			["name"] = "Heart Strike",
		},
		["DARKLEATHERBELT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Dark Leather Belt",
			["icon"] = 136247,
			["id"] = 3766,
		},
		["CLOAKOFFROZENSPIRITS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 56015,
			["icon"] = 136249,
			["name"] = "Cloak of Frozen Spirits",
		},
		["SCROLLOFSPIRITVII"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 50610,
			["icon"] = 132918,
			["name"] = "Scroll of Spirit VII",
		},
		["JORMUNGARLEGARMOR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3763,
			["id"] = 50901,
			["icon"] = 136247,
			["name"] = "Jormungar Leg Armor",
		},
		["LIGHTEMBERFORGEDHAMMER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["id"] = 36128,
			["icon"] = 133058, --136192,
			["name"] = "Light Emberforged Hammer",
		},
		["MECHANIZEDSNOWGOGGLES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 56465,
			["icon"] = 136243,
			["name"] = "Mechanized Snow Goggles",
		},
		["GLYPHOFSCOURGEIMPRISONMENT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 55690,
			["icon"] = 136243,
			["name"] = "Glyph of Scourge Imprisonment",
		},
		["ARTISANCOOK"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Artisan Cook",
			["icon"] = 133971,
			["id"] = 18261,
		},
		["COPPERCHAINBELT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Copper Chain Belt",
			["icon"] = 132491, --136192,
			["id"] = 2661,
		},
		["BRIGHTCLOTHGLOVES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Brightcloth Gloves",
			["icon"] = 132149,
			["id"] = 18415,
		},
		["GLYPHOFICEBLOCK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Ice Block",
			["icon"] = 136243,
			["id"] = 56372,
		},
		["CHESTPLATEOFCONQUEST"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55186,
			["icon"] = 136241,
			["name"] = "Chestplate of Conquest",
		},
		["ROUGHBRONZEBOOTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Rough Bronze Boots",
			["icon"] = 132535, --136192,
			["id"] = 7817,
		},
		["FROSTSTEELTUBE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 56471,
			["icon"] = 136243,
			["name"] = "Froststeel Tube",
		},
		["RADIANTDARKJADE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53931,
			["icon"] = 134071,
			["name"] = "Radiant Dark Jade",
		},
		["DEEPFROZENCORD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 56020,
			["icon"] = 136249,
			["name"] = "Deep Frozen Cord",
		},
		["BLOODBOIL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 48721,
			["icon"] = 237513,
			["name"] = "Blood Boil",
		},
		["PUISSANTSHADOWCRYSTAL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53870,
			["icon"] = 134071,
			["name"] = "Puissant Shadow Crystal",
		},
		["TIGERSFURY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Tiger's Fury",
			["icon"] = 132242,
			["id"] = 5217,
		},
		["BLACKMAGEWEAVEHEADBAND"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Black Mageweave Headband",
			["icon"] = 132149,
			["id"] = 12072,
		},
		["ELIXIROFGREATERDEFENSE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Elixir of Greater Defense",
			["icon"] = 136243,
			["id"] = 11450,
		},
		["SHININGSPELLTHREAD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3763,
			["id"] = 55630,
			["icon"] = 136243,
			["name"] = "Shining Spellthread",
		},
		["GLYPHOFARCANEBLAST"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 56991,
			["icon"] = 132918,
			["name"] = "Glyph of Arcane Blast",
		},
		["WEAKTROLLSBLOODELIXIR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Weak Troll's Blood Elixir",
			["icon"] = 136243,
			["id"] = 3170,
		},
		["GLYPHOFINNERFIRE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Inner Fire",
			["icon"] = 136243,
			["id"] = 55686,
		},
		["BATTLESHOUT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Battle Shout",
			["icon"] = 132333,
			["id"] = 2048,
		},
		["GUNS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Guns",
			["icon"] = 135610,
			["id"] = 266,
		},
		["RETALIATION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Retaliation",
			["icon"] = 132336,
			["id"] = 20230,
		},
		["BLOODMOON"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 50000,
			["id"] = 36261,
			["icon"] = 136241,
			["name"] = "Bloodmoon",
		},
		["GRANDMASTERALCHEMIST"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 51303,
			["icon"] = 136240,
			["name"] = "Grand Master Alchemist",
		},
		["STRONGTROLLSBLOODELIXIR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Strong Troll's Blood Elixir",
			["icon"] = 136243,
			["id"] = 3176,
		},
		["POLYMORPH"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 1411,
			["id"] = 118,
			["icon"] = 136071,
			["name"] = "Polymorph",
		},
		["TELEPORTIRONFORGE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 9408,
			["id"] = 3562,
			["icon"] = 135757,
			["name"] = "Teleport: Ironforge",
		},
		["ENCHANTEDTHORIUMBLADES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Enchanted Thorium Blades",
			["icon"] = 136241,
			["id"] = 34982,
		},
		["PERSONALELECTROMAGNETICPULSEGENERATOR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 54736,
			["icon"] = 136243,
			["name"] = "Personal Electromagnetic Pulse Generator",
		},
		["ROASTEDWORG"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 45552,
			["icon"] = 133971,
			["name"] = "Roasted Worg",
		},
		["SMELTHARDENEDADAMANTITE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 29686,
			["icon"] = 136243,
			["name"] = "Smelt Hardened Adamantite",
		},
		["SKYFORGEDGREATAXE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["id"] = 36135,
			["icon"] = 132436, --136192,
			["name"] = "Skyforged Great Axe",
		},
		["GHOSTWEAVEPANTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Ghostweave Pants",
			["icon"] = 132149,
			["id"] = 18441,
		},
		["ETHEREALOIL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 62408,
			["icon"] = 132798,
			["name"] = "Ethereal Oil",
		},
		["BOOKOFSURVIVAL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Book of Survival",
			["icon"] = 132918,
			["id"] = 59478,
		},
		["RUNEDCOPPERBELT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Runed Copper Belt",
			["icon"] = 132492, --136192,
			["id"] = 2666,
		},
		["SMOOTHGOLDENDRAENITE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 34069,
			["icon"] = 136243,
			["name"] = "Smooth Golden Draenite",
		},
		["FELCLOTHBOOTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Felcloth Boots",
			["icon"] = 132149,
			["id"] = 18437,
		},
		["MAMMOTHCUTTERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 56474,
			["icon"] = 136243,
			["name"] = "Mammoth Cutters",
		},
		["WOOLENCAPE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Woolen Cape",
			["icon"] = 136249,
			["id"] = 2402,
		},
		["FRENZIEDREGENERATION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Frenzied Regeneration",
			["icon"] = 132091,
			["id"] = 22842,
		},
		["ARTISANRIDING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 34091,
			["icon"] = 136103,
			["name"] = "Artisan Riding",
		},
		["WICKEDHUGECITRINE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53886,
			["icon"] = 134071,
			["name"] = "Wicked Huge Citrine",
		},
		["INSECTSWARM"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Insect Swarm",
			["icon"] = 136045,
			["id"] = 5570,
		},
		["SILVEREDBRONZEGAUNTLETS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Silvered Bronze Gauntlets",
			["icon"] = 132939, --136192,
			["id"] = 3333,
		},
		["DREAMWEAVEVEST"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["name"] = "Dreamweave Vest",
			["icon"] = 132149,
			["id"] = 12070,
		},
		["BRONZEBATTLEAXE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Bronze Battle Axe",
			["icon"] = 132415, --136192,
			["id"] = 9987,
		},
		["GRANDMASTERSKINNER"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 50307,
			["icon"] = 134366,
			["name"] = "Grand Master Skinner",
		},
		["SARONITERAZORHEADS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 56475,
			["icon"] = 136243,
			["name"] = "Saronite Razorheads",
		},
		["CRUSADERAURA"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Crusader Aura",
			["icon"] = 135890,
			["id"] = 32223,
		},
		["CRIMSONSILKPANTALOONS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Crimson Silk Pantaloons",
			["icon"] = 132149,
			["id"] = 8799,
		},
		["RUBYPENDANTOFFIRE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Ruby Pendant of Fire",
			["icon"] = 136243,
			["id"] = 26883,
		},
		["JOURNEYMANHERBALIST"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Journeyman Herbalist",
			["icon"] = 136246,
			["id"] = 2373,
		},
		["SPICEDWOLFMEAT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Spiced Wolf Meat",
			["icon"] = 134021, --136192,
			["id"] = 2539,
		},
		["DIPLOMACY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Diplomacy",
			["icon"] = 134328,
			["id"] = 20599,
		},
		["ARCANITESKELETONKEY"] = {
			["maxRange"] = 5,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Arcanite Skeleton Key",
			["icon"] = 136243,
			["id"] = 19657,
		},
		["GREENTINTEDGOGGLES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Green Tinted Goggles",
			["icon"] = 136243,
			["id"] = 3956,
		},
		["WHITESWASHBUCKLERSSHIRT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "White Swashbuckler's Shirt",
			["icon"] = 132149,
			["id"] = 8483,
		},
		["HEALINGINJECTORKIT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 56476,
			["icon"] = 136243,
			["name"] = "Healing Injector Kit",
		},
		["MERCURIALADAMANTITE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 38068,
			["icon"] = 135732,
			["name"] = "Mercurial Adamantite",
		},
		["MASTERENGINEERSGOGGLES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 50000,
			["name"] = "Master Engineer's Goggles",
			["icon"] = 136243,
			["id"] = 19825,
		},
		["ENCHANTEDTEAR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 56531,
			["icon"] = 134071,
			["name"] = "Enchanted Tear",
		},
		["SMELTTITANIUM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 1500,
			["id"] = 55211,
			["icon"] = 135811,
			["name"] = "Smelt Titanium",
		},
		["ARMORVELLUMII"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Armor Vellum II",
			["icon"] = 132918,
			["id"] = 59499,
		},
		["FLASKOFTHEFROSTWYRM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 53755,
			["icon"] = 236878,
			["name"] = "Flask of the Frost Wyrm",
		},
		["TARGETDUMMY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Target Dummy",
			["icon"] = 136243,
			["id"] = 3932,
		},
		["FELIRONBLOODRING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["id"] = 31048,
			["icon"] = 136243,
			["name"] = "Fel Iron Blood Ring",
		},
		["CLOAKOFTORMENTEDSKIES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 55199,
			["icon"] = 136247,
			["name"] = "Cloak of Tormented Skies",
		},
		["JAGGEDDARKJADE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53916,
			["icon"] = 134071,
			["name"] = "Jagged Dark Jade",
		},
		["TREMORTOTEM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 8143,
			["icon"] = 136108,
			["name"] = "Tremor Totem",
		},
		["THESHATTERER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["id"] = 10003,
			["icon"] = 133055, --136192,
			["name"] = "The Shatterer",
		},
		["GOBLINMININGHELMET"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 12717,
			["icon"] = 136243,
			["name"] = "Goblin Mining Helmet",
		},
		["ENERGIZEDDARKJADE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53930,
			["icon"] = 134071,
			["name"] = "Energized Dark Jade",
		},
		["ICEBORNELEGGINGS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 50939,
			["icon"] = 136247,
			["name"] = "Iceborne Leggings",
		},
		["REINCARNATION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 20608,
			["icon"] = 136080,
			["name"] = "Reincarnation",
		},
		["MASTERTAILOR"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 26791,
			["icon"] = 136249,
			["name"] = "Master Tailor",
		},
		["RETRIBUTION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Retribution",
			["icon"] = 135741,
			["id"] = 49898,
		},
		["TIRELESSSKYFLAREDIAMOND"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 55386,
			["icon"] = 134071,
			["name"] = "Tireless Skyflare Diamond",
		},
		["SCORCH"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 1811,
			["icon"] = 135827,
			["name"] = "Scorch",
		},
		["MEDIUMLEATHER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Medium Leather",
			["icon"] = 136243,
			["id"] = 20648,
		},
		["LESSERHEALINGPOTION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Lesser Healing Potion",
			["icon"] = 136243,
			["id"] = 2337,
		},
		["ELIXIROFAGILITY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Elixir of Agility",
			["icon"] = 136243,
			["id"] = 11449,
		},
		["RIGIDSUNCRYSTAL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53854,
			["icon"] = 134071,
			["name"] = "Rigid Sun Crystal",
		},
		["SCROLLOFINTELLECTVII"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 50603,
			["icon"] = 132918,
			["name"] = "Scroll of Intellect VII",
		},
		["SCROLLOFSPIRITVI"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Scroll of Spirit VI",
			["icon"] = 132918,
			["id"] = 50609,
		},
		["MASTERHERBALIST"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 28696,
			["icon"] = 136246,
			["name"] = "Master Herbalist",
		},
		["TEMPEREDTITANSTEELHELM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55373,
			["icon"] = 136241,
			["name"] = "Tempered Titansteel Helm",
		},
		["MINDFREEZE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 47528,
			["icon"] = 237527,
			["name"] = "Mind Freeze",
		},
		["THEHUMANSPIRIT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "The Human Spirit",
			["icon"] = 132874,
			["id"] = 20598,
		},
		["LESSERMANAPOTION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Lesser Mana Potion",
			["icon"] = 136243,
			["id"] = 3173,
		},
		["GLYPHOFREBIRTH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Rebirth",
			["icon"] = 136243,
			["id"] = 54733,
		},
		["BOLTOFSILKCLOTH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Bolt of Silk Cloth",
			["icon"] = 136243,
			["id"] = 3839,
		},
		["AQUAMARINEPENDANTOFTHEWARRIOR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Aquamarine Pendant of the Warrior",
			["icon"] = 133303,
			["id"] = 26562,
		},
		["WICKEDLEATHERARMOR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Wicked Leather Armor",
			["icon"] = 136243,
			["id"] = 19098,
		},
		["FELIRONCHAINTUNIC"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 29556,
			["icon"] = 132636, --136192,
			["name"] = "Fel Iron Chain Tunic",
		},
		["GRANDMASTERFIRSTAID"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 50299,
			["icon"] = 135966,
			["name"] = "Grand Master First Aid",
		},
		["TEARDROPBLOODGARNET"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 28903,
			["icon"] = 136243,
			["name"] = "Teardrop Blood Garnet",
		},
		["FELIRONGREATSWORD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 29565,
			["icon"] = 135327, --136192,
			["name"] = "Fel Iron Greatsword",
		},
		["TITANIUMSKELETONKEY"] = {
			["maxRange"] = 5,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 59403,
			["icon"] = 136243,
			["name"] = "Titanium Skeleton Key",
		},
		["GLYPHOFSINISTERSTRIKE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Sinister Strike",
			["icon"] = 136130,
			["id"] = 56821,
		},
		["RUGGEDARMORKIT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Rugged Armor Kit",
			["icon"] = 136243,
			["id"] = 19058,
		},
		["LIGHTBLESSEDMITTENS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 56022,
			["icon"] = 136249,
			["name"] = "Light Blessed Mittens",
		},
		["HEAVYWOOLBANDAGE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 1500,
			["name"] = "Heavy Wool Bandage",
			["icon"] = 133687, --136192,
			["id"] = 3278,
		},
		["BRILLIANTTITANSTEELTREADS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55377,
			["icon"] = 136241,
			["name"] = "Brilliant Titansteel Treads",
		},
		["JOURNEYMANFISHING"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Journeyman Fishing",
			["icon"] = 136245,
			["id"] = 7734,
		},
		["ELIXIROFSPIRIT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 53747,
			["icon"] = 134713,
			["name"] = "Elixir of Spirit",
		},
		["FROSTWEAVETUNIC"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Frostweave Tunic",
			["icon"] = 132149,
			["id"] = 18403,
		},
		["BLOODSUNNECKLACE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 56196,
			["icon"] = 134071,
			["name"] = "Blood Sun Necklace",
		},
		["INTERCEPT"] = {
			["maxRange"] = 25,
			["minRange"] = 8,
			["castTime"] = 0,
			["name"] = "Intercept",
			["icon"] = 132307,
			["id"] = 20252,
		},
		["MASTERSSPELLTHREAD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 4000,
			["id"] = 56034,
			["icon"] = 136011,
			["name"] = "Master's Spellthread",
		},
		["PURIFY"] = {
			["maxRange"] = 40,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Purify",
			["icon"] = 135949,
			["id"] = 1152,
		},
		["EBONWEAVEGLOVES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 56027,
			["icon"] = 136249,
			["name"] = "Ebonweave Gloves",
		},
		["MARKOFTHEWILD"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Mark of the Wild",
			["icon"] = 136078,
			["id"] = 1126,
		},
		["JEWELCRAFTING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 25229,
			["icon"] = 134071,
			["name"] = "Jewelcrafting",
		},
		["FAERIEFIREFERAL"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Faerie Fire (Feral)",
			["icon"] = 136033,
			["id"] = 16857,
		},
		["POTIONOFNIGHTMARES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53900,
			["icon"] = 136240,
			["name"] = "Potion of Nightmares",
		},
		["GLYPHOFSWEEPINGSTRIKES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 57168,
			["icon"] = 132918,
			["name"] = "Glyph of Sweeping Strikes",
		},
		["BOLDBLOODGARNET"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 28905,
			["icon"] = 136243,
			["name"] = "Bold Blood Garnet",
		},
		["GRANDMASTERENGINEER"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 51305,
			["icon"] = 136243,
			["name"] = "Grand Master Engineer",
		},
		["CYCLONE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Cyclone",
			["icon"] = 136018,
			["id"] = 29538,
		},
		["NATURERESISTANCE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Nature Resistance",
			["icon"] = 135865,
			["id"] = 4081,
		},
		["ANTIMAGICSHELL"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 1411,
			["id"] = 31662,
			["icon"] = 136120,
			["name"] = "Anti-Magic Shell",
		},
		["MORTALSTRIKE"] = {
			["maxRange"] = 50000,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Mortal Strike",
			["icon"] = 132355,
			["id"] = 9347,
		},
		["VEILEDHUGECITRINE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53883,
			["icon"] = 134071,
			["name"] = "Veiled Huge Citrine",
		},
		["LIFETAP"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 1454,
			["icon"] = 136126,
			["name"] = "Life Tap",
		},
		["FELIRONPLATEBELT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 29547,
			["icon"] = 132510, --136192,
			["name"] = "Fel Iron Plate Belt",
		},
		["GOBLINCONSTRUCTIONHELMET"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 12718,
			["icon"] = 136243,
			["name"] = "Goblin Construction Helmet",
		},
		["AVENGERSSHIELD"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Avenger's Shield",
			["icon"] = 135874,
			["id"] = 31935,
		},
		["PROSPECTING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 31252,
			["icon"] = 134081,
			["name"] = "Prospecting",
		},
		["RUNECLOTHBELT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Runecloth Belt",
			["icon"] = 132149,
			["id"] = 18402,
		},
		["CALLOFTHESPIRITS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 66844,
			["icon"] = 310732,
			["name"] = "Call of the Spirits",
		},
		["COUNTERSPELL"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 2139,
			["icon"] = 135856,
			["name"] = "Counterspell",
		},
		["JUDGEMENTOFWISDOM"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Judgement of Wisdom",
			["icon"] = 236255,
			["id"] = 20186,
		},
		["GLYPHOFTOTEMOFWRATH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Totem of Wrath",
			["icon"] = 136243,
			["id"] = 63280,
		},
		["EARTHSHADOWRING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 58143,
			["icon"] = 134071,
			["name"] = "Earthshadow Ring",
		},
		["ARCLIGHTSPANNER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Arclight Spanner",
			["icon"] = 136243,
			["id"] = 7430,
		},
		["LIFEBLOOD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 55428,
			["icon"] = 237556,
			["name"] = "Lifeblood",
		},
		["SCROLLOFAGILITYVII"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 58482,
			["icon"] = 132918,
			["name"] = "Scroll of Agility VII",
		},
		["GRANDMASTERJEWELCRAFTER"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 51310,
			["icon"] = 134072,
			["name"] = "Grand Master Jewelcrafter",
		},
		["COYOTESTEAK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Coyote Steak",
			["icon"] = 134021, --136192,
			["id"] = 2541,
		},
		["VOLATILEBLASTINGTRIGGER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 53281,
			["icon"] = 136243,
			["name"] = "Volatile Blasting Trigger",
		},
		["TEMPEREDSARONITEHELM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 54555,
			["icon"] = 136241,
			["name"] = "Tempered Saronite Helm",
		},
		["AZURESILKVEST"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Azure Silk Vest",
			["icon"] = 132149,
			["id"] = 3859,
		},
		["ROYALINK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Royal Ink",
			["icon"] = 132918,
			["id"] = 57708,
		},
		["GNOMISHROCKETBOOTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 12905,
			["icon"] = 136243,
			["name"] = "Gnomish Rocket Boots",
		},
		["EXPLOSIVESHEEP"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Explosive Sheep",
			["icon"] = 136243,
			["id"] = 3955,
		},
		["FROSTWOVENLEGGINGS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 56030,
			["icon"] = 136249,
			["name"] = "Frostwoven Leggings",
		},
		["SILVERROD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Silver Rod",
			["icon"] = 135138, --136192,
			["id"] = 7818,
		},
		["EXORCISM"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 1411,
			["name"] = "Exorcism",
			["icon"] = 135903,
			["id"] = 879,
		},
		["EXPERTTAILOR"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Expert Tailor",
			["icon"] = 136249,
			["id"] = 3913,
		},
		["STANCEMASTERY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Stance Mastery",
			["icon"] = 136031,
			["id"] = 12678,
		},
		["GREATFEAST"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 45554,
			["icon"] = 133971,
			["name"] = "Great Feast",
		},
		["INLAIDMITHRILCYLINDERPLANS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 12895,
			["icon"] = 136243,
			["name"] = "Inlaid Mithril Cylinder Plans",
		},
		["GYROBALANCEDKHORIUMDESTROYER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 50000,
			["id"] = 41307,
			["icon"] = 136243,
			["name"] = "Gyro-balanced Khorium Destroyer",
		},
		["GLYPHOFTURNEVIL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 54931,
			["icon"] = 136243,
			["name"] = "Glyph of Turn Evil",
		},
		["TURTLESCALEBREASTPLATE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Turtle Scale Breastplate",
			["icon"] = 136247,
			["id"] = 10511,
		},
		["HEAVYBRONZEMACE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Heavy Bronze Mace",
			["icon"] = 133483, --136192,
			["id"] = 3296,
		},
		["CONJUREWATER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2822,
			["id"] = 5504,
			["icon"] = 132793,
			["name"] = "Conjure Water",
		},
		["SHADOWMELD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Shadowmeld",
			["icon"] = 132089,
			["id"] = 58984,
		},
		["ARTISANALCHEMIST"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Artisan Alchemist",
			["icon"] = 136240,
			["id"] = 11612,
		},
		["DUSKWEAVEGLOVES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 55922,
			["icon"] = 136249,
			["name"] = "Duskweave Gloves",
		},
		["SPIKEDCOBALTBOOTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 54918,
			["icon"] = 136241,
			["name"] = "Spiked Cobalt Boots",
		},
		["EXPERTENGINEER"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Expert Engineer",
			["icon"] = 136243,
			["id"] = 4041,
		},
		["ONSLAUGHTRING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["name"] = "Onslaught Ring",
			["icon"] = 136243,
			["id"] = 26907,
		},
		["PURIFICATIONPOTION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Purification Potion",
			["icon"] = 136243,
			["id"] = 17572,
		},
		["HEAVYLEATHERAMMOPOUCH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Heavy Leather Ammo Pouch",
			["icon"] = 136247,
			["id"] = 9194,
		},
		["SAVAGESARONITEPAULDRONS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55306,
			["icon"] = 136241,
			["name"] = "Savage Saronite Pauldrons",
		},
		["SMELTTRUESILVER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 1500,
			["name"] = "Smelt Truesilver",
			["icon"] = 136243,
			["id"] = 10098,
		},
		["GLYPHOFDISPELMAGIC"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Dispel Magic",
			["icon"] = 136243,
			["id"] = 55677,
		},
		["GLYPHOFDETERRENCE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Deterrence",
			["icon"] = 136243,
			["id"] = 56850,
		},
		["SARONITESHIV"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55181,
			["icon"] = 136241,
			["name"] = "Saronite Shiv",
		},
		["COOKING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 2550,
			["icon"] = 133971,
			["name"] = "Cooking",
		},
		["BROWNLINENVEST"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Brown Linen Vest",
			["icon"] = 136243,
			["id"] = 2385,
		},
		["FROSTSCALESHOULDERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 50952,
			["icon"] = 136247,
			["name"] = "Frostscale Shoulders",
		},
		["BLINK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 1953,
			["icon"] = 135736,
			["name"] = "Blink",
		},
		["GLYPHOFENTANGLINGROOTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Glyph of Entangling Roots",
			["icon"] = 132918,
			["id"] = 48121,
		},
		["GIFTOFTHENAARU"] = {
			["maxRange"] = 40,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 28880,
			["icon"] = 135923,
			["name"] = "Gift of the Naaru",
		},
		["RUNEOFSWORDBREAKING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 54446,
			["icon"] = 132269,
			["name"] = "Rune of Swordbreaking",
		},
		["NERUBIANLEGGUARDS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 50957,
			["icon"] = 136247,
			["name"] = "Nerubian Legguards",
		},
		["GLOOMBAND"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Gloom Band",
			["icon"] = 136243,
			["id"] = 25287,
		},
		["HEAVYEARTHFORGEDBREASTPLATE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["id"] = 36129,
			["icon"] = 132741, --136192,
			["name"] = "Heavy Earthforged Breastplate",
		},
		["TURTLESCALELEGGINGS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Turtle Scale Leggings",
			["icon"] = 136247,
			["id"] = 10556,
		},
		["GLACIALSLIPPERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 60994,
			["icon"] = 136249,
			["name"] = "Glacial Slippers",
		},
		["FERALCOMBAT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Feral Combat",
			["icon"] = 135958,
			["id"] = 61977,
		},
		["BLACKPLANAREDGE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 50000,
			["id"] = 34542,
			["icon"] = 136241,
			["name"] = "Black Planar Edge",
		},
		["GNOMISHPOULTRYIZER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["id"] = 30569,
			["icon"] = 136243,
			["name"] = "Gnomish Poultryizer",
		},
		["EARTHENLEGARMOR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3763,
			["id"] = 62447,
			["icon"] = 136247,
			["name"] = "Earthen Leg Armor",
		},
		["FROSTWOVENWRISTWRAPS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 56031,
			["icon"] = 136249,
			["name"] = "Frostwoven Wristwraps",
		},
		["RAISEDEAD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 941,
			["id"] = 17473,
			["icon"] = 136187,
			["name"] = "Raise Dead",
		},
		["APPRENTICECOOK"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Apprentice Cook",
			["icon"] = 133971,
			["id"] = 2551,
		},
		["TIMELESSDARKJADE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53894,
			["icon"] = 134071,
			["name"] = "Timeless Dark Jade",
		},
		["LIGHTSKYFORGEDAXE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["id"] = 36126,
			["icon"] = 132454, --136192,
			["name"] = "Light Skyforged Axe",
		},
		["ICELANCE"] = {
			["maxRange"] = 36,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 30455,
			["icon"] = 135844,
			["name"] = "Ice Lance",
		},
		["HEX"] = {
			["maxRange"] = 20,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 11641,
			["icon"] = 136071,
			["name"] = "Hex",
		},
		["PLATEMAIL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Plate Mail",
			["icon"] = 132736,
			["id"] = 750,
		},
		["SHININGSILVERBREASTPLATE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Shining Silver Breastplate",
			["icon"] = 132750, --136192,
			["id"] = 2675,
		},
		["DEMONSKIN"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 687,
			["icon"] = 136185,
			["name"] = "Demon Skin",
		},
		["WOOLENBOOTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Woolen Boots",
			["icon"] = 136249,
			["id"] = 2401,
		},
		["MYSTICFROSTWOVENWRISTWRAPS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 55913,
			["icon"] = 136249,
			["name"] = "Mystic Frostwoven Wristwraps",
		},
		["SARONITESPELLBLADE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 59442,
			["icon"] = 136241,
			["name"] = "Saronite Spellblade",
		},
		["BLACKCHITINGUARDBOOTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 51568,
			["icon"] = 136247,
			["name"] = "Black Chitinguard Boots",
		},
		["IRONBUCKLE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Iron Buckle",
			["icon"] = 133607, --136192,
			["id"] = 8768,
		},
		["ADVANCEDTARGETDUMMY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Advanced Target Dummy",
			["icon"] = 136243,
			["id"] = 3965,
		},
		["GREENLENS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["name"] = "Green Lens",
			["icon"] = 136243,
			["id"] = 12622,
		},
		["GRILLEDSCULPIN"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 45563,
			["icon"] = 133971,
			["name"] = "Grilled Sculpin",
		},
		["LIVINGEMERALDPENDANT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["name"] = "Living Emerald Pendant",
			["icon"] = 136243,
			["id"] = 26911,
		},
		["HOLYSHIELD"] = {
			["maxRange"] = 40,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Holy Shield",
			["icon"] = 135940,
			["id"] = 9800,
		},
		["DARKFROSTSCALELEGGINGS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 60601,
			["icon"] = 136247,
			["name"] = "Dark Frostscale Leggings",
		},
		["LIGHTLEATHERBRACERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Light Leather Bracers",
			["icon"] = 136247,
			["id"] = 9065,
		},
		["EXPERTRIDING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 34090,
			["icon"] = 136103,
			["name"] = "Expert Riding",
		},
		["BOREANLEATHER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 59926,
			["icon"] = 136247,
			["name"] = "Borean Leather",
		},
		["APPRENTICESCRIBE"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Apprentice Scribe",
			["icon"] = 237171,
			["id"] = 45375,
		},
		["MANASPRINGTOTEM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 5675,
			["icon"] = 136053,
			["name"] = "Mana Spring Totem",
		},
		["FARSIGHT"] = {
			["maxRange"] = 50000,
			["minRange"] = 0,
			["castTime"] = 1882,
			["id"] = 6196,
			["icon"] = 136034,
			["name"] = "Far Sight",
		},
		["RINGOFSILVERMIGHT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Ring of Silver Might",
			["icon"] = 136243,
			["id"] = 25317,
		},
		["NETHERWEAVEBANDAGE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 1500,
			["id"] = 27032,
			["icon"] = 133691, --136192,
			["name"] = "Netherweave Bandage",
		},
		["WIZARDWEAVETURBAN"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Wizardweave Turban",
			["icon"] = 132149,
			["id"] = 18450,
		},
		["WILLTOSURVIVE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Will to Survive",
			["icon"] = 136129,
			["id"] = 59752,
		},
		["SCROLLOFRECALLIII"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 60337,
			["icon"] = 132918,
			["name"] = "Scroll of Recall III",
		},
		["NERUBIANLEGREINFORCEMENTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3763,
			["id"] = 50904,
			["icon"] = 136247,
			["name"] = "Nerubian Leg Reinforcements",
		},
		["FLASKOFPUREMOJO"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 54212,
			["icon"] = 236877,
			["name"] = "Flask of Pure Mojo",
		},
		["HEAVYSTONESTATUE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Heavy Stone Statue",
			["icon"] = 134230,
			["id"] = 32803,
		},
		["FROSTSCALEBRACERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 60599,
			["icon"] = 136247,
			["name"] = "Frostscale Bracers",
		},
		["SCARLETSIGNET"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 64728,
			["icon"] = 134071,
			["name"] = "Scarlet Signet",
		},
		["SUREFIRESHURIKEN"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55202,
			["icon"] = 136241,
			["name"] = "Sure-fire Shuriken",
		},
		["FOREMANSENCHANTEDHELMET"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 30515,
			["icon"] = 135933,
			["name"] = "Foreman's Enchanted Helmet",
		},
		["GLYPHOFWRATH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Wrath",
			["icon"] = 136243,
			["id"] = 54756,
		},
		["SILKHEADBAND"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Silk Headband",
			["icon"] = 132149,
			["id"] = 8762,
		},
		["SIMPLEKILT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Simple Kilt",
			["icon"] = 132149,
			["id"] = 12046,
		},
		["FELSHARPENINGSTONE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 29654,
			["icon"] = 135253, --136192,
			["name"] = "Fel Sharpening Stone",
		},
		["LAVAFORGEDWARHAMMER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["id"] = 36136,
			["icon"] = 133054, --136192,
			["name"] = "Lavaforged Warhammer",
		},
		["JADEFIREINK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Jadefire Ink",
			["icon"] = 132918,
			["id"] = 57707,
		},
		["GHOSTWEAVEGLOVES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Ghostweave Gloves",
			["icon"] = 132149,
			["id"] = 18413,
		},
		["FELCLOTHSHOULDERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Felcloth Shoulders",
			["icon"] = 132149,
			["id"] = 18453,
		},
		["BRONZESETTING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Bronze Setting",
			["icon"] = 136243,
			["id"] = 25278,
		},
		["FROSTNOVA"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 122,
			["icon"] = 135848,
			["name"] = "Frost Nova",
		},
		["SPIKEDTITANSTEELTREADS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55375,
			["icon"] = 136241,
			["name"] = "Spiked Titansteel Treads",
		},
		["LIGHTLEATHERQUIVER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Light Leather Quiver",
			["icon"] = 136247,
			["id"] = 9060,
		},
		["SHIELDSLAM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Shield Slam",
			["icon"] = 132357,
			["id"] = 8242,
		},
		["RUGGEDLEATHER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Rugged Leather",
			["icon"] = 136243,
			["id"] = 22331,
		},
		["ORANGEMAGEWEAVESHIRT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Orange Mageweave Shirt",
			["icon"] = 132149,
			["id"] = 12061,
		},
		["ADAMANTITEFRAME"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 30306,
			["icon"] = 136243,
			["name"] = "Adamantite Frame",
		},
		["CRYSTALCITRINENECKLACE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 58141,
			["icon"] = 134071,
			["name"] = "Crystal Citrine Necklace",
		},
		["SIMPLEOPALRING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["name"] = "Simple Opal Ring",
			["icon"] = 136243,
			["id"] = 26902,
		},
		["LUMINOUSHUGECITRINE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53881,
			["icon"] = 134071,
			["name"] = "Luminous Huge Citrine",
		},
		["DEFTHUGECITRINE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53880,
			["icon"] = 134071,
			["name"] = "Deft Huge Citrine",
		},
		["FURY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Fury",
			["icon"] = 135882,
			["id"] = 40601,
		},
		["GOBLINROCKETBOOTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 8892,
			["icon"] = 135805,
			["name"] = "Goblin Rocket Boots",
		},
		["COBALTFRAGBOMB"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 56460,
			["icon"] = 136243,
			["name"] = "Cobalt Frag Bomb",
		},
		["GLYPHOFSERPENTSTING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Serpent Sting",
			["icon"] = 136243,
			["id"] = 56832,
		},
		["RECKLESSNESS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Recklessness",
			["icon"] = 132109,
			["id"] = 1719,
		},
		["TURNEVIL"] = {
			["maxRange"] = 20,
			["minRange"] = 0,
			["castTime"] = 1411,
			["name"] = "Turn Evil",
			["icon"] = 135983,
			["id"] = 10326,
		},
		["GLYPHOFPAINSUPPRESSION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Pain Suppression",
			["icon"] = 136243,
			["id"] = 63248,
		},
		["GLYPHOFMANAGEM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Mana Gem",
			["icon"] = 136243,
			["id"] = 56367,
		},
		["DEMORALIZINGSHOUT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Demoralizing Shout",
			["icon"] = 132366,
			["id"] = 1160,
		},
		["ORNATESARONITEBRACERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 56549,
			["icon"] = 136241,
			["name"] = "Ornate Saronite Bracers",
		},
		["AUTOSHOT"] = {
			["maxRange"] = 35,
			["minRange"] = 5,
			["castTime"] = -999500,
			["id"] = 75,
			["icon"] = 135485,
			["name"] = "Auto Shot",
		},
		["GOLDENSKELETONKEY"] = {
			["maxRange"] = 5,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Golden Skeleton Key",
			["icon"] = 136243,
			["id"] = 19649,
		},
		["GLYPHOFMUTILATE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Mutilate",
			["icon"] = 136243,
			["id"] = 63268,
		},
		["POUNCE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Pounce",
			["icon"] = 132142,
			["id"] = 9005,
		},
		["GLYPHOFSHADOWWORDPAIN"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 55681,
			["icon"] = 136243,
			["name"] = "Glyph of Shadow Word: Pain",
		},
		["WHITEWOOLENDRESS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "White Woolen Dress",
			["icon"] = 132149,
			["id"] = 8467,
		},
		["GREENIRONHELM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Green Iron Helm",
			["icon"] = 133071, --136192,
			["id"] = 3502,
		},
		["ELIXIROFGREATERAGILITY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Elixir of Greater Agility",
			["icon"] = 136243,
			["id"] = 11467,
		},
		["HEAVYSILKBANDAGE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 1500,
			["name"] = "Heavy Silk Bandage",
			["icon"] = 133672, --136192,
			["id"] = 7929,
		},
		["GLYPHOFSEALOFCOMMAND"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 54925,
			["icon"] = 136243,
			["name"] = "Glyph of Seal of Command",
		},
		["FIERYINK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Fiery Ink",
			["icon"] = 132918,
			["id"] = 57710,
		},
		["FURLININGSTAMINA"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 57690,
			["icon"] = 136247,
			["name"] = "Fur Lining - Stamina",
		},
		["POACHEDNETTLEFISH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 45565,
			["icon"] = 133971,
			["name"] = "Poached Nettlefish",
		},
		["TITANSTEELGUARDIAN"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55371,
			["icon"] = 136241,
			["name"] = "Titansteel Guardian",
		},
		["BRILLIANTSARONITEBELT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 59436,
			["icon"] = 136241,
			["name"] = "Brilliant Saronite Belt",
		},
		["COPPERCHAINPANTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Copper Chain Pants",
			["icon"] = 134583, --136192,
			["id"] = 2662,
		},
		["ELEGANTSILVERRING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Elegant Silver Ring",
			["icon"] = 136243,
			["id"] = 25280,
		},
		["GLYPHOFRAKE"] = {
			["maxRange"] = 50000,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 54820,
			["icon"] = 236164,
			["name"] = "Glyph of Rake",
		},
		["LESSERRUNEOFWARDING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2822,
			["id"] = 32274,
			["icon"] = 134424,
			["name"] = "Lesser Rune of Warding",
		},
		["SAVAGESARONITEWALKERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55308,
			["icon"] = 136241,
			["name"] = "Savage Saronite Walkers",
		},
		["GLYPHOFCORRUPTION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Corruption",
			["icon"] = 136243,
			["id"] = 56218,
		},
		["LINENBAG"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Linen Bag",
			["icon"] = 136243,
			["id"] = 3755,
		},
		["DUSKWEAVECOWL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 55919,
			["icon"] = 136249,
			["name"] = "Duskweave Cowl",
		},
		["THEBIGGERONE"] = {
			["maxRange"] = 15,
			["minRange"] = 0,
			["castTime"] = 2822,
			["id"] = 30461,
			["icon"] = 135826,
			["name"] = "The Bigger One",
		},
		["TENUOUSSHADOWCRYSTAL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53861,
			["icon"] = 134071,
			["name"] = "Tenuous Shadow Crystal",
		},
		["FROSTWEAVENET"] = {
			["maxRange"] = 25,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 55536,
			["icon"] = 134325,
			["name"] = "Frostweave Net",
		},
		["SIMPLEPEARLRING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Simple Pearl Ring",
			["icon"] = 136243,
			["id"] = 25284,
		},
		["BOLTOFIMBUEDFROSTWEAVE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 55900,
			["icon"] = 136249,
			["name"] = "Bolt of Imbued Frostweave",
		},
		["FROSTSCALEBELT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 50955,
			["icon"] = 136247,
			["name"] = "Frostscale Belt",
		},
		["RUNECLOTHGLOVES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Runecloth Gloves",
			["icon"] = 132149,
			["id"] = 18417,
		},
		["GNOMISHGOGGLES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["id"] = 12897,
			["icon"] = 136243,
			["name"] = "Gnomish Goggles",
		},
		["SMOKEDSALMON"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 45564,
			["icon"] = 133971,
			["name"] = "Smoked Salmon",
		},
		["MASTERSINSCRIPTIONOFTHEPINNACLE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 4000,
			["id"] = 61119,
			["icon"] = 237171,
			["name"] = "Master's Inscription of the Pinnacle",
		},
		["GLYPHOFSUNDERARMOR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Glyph of Sunder Armor",
			["icon"] = 132918,
			["id"] = 57167,
		},
		["BLUELINENSHIRT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Blue Linen Shirt",
			["icon"] = 132149,
			["id"] = 2394,
		},
		["GLYPHOFEXPOSEARMOR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Expose Armor",
			["icon"] = 136243,
			["id"] = 56803,
		},
		["BRILLIANTPEARLBAND"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 41414,
			["icon"] = 134071,
			["name"] = "Brilliant Pearl Band",
		},
		["MOCKINGBLOW"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Mocking Blow",
			["icon"] = 132350,
			["id"] = 694,
		},
		["SPELLWEAVE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 56003,
			["icon"] = 132149,
			["name"] = "Spellweave",
		},
		["EXPERTHERBALIST"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Expert Herbalist",
			["icon"] = 136246,
			["id"] = 3571,
		},
		["MOONSHROUDGLOVES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 56025,
			["icon"] = 136249,
			["name"] = "Moonshroud Gloves",
		},
		["INTERVENE"] = {
			["maxRange"] = 25,
			["minRange"] = 8,
			["castTime"] = 0,
			["name"] = "Intervene",
			["icon"] = 132365,
			["id"] = 3411,
		},
		["GLYPHOFARCANEPOWER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 56381,
			["icon"] = 136243,
			["name"] = "Glyph of Arcane Power",
		},
		["CLEAVE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Cleave",
			["icon"] = 132338,
			["id"] = 845,
		},
		["TEMPEREDSARONITEBELT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 54551,
			["icon"] = 136241,
			["name"] = "Tempered Saronite Belt",
		},
		["FROSTSCALEHELM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 60600,
			["icon"] = 136247,
			["name"] = "Frostscale Helm",
		},
		["FROSTWEAVEPANTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Frostweave Pants",
			["icon"] = 132149,
			["id"] = 18424,
		},
		["STARFALL"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Starfall",
			["icon"] = 135753,
			["id"] = 20687,
		},
		["GLYPHOFCLAW"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Claw",
			["icon"] = 136243,
			["id"] = 67598,
		},
		["GOBLINROCKETHELMET"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 12758,
			["icon"] = 136243,
			["name"] = "Goblin Rocket Helmet",
		},
		["BLOODPRESENCE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 48266,
			["icon"] = 135770,
			["name"] = "Blood Presence",
		},
		["CLEANSINGTOTEM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 8170,
			["icon"] = 136019,
			["name"] = "Cleansing Totem",
		},
		["BRILLIANTNECKLACE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Brilliant Necklace",
			["icon"] = 136243,
			["id"] = 36523,
		},
		["HEALINGSTREAMTOTEM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 5394,
			["icon"] = 135127,
			["name"] = "Healing Stream Totem",
		},
		["SUNDEREDDARKJADE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53927,
			["icon"] = 134071,
			["name"] = "Sundered Dark Jade",
		},
		["GLYPHOFEVASION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Evasion",
			["icon"] = 136243,
			["id"] = 56799,
		},
		["HEAVYLEATHER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Heavy Leather",
			["icon"] = 136243,
			["id"] = 20649,
		},
		["SPIKEDCOBALTSHOULDERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 54941,
			["icon"] = 136241,
			["name"] = "Spiked Cobalt Shoulders",
		},
		["RUNEOFTHENERUBIANCARAPACE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 70164,
			["icon"] = 135371,
			["name"] = "Rune of the Nerubian Carapace",
		},
		["HEAVYLINENGLOVES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Heavy Linen Gloves",
			["icon"] = 132149,
			["id"] = 3840,
		},
		["FLASHOFLIGHT"] = {
			["maxRange"] = 40,
			["minRange"] = 0,
			["castTime"] = 1411,
			["name"] = "Flash of Light",
			["icon"] = 135907,
			["id"] = 19750,
		},
		["GHOSTWEAVEVEST"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Ghostweave Vest",
			["icon"] = 132149,
			["id"] = 18416,
		},
		["BLESSINGOFKINGS"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Blessing of Kings",
			["icon"] = 135995,
			["id"] = 20217,
		},
		["DRAGONSTRIKE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 50000,
			["id"] = 36262,
			["icon"] = 136241,
			["name"] = "Dragonstrike",
		},
		["GREATERHEALINGPOTION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Greater Healing Potion",
			["icon"] = 136243,
			["id"] = 7181,
		},
		["FIRERESISTANCEAURA"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Fire Resistance Aura",
			["icon"] = 135824,
			["id"] = 19891,
		},
		["CUDGELOFSARONITEJUSTICE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 56280,
			["icon"] = 136241,
			["name"] = "Cudgel of Saronite Justice",
		},
		["ELIXIROFMIGHTYFORTITUDE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 53751,
			["icon"] = 134792,
			["name"] = "Elixir of Mighty Fortitude",
		},
		["RIP"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Rip",
			["icon"] = 132152,
			["id"] = 1079,
		},
		["ICEBORNEBELT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 50943,
			["icon"] = 136247,
			["name"] = "Iceborne Belt",
		},
		["REDSWASHBUCKLERSSHIRT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Red Swashbuckler's Shirt",
			["icon"] = 132149,
			["id"] = 8489,
		},
		["NESINGWARY4000"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 50000,
			["id"] = 60874,
			["icon"] = 136243,
			["name"] = "Nesingwary 4000",
		},
		["FROSTSAVAGEBRACERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 59583,
			["icon"] = 136249,
			["name"] = "Frostsavage Bracers",
		},
		["MAGEARMOR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 6117,
			["icon"] = 135991,
			["name"] = "Mage Armor",
		},
		["TRACKHUMANOIDS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Track Humanoids",
			["icon"] = 132328,
			["id"] = 5225,
		},
		["GLYPHOFRENEW"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Renew",
			["icon"] = 136243,
			["id"] = 55674,
		},
		["ASTRALRECALL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 9408,
			["id"] = 556,
			["icon"] = 136010,
			["name"] = "Astral Recall",
		},
		["ONEHANDEDAXES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "One-Handed Axes",
			["icon"] = 132392,
			["id"] = 196,
		},
		["SCROLLOFSPIRITIII"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Scroll of Spirit III",
			["icon"] = 132918,
			["id"] = 50606,
		},
		["PORTALSTORMWIND"] = {
			["maxRange"] = 10,
			["minRange"] = 0,
			["castTime"] = 9408,
			["id"] = 10059,
			["icon"] = 135748,
			["name"] = "Portal: Stormwind",
		},
		["GLYPHOFHEROICSTRIKE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Glyph of Heroic Strike",
			["icon"] = 132918,
			["id"] = 57158,
		},
		["HEAVYRUNECLOTHBANDAGE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 1500,
			["name"] = "Heavy Runecloth Bandage",
			["icon"] = 133682, --136192,
			["id"] = 18630,
		},
		["LIONHEARTBLADE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 50000,
			["id"] = 34538,
			["icon"] = 136241,
			["name"] = "Lionheart Blade",
		},
		["WICKEDLEATHERPANTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Wicked Leather Pants",
			["icon"] = 136243,
			["id"] = 19083,
		},
		["GREENLEATHERBRACERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Green Leather Bracers",
			["icon"] = 136247,
			["id"] = 3776,
		},
		["HEAVYMITHRILAXE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Heavy Mithril Axe",
			["icon"] = 132405, --136192,
			["id"] = 9993,
		},
		["EXECUTE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Execute",
			["icon"] = 135358,
			["id"] = 5308,
		},
		["DODGE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 81,
			["icon"] = 136047,
			["name"] = "Dodge",
		},
		["HEAVYCOPPERBROADSWORD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Heavy Copper Broadsword",
			["icon"] = 135312, --136192,
			["id"] = 3292,
		},
		["TEMPEREDSARONITEGAUNTLETS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55015,
			["icon"] = 136241,
			["name"] = "Tempered Saronite Gauntlets",
		},
		["EARTHSHIELD"] = {
			["maxRange"] = 50000,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 379,
			["icon"] = 136089,
			["name"] = "Earth Shield",
		},
		["THROW"] = {
			["maxRange"] = 30,
			["minRange"] = 5,
			["castTime"] = 500,
			["id"] = 2764,
			["icon"] = 132324,
			["name"] = "Throw",
		},
		["NERUBIANBOOTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 50960,
			["icon"] = 136247,
			["name"] = "Nerubian Boots",
		},
		["GNOMISHNETOMATICPROJECTOR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 12902,
			["icon"] = 136243,
			["name"] = "Gnomish Net-o-Matic Projector",
		},
		["HEAVYMITHRILSHOULDER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Heavy Mithril Shoulder",
			["icon"] = 135053, --136192,
			["id"] = 9926,
		},
		["GLYPHOFARCANESHOT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Arcane Shot",
			["icon"] = 136243,
			["id"] = 56841,
		},
		["FROSTMOONPANTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 56021,
			["icon"] = 136249,
			["name"] = "Frostmoon Pants",
		},
		["NERUBIANHELM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 60624,
			["icon"] = 136247,
			["name"] = "Nerubian Helm",
		},
		["MAUL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Maul",
			["icon"] = 132136,
			["id"] = 6807,
		},
		["MITHRILGYROSHOT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Mithril Gyro-Shot",
			["icon"] = 136243,
			["id"] = 12621,
		},
		["FELWEIGHTSTONE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 34607,
			["icon"] = 135260, --136192,
			["name"] = "Fel Weightstone",
		},
		["SCROLLOFINTELLECTV"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Scroll of Intellect V",
			["icon"] = 132918,
			["id"] = 50601,
		},
		["TEMPEREDSARONITEBOOTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 54552,
			["icon"] = 136241,
			["name"] = "Tempered Saronite Boots",
		},
		["BARBARICLINENVEST"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Barbaric Linen Vest",
			["icon"] = 132149,
			["id"] = 2395,
		},
		["FELIRONCHAINBRACERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 29553,
			["icon"] = 132612, --136192,
			["name"] = "Fel Iron Chain Bracers",
		},
		["FLASKOFENDLESSRAGE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 53760,
			["icon"] = 236880,
			["name"] = "Flask of Endless Rage",
		},
		["FINDTREASURE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 2481,
			["icon"] = 135725,
			["name"] = "Find Treasure",
		},
		["MANAINJECTORKIT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 56477,
			["icon"] = 136243,
			["name"] = "Mana Injector Kit",
		},
		["GOLDPOWERCORE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Gold Power Core",
			["icon"] = 136243,
			["id"] = 12584,
		},
		["SCROLLOFINTELLECTVIII"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 50604,
			["icon"] = 132918,
			["name"] = "Scroll of Intellect VIII",
		},
		["ROUGHBRONZESHOULDERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Rough Bronze Shoulders",
			["icon"] = 135036, --136192,
			["id"] = 3328,
		},
		["SILKBANDAGE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 1500,
			["name"] = "Silk Bandage",
			["icon"] = 133671, --136192,
			["id"] = 7928,
		},
		["PROTECTION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Protection",
			["icon"] = 132360,
			["id"] = 42206,
		},
		["RAPTORSTRIKE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 2973,
			["icon"] = 132223,
			["name"] = "Raptor Strike",
		},
		["JOURNEYMANLEATHERWORKER"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Journeyman Leatherworker",
			["icon"] = 133611,
			["id"] = 2154,
		},
		["GNOMISHFLAMETURRET"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 30568,
			["icon"] = 136243,
			["name"] = "Gnomish Flame Turret",
		},
		["SOLIDGRINDINGSTONE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Solid Grinding Stone",
			["icon"] = 135246, --136192,
			["id"] = 9920,
		},
		["MASTERSCRIBE"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 45379,
			["icon"] = 237171,
			["name"] = "Master Scribe",
		},
		["GLYPHOFEVOCATION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Evocation",
			["icon"] = 136243,
			["id"] = 56380,
		},
		["EARTHLIVINGWEAPON"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 51730,
			["icon"] = 237575,
			["name"] = "Earthliving Weapon",
		},
		["REDLINENSHIRT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Red Linen Shirt",
			["icon"] = 132149,
			["id"] = 2392,
		},
		["GLYPHOFREVENGE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Glyph of Revenge",
			["icon"] = 132918,
			["id"] = 57165,
		},
		["LIGHTNINGBOLT"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 1411,
			["id"] = 403,
			["icon"] = 136048,
			["name"] = "Lightning Bolt",
		},
		["JUDGEMENTOFJUSTICE"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Judgement of Justice",
			["icon"] = 236258,
			["id"] = 20184,
		},
		["AUTOATTACK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 6603,
			["icon"] = 135703,
			["name"] = "Auto Attack",
		},
		["DARKICEBORNELEGGINGS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 60611,
			["icon"] = 136247,
			["name"] = "Dark Iceborne Leggings",
		},
		["SOLIDBLASTINGPOWDER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Solid Blasting Powder",
			["icon"] = 136243,
			["id"] = 12585,
		},
		["GLYPHOFEARTHSHIELD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Earth Shield",
			["icon"] = 136243,
			["id"] = 63279,
		},
		["GLYPHOFMOONFIRE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Moonfire",
			["icon"] = 237171,
			["id"] = 52085,
		},
		["GLYPHOFHEALTHFUNNEL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Health Funnel",
			["icon"] = 136243,
			["id"] = 56238,
		},
		["HYPERSPEEDACCELERATORS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 54999,
			["icon"] = 136243,
			["name"] = "Hyperspeed Accelerators",
		},
		["FELIRONBREASTPLATE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 29550,
			["icon"] = 132742, --136192,
			["name"] = "Fel Iron Breastplate",
		},
		["GLOVESOFMEDITATION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Gloves of Meditation",
			["icon"] = 132149,
			["id"] = 3852,
		},
		["HEAVYBOREANARMORKIT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2822,
			["id"] = 50909,
			["icon"] = 136247,
			["name"] = "Heavy Borean Armor Kit",
		},
		["GROUNDINGTOTEM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 8177,
			["icon"] = 136039,
			["name"] = "Grounding Totem",
		},
		["DUSKWEAVEBOOTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 55924,
			["icon"] = 136249,
			["name"] = "Duskweave Boots",
		},
		["FROSTRESISTANCEAURA"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Frost Resistance Aura",
			["icon"] = 135865,
			["id"] = 19888,
		},
		["PEARLHANDLEDDAGGER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Pearl-handled Dagger",
			["icon"] = 135641, --136192,
			["id"] = 6517,
		},
		["CRABCAKE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Crab Cake",
			["icon"] = 133950, --136192,
			["id"] = 2544,
		},
		["NETHERWEAVENET"] = {
			["maxRange"] = 25,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 31367,
			["icon"] = 134325,
			["name"] = "Netherweave Net",
		},
		["FIREGUARD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 50000,
			["id"] = 34535,
			["icon"] = 136241,
			["name"] = "Fireguard",
		},
		["NIGHTSCAPEPANTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Nightscape Pants",
			["icon"] = 136247,
			["id"] = 10548,
		},
		["GREENIRONBRACERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Green Iron Bracers",
			["icon"] = 132605, --136192,
			["id"] = 3501,
		},
		["MONARCHCRAB"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 59759,
			["icon"] = 134071,
			["name"] = "Monarch Crab",
		},
		["FROSTGRENADES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 39973,
			["icon"] = 136243,
			["name"] = "Frost Grenades",
		},
		["ARCTICCHESTPIECE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 50944,
			["icon"] = 136247,
			["name"] = "Arctic Chestpiece",
		},
		["GRANDMASTERMINER"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 50309,
			["icon"] = 136248,
			["name"] = "Grand Master Miner",
		},
		["HEAVYADAMANTITERING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["id"] = 31052,
			["icon"] = 136243,
			["name"] = "Heavy Adamantite Ring",
		},
		["SMELTFELSTEEL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 1500,
			["id"] = 29360,
			["icon"] = 136243,
			["name"] = "Smelt Felsteel",
		},
		["EARTHFORGEDLEGGINGS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["id"] = 36122,
			["icon"] = 134693, --136192,
			["name"] = "Earthforged Leggings",
		},
		["GLYPHOFUNBREAKABLEARMOR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 57226,
			["icon"] = 132918,
			["name"] = "Glyph of Unbreakable Armor",
		},
		["THICKLEATHER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Thick Leather",
			["icon"] = 136243,
			["id"] = 20650,
		},
		["LAMBENTDARKJADE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53928,
			["icon"] = 134071,
			["name"] = "Lambent Dark Jade",
		},
		["JOURNEYMANJEWELCRAFTER"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Journeyman Jewelcrafter",
			["icon"] = 134073,
			["id"] = 25246,
		},
		["DIVINESHIELD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Divine Shield",
			["icon"] = 135896,
			["id"] = 642,
		},
		["MECHANICALREPAIRKIT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Mechanical Repair Kit",
			["icon"] = 136243,
			["id"] = 15255,
		},
		["BRILLIANTSARONITELEGPLATES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55055,
			["icon"] = 136241,
			["name"] = "Brilliant Saronite Legplates",
		},
		["MIXOLOGY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Mixology",
			["icon"] = 134735,
			["id"] = 53042,
		},
		["ARCANITEROD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Arcanite Rod",
			["icon"] = 135156, --136192,
			["id"] = 20201,
		},
		["ETHEREALINK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Ethereal Ink",
			["icon"] = 132918,
			["id"] = 57713,
		},
		["SMELTTIN"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 1500,
			["name"] = "Smelt Tin",
			["icon"] = 136243,
			["id"] = 3304,
		},
		["TITANIUMSHIELDSPIKE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2822,
			["id"] = 56353,
			["icon"] = 136243,
			["name"] = "Titanium Shield Spike",
		},
		["EBONWEAVEROBE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 56026,
			["icon"] = 136249,
			["name"] = "Ebonweave Robe",
		},
		["CUREDMEDIUMHIDE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Cured Medium Hide",
			["icon"] = 136247,
			["id"] = 3817,
		},
		["RUNICLEATHERBELT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Runic Leather Belt",
			["icon"] = 136243,
			["id"] = 19072,
		},
		["GLYPHOFSHRED"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Shred",
			["icon"] = 136243,
			["id"] = 54815,
		},
		["ROUGHBRONZECUIRASS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Rough Bronze Cuirass",
			["icon"] = 132630, --136192,
			["id"] = 2670,
		},
		["AZURESILKPANTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Azure Silk Pants",
			["icon"] = 132149,
			["id"] = 8758,
		},
		["STONEGUARDBAND"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 58145,
			["icon"] = 134071,
			["name"] = "Stoneguard Band",
		},
		["NETHERWEAVEPANTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 26771,
			["icon"] = 132149,
			["name"] = "Netherweave Pants",
		},
		["GLYPHOFHAMMEROFJUSTICE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Hammer of Justice",
			["icon"] = 136243,
			["id"] = 54923,
		},
		["ICYPRISM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 62242,
			["icon"] = 134071,
			["name"] = "Icy Prism",
		},
		["GLYPHOFICYTOUCH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Glyph of Icy Touch",
			["icon"] = 132918,
			["id"] = 57219,
		},
		["APPRENTICEMINER"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Apprentice Miner",
			["icon"] = 136248,
			["id"] = 2581,
		},
		["TELEPORTDALARAN"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 941,
			["id"] = 3578,
			["icon"] = 135760,
			["name"] = "Teleport: Dalaran",
		},
		["DEEPFREEZE"] = {
			["maxRange"] = 36,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 44572,
			["icon"] = 236214,
			["name"] = "Deep Freeze",
		},
		["SANCTIFIEDSPELLTHREAD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 4000,
			["id"] = 56039,
			["icon"] = 136011,
			["name"] = "Sanctified Spellthread",
		},
		["SPIDERSILKDRAPE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Spidersilk Drape",
			["icon"] = 132149,
			["id"] = 63742,
		},
		["ARCANEBRILLIANCE"] = {
			["maxRange"] = 40,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 23028,
			["icon"] = 135869,
			["name"] = "Arcane Brilliance",
		},
		["FOREMANSREINFORCEDHELMET"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 30519,
			["icon"] = 135933,
			["name"] = "Foreman's Reinforced Helmet",
		},
		["FIREELEMENTALTOTEM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 2894,
			["icon"] = 135790,
			["name"] = "Fire Elemental Totem",
		},
		["MASTERSINSCRIPTIONOFTHESTORM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 4000,
			["id"] = 61120,
			["icon"] = 237171,
			["name"] = "Master's Inscription of the Storm",
		},
		["FIRESUNDER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 35751,
			["icon"] = 132839,
			["name"] = "Fire Sunder",
		},
		["GLOWINGSHADOWCRYSTAL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53862,
			["icon"] = 134071,
			["name"] = "Glowing Shadow Crystal",
		},
		["FLASKOFTHENORTH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 67016,
			["icon"] = 236879,
			["name"] = "Flask of the North",
		},
		["BALANCEDSHADOWCRYSTAL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53866,
			["icon"] = 134071,
			["name"] = "Balanced Shadow Crystal",
		},
		["EVOCATION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 12051,
			["icon"] = 136075,
			["name"] = "Evocation",
		},
		["FLAMETONGUETOTEM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 8227,
			["icon"] = 136040,
			["name"] = "Flametongue Totem",
		},
		["SOOTHEANIMAL"] = {
			["maxRange"] = 40,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Soothe Animal",
			["icon"] = 132163,
			["id"] = 2908,
		},
		["MAGEWEAVEBAG"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Mageweave Bag",
			["icon"] = 132149,
			["id"] = 12065,
		},
		["BANDOFNATURALFIRE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["id"] = 26916,
			["icon"] = 136243,
			["name"] = "Band of Natural Fire",
		},
		["SEALOFVENGEANCE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Seal of Vengeance",
			["icon"] = 135969,
			["id"] = 31801,
		},
		["BRIGHTCLOTHCLOAK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Brightcloth Cloak",
			["icon"] = 132149,
			["id"] = 18420,
		},
		["BRONZETUBE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Bronze Tube",
			["icon"] = 136243,
			["id"] = 3938,
		},
		["FIREWARD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 543,
			["icon"] = 135806,
			["name"] = "Fire Ward",
		},
		["HEAVYDYNAMITE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Heavy Dynamite",
			["icon"] = 136243,
			["id"] = 3946,
		},
		["NATURERESISTANCETOTEM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 10595,
			["icon"] = 136061,
			["name"] = "Nature Resistance Totem",
		},
		["DUSKYBELT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Dusky Belt",
			["icon"] = 136247,
			["id"] = 9206,
		},
		["SCROLLOFSTRENGTHVII"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 58490,
			["icon"] = 132918,
			["name"] = "Scroll of Strength VII",
		},
		["BOLTOFMAGEWEAVE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Bolt of Mageweave",
			["icon"] = 136249,
			["id"] = 3865,
		},
		["ORNATESARONITESKULLSHIELD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 56556,
			["icon"] = 136241,
			["name"] = "Ornate Saronite Skullshield",
		},
		["BRIGHTCLOTHROBE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Brightcloth Robe",
			["icon"] = 132149,
			["id"] = 18414,
		},
		["GLYPHOFHAMSTRING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Glyph of Hamstring",
			["icon"] = 132918,
			["id"] = 57157,
		},
		["DALARANCLAMCHOWDER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 58065,
			["icon"] = 133971,
			["name"] = "Dalaran Clam Chowder",
		},
		["ESCAPEARTIST"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 20589,
			["icon"] = 132309,
			["name"] = "Escape Artist",
		},
		["SMALLLEATHERAMMOPOUCH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Small Leather Ammo Pouch",
			["icon"] = 136247,
			["id"] = 9062,
		},
		["GLYPHOFAIMEDSHOT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Aimed Shot",
			["icon"] = 136243,
			["id"] = 56824,
		},
		["FLYINGTIGERGOGGLES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Flying Tiger Goggles",
			["icon"] = 136243,
			["id"] = 3934,
		},
		["TEMPEREDSARONITESHOULDERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 54556,
			["icon"] = 136241,
			["name"] = "Tempered Saronite Shoulders",
		},
		["SWIFTSKYFLAREDIAMOND"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 55394,
			["icon"] = 134071,
			["name"] = "Swift Skyflare Diamond",
		},
		["GLYPHOFBARBARICINSULTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Glyph of Barbaric Insults",
			["icon"] = 132918,
			["id"] = 57151,
		},
		["DOUBLESTITCHEDWOOLENSHOULDERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Double-stitched Woolen Shoulders",
			["icon"] = 132149,
			["id"] = 3848,
		},
		["SAVAGESARONITEGAUNTLETS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55309,
			["icon"] = 136241,
			["name"] = "Savage Saronite Gauntlets",
		},
		["TEMPEREDSARONITEBRACERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55017,
			["icon"] = 136241,
			["name"] = "Tempered Saronite Bracers",
		},
		["SHADOWCRYSTALFOCUSINGLENS"] = {
			["maxRange"] = 60,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 56190,
			["icon"] = 134071,
			["name"] = "Shadow Crystal Focusing Lens",
		},
		["FROSTWOVENBELT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 55908,
			["icon"] = 136249,
			["name"] = "Frostwoven Belt",
		},
		["BLASTWAVE"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 1831,
			["icon"] = 135903,
			["name"] = "Blast Wave",
		},
		["GREENLINENBRACERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Green Linen Bracers",
			["icon"] = 132149,
			["id"] = 3841,
		},
		["ROUGHCOPPERBOMB"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Rough Copper Bomb",
			["icon"] = 136243,
			["id"] = 3923,
		},
		["HORNEDCOBALTHELM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 54949,
			["icon"] = 136241,
			["name"] = "Horned Cobalt Helm",
		},
		["GLYPHOFDISENGAGE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Disengage",
			["icon"] = 136243,
			["id"] = 56844,
		},
		["POWERFULEARTHSIEGEDIAMOND"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 55399,
			["icon"] = 134071,
			["name"] = "Powerful Earthsiege Diamond",
		},
		["RUNICFOCUS"] = {
			["maxRange"] = 35,
			["minRange"] = 5,
			["castTime"] = 0,
			["id"] = 53069,
			["icon"] = 136057,
			["name"] = "Runic Focus",
		},
		["SCROLLOFSTRENGTHIII"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Scroll of Strength III",
			["icon"] = 132918,
			["id"] = 58486,
		},
		["HAMSTRING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Hamstring",
			["icon"] = 132316,
			["id"] = 1715,
		},
		["NERUBIANSHOULDERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 50958,
			["icon"] = 136247,
			["name"] = "Nerubian Shoulders",
		},
		["BOOKOFSTARS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Book of Stars",
			["icon"] = 132918,
			["id"] = 59490,
		},
		["SLAM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 1500,
			["name"] = "Slam",
			["icon"] = 132340,
			["id"] = 1464,
		},
		["GREATERMANAPOTION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Greater Mana Potion",
			["icon"] = 136243,
			["id"] = 11448,
		},
		["HILLMANSLEATHERGLOVES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Hillman's Leather Gloves",
			["icon"] = 136247,
			["id"] = 3764,
		},
		["WATERWALKING"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 546,
			["icon"] = 135863,
			["name"] = "Water Walking",
		},
		["RESTORATION"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 11359,
			["icon"] = 135894,
			["name"] = "Restoration",
		},
		["THUNDERSTORM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 51490,
			["icon"] = 237589,
			["name"] = "Thunderstorm",
		},
		["MOONSHROUDROBE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 56024,
			["icon"] = 136249,
			["name"] = "Moonshroud Robe",
		},
		["WISPSPIRIT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Wisp Spirit",
			["icon"] = 136116,
			["id"] = 20585,
		},
		["DARKNERUBIANLEGGINGS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 60627,
			["icon"] = 136247,
			["name"] = "Dark Nerubian Leggings",
		},
		["SCROLLOFSTRENGTHIV"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Scroll of Strength IV",
			["icon"] = 132918,
			["id"] = 58487,
		},
		["RUNEOFTHESTONESKINGARGOYLE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 62158,
			["icon"] = 237480,
			["name"] = "Rune of the Stoneskin Gargoyle",
		},
		["SMELTSTEEL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 1500,
			["name"] = "Smelt Steel",
			["icon"] = 136243,
			["id"] = 3569,
		},
		["GLYPHOFJUDGEMENT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Judgement",
			["icon"] = 136243,
			["id"] = 54922,
		},
		["DUSKWEAVEBELT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 55914,
			["icon"] = 136249,
			["name"] = "Duskweave Belt",
		},
		["PLAGUESTRIKE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 45462,
			["icon"] = 237519,
			["name"] = "Plague Strike",
		},
		["GURUSELIXIR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 53749,
			["icon"] = 134735,
			["name"] = "Guru's Elixir",
		},
		["TOTEMOFWRATH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 30706,
			["icon"] = 135829,
			["name"] = "Totem of Wrath",
		},
		["MASTERSKINNER"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 32679,
			["icon"] = 134366,
			["name"] = "Master Skinner",
		},
		["COPPERMODULATOR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Copper Modulator",
			["icon"] = 136243,
			["id"] = 3926,
		},
		["RUNECLOTHBOOTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Runecloth Boots",
			["icon"] = 132149,
			["id"] = 18423,
		},
		["ETCHEDHUGECITRINE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53873,
			["icon"] = 134071,
			["name"] = "Etched Huge Citrine",
		},
		["CONJUREMANAGEM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2822,
			["id"] = 759,
			["icon"] = 134104,
			["name"] = "Conjure Mana Gem",
		},
		["FELIRONCASING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 30304,
			["icon"] = 136243,
			["name"] = "Fel Iron Casing",
		},
		["HEAVYLINENBANDAGE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 1500,
			["name"] = "Heavy Linen Bandage",
			["icon"] = 133688, --136192,
			["id"] = 3276,
		},
		["GRANDMASTERCOOK"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 51295,
			["icon"] = 133971,
			["name"] = "Grand Master Cook",
		},
		["SPELLREFLECTION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Spell Reflection",
			["icon"] = 136222,
			["id"] = 9941,
		},
		["ICYVEINS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 12472,
			["icon"] = 135838,
			["name"] = "Icy Veins",
		},
		["COARSEGRINDINGSTONE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Coarse Grinding Stone",
			["icon"] = 135244, --136192,
			["id"] = 3326,
		},
		["HEAVYARMORKIT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Heavy Armor Kit",
			["icon"] = 136247,
			["id"] = 3780,
		},
		["OVERPOWER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Overpower",
			["icon"] = 132223,
			["id"] = 7384,
		},
		["ICESTRIKERSCLOAK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 60637,
			["icon"] = 136247,
			["name"] = "Ice Striker's Cloak",
		},
		["WHIRLINGSTEELAXES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Whirling Steel Axes",
			["icon"] = 136241,
			["id"] = 34981,
		},
		["NOTCHEDCOBALTWARAXE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55204,
			["icon"] = 136241,
			["name"] = "Notched Cobalt War Axe",
		},
		["TOUGHENEDLEATHERGLOVES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Toughened Leather Gloves",
			["icon"] = 136247,
			["id"] = 3770,
		},
		["SHININGDARKJADE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53923,
			["icon"] = 134071,
			["name"] = "Shining Dark Jade",
		},
		["EMBOSSEDLEATHERVEST"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Embossed Leather Vest",
			["icon"] = 136247,
			["id"] = 2160,
		},
		["GLYPHOFDEATHSTRIKE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 59336,
			["icon"] = 136243,
			["name"] = "Glyph of Death Strike",
		},
		["BLIZZARD"] = {
			["maxRange"] = 36,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 10,
			["icon"] = 135857,
			["name"] = "Blizzard",
		},
		["GLYPHOFGARROTE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Garrote",
			["icon"] = 136243,
			["id"] = 56812,
		},
		["ENDLESSMANAPOTION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 58868,
			["icon"] = 136240,
			["name"] = "Endless Mana Potion",
		},
		["SMELTING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 2656,
			["icon"] = 135811,
			["name"] = "Smelting",
		},
		["DRUMSOFTHEWILD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 69388,
			["icon"] = 136247,
			["name"] = "Drums of the Wild",
		},
		["GLYPHOFFLAMETONGUEWEAPON"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Flametongue Weapon",
			["icon"] = 136243,
			["id"] = 55451,
		},
		["ICEBANETREADS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 61010,
			["icon"] = 136241,
			["name"] = "Icebane Treads",
		},
		["SCROLLOFINTELLECTIII"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Scroll of Intellect III",
			["icon"] = 132918,
			["id"] = 50599,
		},
		["GREENLINENSHIRT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Green Linen Shirt",
			["icon"] = 132149,
			["id"] = 2396,
		},
		["GLINTINGSTEELDAGGER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Glinting Steel Dagger",
			["icon"] = 135641, --136192,
			["id"] = 15972,
		},
		["SILVERCONTACT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Silver Contact",
			["icon"] = 136243,
			["id"] = 3973,
		},
		["GOLDENDRAENITERING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["id"] = 31049,
			["icon"] = 136243,
			["name"] = "Golden Draenite Ring",
		},
		["FROSTSAVAGECOWL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 59589,
			["icon"] = 136249,
			["name"] = "Frostsavage Cowl",
		},
		["GLYPHOFAMBUSH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 56813,
			["icon"] = 136243,
			["name"] = "Glyph of Ambush",
		},
		["MIRRORIMAGE"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 36847,
			["icon"] = 135757,
			["name"] = "Mirror Image",
		},
		["ETERNALBELTBUCKLE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3763,
			["id"] = 55655,
			["icon"] = 132525,
			["name"] = "Eternal Belt Buckle",
		},
		["MAGNIFICENTFLYINGCARPET"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 20000,
			["id"] = 60971,
			["icon"] = 136249,
			["name"] = "Magnificent Flying Carpet",
		},
		["BULWARKOFKINGS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 50000,
			["id"] = 34534,
			["icon"] = 136241,
			["name"] = "Bulwark of Kings",
		},
		["FRAGBELT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 54793,
			["icon"] = 136243,
			["name"] = "Frag Belt",
		},
		["APPRENTICERIDING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 33388,
			["icon"] = 136103,
			["name"] = "Apprentice Riding",
		},
		["CERTIFICATEOFOWNERSHIP"] = {
			["maxRange"] = 15,
			["minRange"] = 0,
			["castTime"] = 9408,
			["name"] = "Certificate of Ownership",
			["icon"] = 237446,
			["id"] = 59385,
		},
		["VENGEANCEBINDINGS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55298,
			["icon"] = 136241,
			["name"] = "Vengeance Bindings",
		},
		["ELIXIROFMINORACCURACY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Elixir of Minor Accuracy",
			["icon"] = 134786,
			["id"] = 63729,
		},
		["GLYPHOFDISEASE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Disease",
			["icon"] = 136243,
			["id"] = 63334,
		},
		["GLINTINGHUGECITRINE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53878,
			["icon"] = 134071,
			["name"] = "Glinting Huge Citrine",
		},
		["GYROCHRONATOM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Gyrochronatom",
			["icon"] = 136243,
			["id"] = 3961,
		},
		["EXPERTSKINNER"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Expert Skinner",
			["icon"] = 134366,
			["id"] = 8620,
		},
		["HEALINGWAVE"] = {
			["maxRange"] = 40,
			["minRange"] = 0,
			["castTime"] = 1411,
			["id"] = 331,
			["icon"] = 136052,
			["name"] = "Healing Wave",
		},
		["DARKFLAMEINK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 57714,
			["icon"] = 132918,
			["name"] = "Darkflame Ink",
		},
		["AZURESILKBELT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Azure Silk Belt",
			["icon"] = 132149,
			["id"] = 8766,
		},
		["SACREDSHIELD"] = {
			["maxRange"] = 40,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Sacred Shield",
			["icon"] = 236249,
			["id"] = 53601,
		},
		["SAVAGESARONITEWAISTGUARD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55307,
			["icon"] = 136241,
			["name"] = "Savage Saronite Waistguard",
		},
		["SAUTEEDGOBY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 45562,
			["icon"] = 133971,
			["name"] = "Sauteed Goby",
		},
		["ROUGHBRONZELEGGINGS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Rough Bronze Leggings",
			["icon"] = 134583, --136192,
			["id"] = 2668,
		},
		["LESSERWIZARDSROBE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Lesser Wizard's Robe",
			["icon"] = 132149,
			["id"] = 6690,
		},
		["HEAVYNETHERWEAVEBANDAGE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 1500,
			["id"] = 27033,
			["icon"] = 133692, --136192,
			["name"] = "Heavy Netherweave Bandage",
		},
		["IRONBOUNDTOME"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 59497,
			["icon"] = 132918,
			["name"] = "Iron-bound Tome",
		},
		["ANCESTRALSPIRIT"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 9408,
			["id"] = 2008,
			["icon"] = 136077,
			["name"] = "Ancestral Spirit",
		},
		["MASTERFISHING"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 18816,
			["id"] = 33100,
			["icon"] = 136245,
			["name"] = "Master Fishing",
		},
		["SOVEREIGNSHADOWCRYSTAL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53859,
			["icon"] = 134071,
			["name"] = "Sovereign Shadow Crystal",
		},
		["RIGHTEOUSFURY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Righteous Fury",
			["icon"] = 135962,
			["id"] = 25780,
		},
		["ARCANETAROT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Arcane Tarot",
			["icon"] = 132918,
			["id"] = 59487,
		},
		["ARCTICBOOTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 50948,
			["icon"] = 136247,
			["name"] = "Arctic Boots",
		},
		["WINDFORGEDLEGGINGS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["id"] = 36124,
			["icon"] = 134662, --136192,
			["name"] = "Windforged Leggings",
		},
		["JOURNEYMANSKINNER"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Journeyman Skinner",
			["icon"] = 134366,
			["id"] = 8619,
		},
		["GLYPHOFHEALINGTOUCH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Healing Touch",
			["icon"] = 136243,
			["id"] = 54825,
		},
		["RUNICLEATHERHEADBAND"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Runic Leather Headband",
			["icon"] = 136243,
			["id"] = 19082,
		},
		["AMPLIFYMAGIC"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 1008,
			["icon"] = 135907,
			["name"] = "Amplify Magic",
		},
		["GREATERBLESSINGOFMIGHT"] = {
			["maxRange"] = 40,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Greater Blessing of Might",
			["icon"] = 135908,
			["id"] = 25782,
		},
		["DARKCOMMAND"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 56222,
			["icon"] = 136088,
			["name"] = "Dark Command",
		},
		["HAMMEROFJUSTICE"] = {
			["maxRange"] = 10,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Hammer of Justice",
			["icon"] = 135963,
			["id"] = 853,
		},
		["DREAMSIGNET"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 56197,
			["icon"] = 134071,
			["name"] = "Dream Signet",
		},
		["BRILLIANTGLASS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 47280,
			["icon"] = 134071,
			["name"] = "Brilliant Glass",
		},
		["PURGE"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 370,
			["icon"] = 136075,
			["name"] = "Purge",
		},
		["GLINTINGFLAMESPESSARITE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 28914,
			["icon"] = 136243,
			["name"] = "Glinting Flame Spessarite",
		},
		["FROSTARMOR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 168,
			["icon"] = 135843,
			["name"] = "Frost Armor",
		},
		["ENRAGE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Enrage",
			["icon"] = 132126,
			["id"] = 5229,
		},
		["SMELTSARONITE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 1500,
			["id"] = 49258,
			["icon"] = 135811,
			["name"] = "Smelt Saronite",
		},
		["INDESTRUCTIBLEALCHEMISTSTONE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 50000,
			["id"] = 60403,
			["icon"] = 136240,
			["name"] = "Indestructible Alchemist Stone",
		},
		["FELINEGRACE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Feline Grace",
			["icon"] = 132914,
			["id"] = 20719,
		},
		["DARKJADEFOCUSINGLENS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 56205,
			["icon"] = 134071,
			["name"] = "Dark Jade Focusing Lens",
		},
		["FROSTWOVENCOWL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 55907,
			["icon"] = 136249,
			["name"] = "Frostwoven Cowl",
		},
		["CINDERCLOTHBOOTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Cindercloth Boots",
			["icon"] = 132149,
			["id"] = 12088,
		},
		["FLASHINGBLOODSTONE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53844,
			["icon"] = 134071,
			["name"] = "Flashing Bloodstone",
		},
		["EMBRACEOFTHETWISTINGNETHER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 50000,
			["id"] = 36256,
			["icon"] = 136241,
			["name"] = "Embrace of the Twisting Nether",
		},
		["DEATHGATE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 9408,
			["id"] = 50977,
			["icon"] = 135766,
			["name"] = "Death Gate",
		},
		["DEVOTIONAURA"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Devotion Aura",
			["icon"] = 135893,
			["id"] = 465,
		},
		["GLYPHOFDIVINITY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Divinity",
			["icon"] = 136243,
			["id"] = 54939,
		},
		["SHATTERINGTHROW"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 1500,
			["name"] = "Shattering Throw",
			["icon"] = 311430,
			["id"] = 64380,
		},
		["GLYPHOFSPIRITUALATTUNEMENT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Spiritual Attunement",
			["icon"] = 136243,
			["id"] = 54924,
		},
		["SUBTLETY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 17118,
			["icon"] = 132150,
			["name"] = "Subtlety",
		},
		["INKOFTHESKY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Ink of the Sky",
			["icon"] = 132918,
			["id"] = 57712,
		},
		["DIAMONDFOCUSRING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Diamond Focus Ring",
			["icon"] = 136243,
			["id"] = 36526,
		},
		["QUICKNESS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Quickness",
			["icon"] = 136057,
			["id"] = 4281,
		},
		["DAUNTINGHANDGUARDS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55301,
			["icon"] = 136241,
			["name"] = "Daunting Handguards",
		},
		["BRILLIANTSARONITEBOOTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55057,
			["icon"] = 136241,
			["name"] = "Brilliant Saronite Boots",
		},
		["ENGRAVEDTRUESILVERRING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Engraved Truesilver Ring",
			["icon"] = 136243,
			["id"] = 25620,
		},
		["HORNOFWINTER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 57330,
			["icon"] = 134228,
			["name"] = "Horn of Winter",
		},
		["TRUESILVERBREASTPLATE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["id"] = 9974,
			["icon"] = 132739, --136192,
			["name"] = "Truesilver Breastplate",
		},
		["PACKOFENDLESSPOCKETS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 60643,
			["icon"] = 136247,
			["name"] = "Pack of Endless Pockets",
		},
		["WIZARDWEAVEROBE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Wizardweave Robe",
			["icon"] = 132149,
			["id"] = 18446,
		},
		["BLOCK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 107,
			["icon"] = 132110,
			["name"] = "Block",
		},
		["BOXOFBOMBS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 56468,
			["icon"] = 136243,
			["name"] = "Box of Bombs",
		},
		["NETHERCHAINSHIRT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 50000,
			["id"] = 34529,
			["icon"] = 136241,
			["name"] = "Nether Chain Shirt",
		},
		["INLAIDMALACHITERING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Inlaid Malachite Ring",
			["icon"] = 136243,
			["id"] = 25283,
		},
		["ROYALSHADOWCRYSTAL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53864,
			["icon"] = 134071,
			["name"] = "Royal Shadow Crystal",
		},
		["THORIUMHELM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Thorium Helm",
			["icon"] = 133125, --136192,
			["id"] = 16653,
		},
		["GOLDENRINGOFPOWER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["name"] = "Golden Ring of Power",
			["icon"] = 134072,
			["id"] = 34955,
		},
		["SCROLLOFAGILITYIV"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Scroll of Agility IV",
			["icon"] = 132918,
			["id"] = 58478,
		},
		["HARDENEDADAMANTITETUBE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 30307,
			["icon"] = 136243,
			["name"] = "Hardened Adamantite Tube",
		},
		["THICKWARAXE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Thick War Axe",
			["icon"] = 135419, --136192,
			["id"] = 3294,
		},
		["BLACKMAGEWEAVELEGGINGS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Black Mageweave Leggings",
			["icon"] = 132149,
			["id"] = 12049,
		},
		["REINFORCEDLINENCAPE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Reinforced Linen Cape",
			["icon"] = 132149,
			["id"] = 2397,
		},
		["GLYPHOFPOLYMORPH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 56375,
			["icon"] = 136243,
			["name"] = "Glyph of Polymorph",
		},
		["DELICATEBLOODSTONE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53832,
			["icon"] = 134071,
			["name"] = "Delicate Bloodstone",
		},
		["DARKLEATHERCLOAK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Dark Leather Cloak",
			["icon"] = 136247,
			["id"] = 2168,
		},
		["APPRENTICESKINNER"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Apprentice Skinner",
			["icon"] = 134366,
			["id"] = 8615,
		},
		["GLYPHOFIMMOLATIONTRAP"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Immolation Trap",
			["icon"] = 136243,
			["id"] = 56846,
		},
		["FROSTBOLT"] = {
			["maxRange"] = 36,
			["minRange"] = 0,
			["castTime"] = 753,
			["id"] = 116,
			["icon"] = 135846,
			["name"] = "Frostbolt",
		},
		["STORMFORGEDHAUBERK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["id"] = 36130,
			["icon"] = 132639, --136192,
			["name"] = "Stormforged Hauberk",
		},
		["CATFORM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Cat Form",
			["icon"] = 132115,
			["id"] = 768,
		},
		["SARONITEAMBUSHER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55179,
			["icon"] = 136241,
			["name"] = "Saronite Ambusher",
		},
		["PATTERNEDBRONZEBRACERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Patterned Bronze Bracers",
			["icon"] = 132606, --136192,
			["id"] = 2672,
		},
		["MYSTICTOME"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Mystic Tome",
			["icon"] = 132918,
			["id"] = 58565,
		},
		["SPRINGYARACHNOWEAVE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 63765,
			["icon"] = 136243,
			["name"] = "Springy Arachnoweave",
		},
		["SARONITEPROTECTOR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55013,
			["icon"] = 136241,
			["name"] = "Saronite Protector",
		},
		["HEAVYCOPPERRING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Heavy Copper Ring",
			["icon"] = 136243,
			["id"] = 26926,
		},
		["THORIUMLEGGINGS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Thorium Leggings",
			["icon"] = 134584, --136192,
			["id"] = 16662,
		},
		["DARKLEATHERBOOTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Dark Leather Boots",
			["icon"] = 136247,
			["id"] = 2167,
		},
		["BLOODSTRIKE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 45902,
			["icon"] = 135772,
			["name"] = "Blood Strike",
		},
		["COPPERCHAINBOOTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Copper Chain Boots",
			["icon"] = 132535, --136192,
			["id"] = 3319,
		},
		["FROSTWEAVEBANDAGE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 1500,
			["id"] = 45545,
			["icon"] = 133675,
			["name"] = "Frostweave Bandage",
		},
		["DENSEBLASTINGPOWDER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Dense Blasting Powder",
			["icon"] = 136243,
			["id"] = 19788,
		},
		["BRILLIANTSARONITEBREASTPLATE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55058,
			["icon"] = 136241,
			["name"] = "Brilliant Saronite Breastplate",
		},
		["STARFIRE"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 3293,
			["name"] = "Starfire",
			["icon"] = 135753,
			["id"] = 2912,
		},
		["VICTORYRUSH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Victory Rush",
			["icon"] = 132342,
			["id"] = 34428,
		},
		["SMELTGOLD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 1500,
			["name"] = "Smelt Gold",
			["icon"] = 136243,
			["id"] = 3308,
		},
		["DARKFROSTSCALEBREASTPLATE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 60604,
			["icon"] = 136247,
			["name"] = "Dark Frostscale Breastplate",
		},
		["SHIELDBLOCK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Shield Block",
			["icon"] = 132110,
			["id"] = 2565,
		},
		["ARCANEEXPLOSION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 1449,
			["icon"] = 136116,
			["name"] = "Arcane Explosion",
		},
		["SONICBOOSTER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["id"] = 56466,
			["icon"] = 136243,
			["name"] = "Sonic Booster",
		},
		["LUSTROUSCHALCEDONY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53941,
			["icon"] = 134071,
			["name"] = "Lustrous Chalcedony",
		},
		["ELIXIROFFORTITUDE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Elixir of Fortitude",
			["icon"] = 136243,
			["id"] = 3450,
		},
		["BARBARICSHOULDERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Barbaric Shoulders",
			["icon"] = 136247,
			["id"] = 7151,
		},
		["GLYPHOFICYVEINS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Icy Veins",
			["icon"] = 136243,
			["id"] = 56374,
		},
		["RINGOFTWILIGHTSHADOWS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Ring of Twilight Shadows",
			["icon"] = 136243,
			["id"] = 25318,
		},
		["CRIMSONSILKGLOVES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Crimson Silk Gloves",
			["icon"] = 132149,
			["id"] = 8804,
		},
		["THEBIGONE"] = {
			["maxRange"] = 15,
			["minRange"] = 0,
			["castTime"] = 2822,
			["id"] = 12562,
			["icon"] = 135826,
			["name"] = "The Big One",
		},
		["CONSECRATION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Consecration",
			["icon"] = 135926,
			["id"] = 20116,
		},
		["GUARDIANSSHADOWCRYSTAL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53871,
			["icon"] = 134071,
			["name"] = "Guardian's Shadow Crystal",
		},
		["SHRED"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Shred",
			["icon"] = 132152,
			["id"] = 3252,
		},
		["ARTISANMINER"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Artisan Miner",
			["icon"] = 136248,
			["id"] = 10249,
		},
		["TRUESILVERROD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Truesilver Rod",
			["icon"] = 135148, --136192,
			["id"] = 14380,
		},
		["SCROLLOFSTAMINAVI"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Scroll of Stamina VI",
			["icon"] = 132918,
			["id"] = 50618,
		},
		["GLYPHOFWINDFURYWEAPON"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 55445,
			["icon"] = 136243,
			["name"] = "Glyph of Windfury Weapon",
		},
		["INTIMIDATINGSHOUT"] = {
			["maxRange"] = 8,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Intimidating Shout",
			["icon"] = 132154,
			["id"] = 5246,
		},
		["GLYPHOFVOIDWALKER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Voidwalker",
			["icon"] = 136243,
			["id"] = 56247,
		},
		["DRAGONSTOMPERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 60605,
			["icon"] = 136247,
			["name"] = "Dragonstompers",
		},
		["APPRENTICETAILOR"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Apprentice Tailor",
			["icon"] = 136249,
			["id"] = 3911,
		},
		["SWIFTFLIGHTFORM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Swift Flight Form",
			["icon"] = 132128,
			["id"] = 40120,
		},
		["GLYPHOFARCANEEXPLOSION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Arcane Explosion",
			["icon"] = 136243,
			["id"] = 56360,
		},
		["MITHRILSCALEPANTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Mithril Scale Pants",
			["icon"] = 134583, --136192,
			["id"] = 9931,
		},
		["THICKBRONZENECKLACE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Thick Bronze Necklace",
			["icon"] = 136243,
			["id"] = 26927,
		},
		["SPELLWEAVEGLOVES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 56029,
			["icon"] = 136249,
			["name"] = "Spellweave Gloves",
		},
		["GNOMISHPOWERGOGGLES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 50000,
			["id"] = 30574,
			["icon"] = 136243,
			["name"] = "Gnomish Power Goggles",
		},
		["EXPERTALCHEMIST"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Expert Alchemist",
			["icon"] = 136240,
			["id"] = 3465,
		},
		["FURLININGSPELLPOWER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 57691,
			["icon"] = 136247,
			["name"] = "Fur Lining - Spell Power",
		},
		["ELUSIVENESS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Elusiveness",
			["icon"] = 135994,
			["id"] = 13981,
		},
		["MIDNIGHTINK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Midnight Ink",
			["icon"] = 132918,
			["id"] = 53462,
		},
		["FELIRONHAMMER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 29558,
			["icon"] = 133043, --136192,
			["name"] = "Fel Iron Hammer",
		},
		["SOLIDDYNAMITE"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 941,
			["name"] = "Solid Dynamite",
			["icon"] = 135826,
			["id"] = 12419,
		},
		["REMOVECURSE"] = {
			["maxRange"] = 40,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 475,
			["icon"] = 136082,
			["name"] = "Remove Curse",
		},
		["FINELEATHERCLOAK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Fine Leather Cloak",
			["icon"] = 136247,
			["id"] = 2159,
		},
		["RUNEDBLOODSTONE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53834,
			["icon"] = 134071,
			["name"] = "Runed Bloodstone",
		},
		["GLYPHOFRIP"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Rip",
			["icon"] = 136243,
			["id"] = 54818,
		},
		["HANDOFRECKONING"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Hand of Reckoning",
			["icon"] = 135984,
			["id"] = 62124,
		},
		["THORNS"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Thorns",
			["icon"] = 136104,
			["id"] = 467,
		},
		["FROSTGUARDDRAPE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 64729,
			["icon"] = 136249,
			["name"] = "Frostguard Drape",
		},
		["WRATH"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 1411,
			["name"] = "Wrath",
			["icon"] = 136006,
			["id"] = 5176,
		},
		["SCROLLOFSTRENGTHII"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Scroll of Strength II",
			["icon"] = 132918,
			["id"] = 58485,
		},
		["BREASTPLATEOFKINGS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 50000,
			["id"] = 34533,
			["icon"] = 136241,
			["name"] = "Breastplate of Kings",
		},
		["BLACKMAGEWEAVEVEST"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Black Mageweave Vest",
			["icon"] = 132149,
			["id"] = 12048,
		},
		["CURSEOFWEAKNESS"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 702,
			["icon"] = 136138,
			["name"] = "Curse of Weakness",
		},
		["SCROLLOFSPIRITIV"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Scroll of Spirit IV",
			["icon"] = 132918,
			["id"] = 50607,
		},
		["SCROLLOFAGILITY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 1500,
			["name"] = "Scroll of Agility",
			["icon"] = 132918,
			["id"] = 58472,
		},
		["FROSTSAVAGELEGGINGS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 59588,
			["icon"] = 136249,
			["name"] = "Frostsavage Leggings",
		},
		["BLOODSTONEBAND"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 56193,
			["icon"] = 134071,
			["name"] = "Bloodstone Band",
		},
		["DEVASTATE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Devastate",
			["icon"] = 135291,
			["id"] = 20243,
		},
		["PROWL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Prowl",
			["icon"] = 132089,
			["id"] = 5215,
		},
		["EMERALDLIONRING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Emerald Lion Ring",
			["icon"] = 134072,
			["id"] = 34961,
		},
		["FAERIEFIRE"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Faerie Fire",
			["icon"] = 136033,
			["id"] = 770,
		},
		["WEAPONVELLUM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Weapon Vellum",
			["icon"] = 132918,
			["id"] = 52840,
		},
		["SMELTFELIRON"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 1500,
			["id"] = 29356,
			["icon"] = 136243,
			["name"] = "Smelt Fel Iron",
		},
		["BLACKDUSKWEAVEROBE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 55941,
			["icon"] = 136249,
			["name"] = "Black Duskweave Robe",
		},
		["GLYPHOFSCOURGESTRIKE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 57224,
			["icon"] = 132918,
			["name"] = "Glyph of Scourge Strike",
		},
		["COPPERCLAYMORE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Copper Claymore",
			["icon"] = 135322, --136192,
			["id"] = 9983,
		},
		["FIREOIL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 1500,
			["name"] = "Fire Oil",
			["icon"] = 136243,
			["id"] = 7837,
		},
		["ABOLISHPOISON"] = {
			["maxRange"] = 40,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Abolish Poison",
			["icon"] = 136068,
			["id"] = 2893,
		},
		["GLYPHOFFADE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Fade",
			["icon"] = 136243,
			["id"] = 55684,
		},
		["GLYPHOFSUCCUBUS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 56250,
			["icon"] = 136243,
			["name"] = "Glyph of Succubus",
		},
		["ROUGHGRINDINGSTONE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Rough Grinding Stone",
			["icon"] = 135243, --136192,
			["id"] = 3320,
		},
		["SCOURGESTRIKE"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 28265,
			["icon"] = 135990,
			["name"] = "Scourge Strike",
		},
		["SOLIDWEIGHTSTONE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Solid Weightstone",
			["icon"] = 135258, --136192,
			["id"] = 9921,
		},
		["GNOMISHBATTLECHICKEN"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 12906,
			["icon"] = 136243,
			["name"] = "Gnomish Battle Chicken",
		},
		["GLYPHOFIMP"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Imp",
			["icon"] = 136243,
			["id"] = 56248,
		},
		["FROSTWOVENGLOVES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 55904,
			["icon"] = 136249,
			["name"] = "Frostwoven Gloves",
		},
		["POTENTHUGECITRINE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53882,
			["icon"] = 134071,
			["name"] = "Potent Huge Citrine",
		},
		["ORNATESARONITEGAUNTLETS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 56553,
			["icon"] = 136241,
			["name"] = "Ornate Saronite Gauntlets",
		},
		["RUNICMANAPOTION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53837,
			["icon"] = 136240,
			["name"] = "Runic Mana Potion",
		},
		["BLADEDPICKAXE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 56461,
			["icon"] = 136243,
			["name"] = "Bladed Pickaxe",
		},
		["NIGHTSCAPEHEADBAND"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Nightscape Headband",
			["icon"] = 136247,
			["id"] = 10507,
		},
		["GLYPHOFCLEAVING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Glyph of Cleaving",
			["icon"] = 132918,
			["id"] = 57154,
		},
		["CHALLENGINGROAR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Challenging Roar",
			["icon"] = 132117,
			["id"] = 5209,
		},
		["GLYPHOFSOULSTONE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Soulstone",
			["icon"] = 136243,
			["id"] = 56231,
		},
		["REDEMPTION"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 9408,
			["name"] = "Redemption",
			["icon"] = 135955,
			["id"] = 7328,
		},
		["BRONZEBANDOFFORCE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Bronze Band of Force",
			["icon"] = 134072,
			["id"] = 37818,
		},
		["FORCEFULDEFLECTION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 49410,
			["icon"] = 132269,
			["name"] = "Forceful Deflection",
		},
		["FELCLOTHHOOD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Felcloth Hood",
			["icon"] = 132149,
			["id"] = 18442,
		},
		["TURTLESCALEHELM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Turtle Scale Helm",
			["icon"] = 136247,
			["id"] = 10552,
		},
		["HELLFIRETOME"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 59495,
			["icon"] = 132918,
			["name"] = "Hellfire Tome",
		},
		["DENSEDYNAMITE"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 941,
			["name"] = "Dense Dynamite",
			["icon"] = 135826,
			["id"] = 23063,
		},
		["MITHRILFILIGREE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Mithril Filigree",
			["icon"] = 136243,
			["id"] = 25615,
		},
		["ONEHANDEDMACES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "One-Handed Maces",
			["icon"] = 133476,
			["id"] = 198,
		},
		["MASTEROFANATOMY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 53125,
			["icon"] = 134338,
			["name"] = "Master of Anatomy",
		},
		["TELEPORTSHATTRATH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 9408,
			["id"] = 33690,
			["icon"] = 135760,
			["name"] = "Teleport: Shattrath",
		},
		["SCROLLOFSTRENGTHVIII"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 58491,
			["icon"] = 132918,
			["name"] = "Scroll of Strength VIII",
		},
		["INDESTRUCTIBLEPOTION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53905,
			["icon"] = 136240,
			["name"] = "Indestructible Potion",
		},
		["RIGHTEOUSGAUNTLETS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55300,
			["icon"] = 136241,
			["name"] = "Righteous Gauntlets",
		},
		["THORIUMARMOR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Thorium Armor",
			["icon"] = 132743, --136192,
			["id"] = 16642,
		},
		["DEATHCOIL"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 6789,
			["icon"] = 136145,
			["name"] = "Death Coil",
		},
		["HANDFULOFCOBALTBOLTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 56349,
			["icon"] = 136243,
			["name"] = "Handful of Cobalt Bolts",
		},
		["SAVAGEROAR"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Savage Roar",
			["icon"] = 236167,
			["id"] = 52610,
		},
		["HEAVYMITHRILBOOTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Heavy Mithril Boots",
			["icon"] = 132582, --136192,
			["id"] = 9968,
		},
		["CATSEYEELIXIR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Catseye Elixir",
			["icon"] = 136243,
			["id"] = 12609,
		},
		["CRIMSONSILKVEST"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Crimson Silk Vest",
			["icon"] = 132149,
			["id"] = 8791,
		},
		["SCROLLOFSPIRITVIII"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 50611,
			["icon"] = 132918,
			["name"] = "Scroll of Spirit VIII",
		},
		["TWOHANDEDAXES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Two-Handed Axes",
			["icon"] = 132395,
			["id"] = 197,
		},
		["JORMUNGARLEGREINFORCEMENTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3763,
			["id"] = 50903,
			["icon"] = 136247,
			["name"] = "Jormungar Leg Reinforcements",
		},
		["MIGHTYRAGEPOTION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Mighty Rage Potion",
			["icon"] = 136243,
			["id"] = 17527,
		},
		["FELIRONPLATEPANTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 29549,
			["icon"] = 134694, --136192,
			["name"] = "Fel Iron Plate Pants",
		},
		["GLIMMERINGHUGECITRINE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53891,
			["icon"] = 134071,
			["name"] = "Glimmering Huge Citrine",
		},
		["GLYPHOFICEARMOR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Ice Armor",
			["icon"] = 136243,
			["id"] = 56384,
		},
		["BRONZESHORTSWORD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Bronze Shortsword",
			["icon"] = 135274, --136192,
			["id"] = 2742,
		},
		["TRACKBEASTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 1494,
			["icon"] = 132328,
			["name"] = "Track Beasts",
		},
		["TITANSTEELSPELLBLADE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 63182,
			["icon"] = 136241,
			["name"] = "Titansteel Spellblade",
		},
		["BOLTOFNETHERWEAVE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 26745,
			["icon"] = 132149,
			["name"] = "Bolt of Netherweave",
		},
		["COARSESTONESTATUE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Coarse Stone Statue",
			["icon"] = 136243,
			["id"] = 32801,
		},
		["ICEBOUNDFORTITUDE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 48792,
			["icon"] = 237525,
			["name"] = "Icebound Fortitude",
		},
		["COBALTBRACERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55834,
			["icon"] = 136241,
			["name"] = "Cobalt Bracers",
		},
		["SCROLLOFSTRENGTH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 1500,
			["name"] = "Scroll of Strength",
			["icon"] = 132918,
			["id"] = 58484,
		},
		["GLYPHOFEARTHLIVINGWEAPON"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 55439,
			["icon"] = 136243,
			["name"] = "Glyph of Earthliving Weapon",
		},
		["JOURNEYMANALCHEMIST"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Journeyman Alchemist",
			["icon"] = 136240,
			["id"] = 2280,
		},
		["BLACKMAGEWEAVESHOULDERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Black Mageweave Shoulders",
			["icon"] = 132149,
			["id"] = 12074,
		},
		["POWERWORDFORTITUDE"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 1243,
			["icon"] = 135987,
			["name"] = "Power Word: Fortitude",
		},
		["FROSTBRANDWEAPON"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 8033,
			["icon"] = 135847,
			["name"] = "Frostbrand Weapon",
		},
		["TRANSMUTESKYFLAREDIAMOND"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 57425,
			["icon"] = 134085,
			["name"] = "Transmute: Skyflare Diamond",
		},
		["EXPERTMINER"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Expert Miner",
			["icon"] = 136248,
			["id"] = 3568,
		},
		["RAKE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Rake",
			["icon"] = 132122,
			["id"] = 1822,
		},
		["PORTALDARNASSUS"] = {
			["maxRange"] = 10,
			["minRange"] = 0,
			["castTime"] = 9408,
			["id"] = 11419,
			["icon"] = 135741,
			["name"] = "Portal: Darnassus",
		},
		["ARTISANBLACKSMITH"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Artisan Blacksmith",
			["icon"] = 136241,
			["id"] = 9786,
		},
		["DRUMSOFFORGOTTENKINGS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 69386,
			["icon"] = 136247,
			["name"] = "Drums of Forgotten Kings",
		},
		["GLYPHOFMINDFLAY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Mind Flay",
			["icon"] = 136243,
			["id"] = 55687,
		},
		["ORNATETIGERSEYENECKLACE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Ornate Tigerseye Necklace",
			["icon"] = 136243,
			["id"] = 26928,
		},
		["GLYPHOFPSYCHICSCREAM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Psychic Scream",
			["icon"] = 136243,
			["id"] = 55676,
		},
		["SPIKEDCOBALTLEGPLATES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 54947,
			["icon"] = 136241,
			["name"] = "Spiked Cobalt Legplates",
		},
		["THORIUMGRENADE"] = {
			["maxRange"] = 45,
			["minRange"] = 0,
			["castTime"] = 941,
			["name"] = "Thorium Grenade",
			["icon"] = 135826,
			["id"] = 19769,
		},
		["MAGMATOTEM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 8187,
			["icon"] = 135826,
			["name"] = "Magma Totem",
		},
		["LIGHTWEAVEEMBROIDERY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 55640,
			["icon"] = 136243,
			["name"] = "Lightweave Embroidery",
		},
		["ENTANGLINGROOTS"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 1411,
			["name"] = "Entangling Roots",
			["icon"] = 136100,
			["id"] = 339,
		},
		["MITHRILCOIF"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Mithril Coif",
			["icon"] = 133137, --136192,
			["id"] = 9961,
		},
		["BRILLIANTSARONITEHELM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 59441,
			["icon"] = 136241,
			["name"] = "Brilliant Saronite Helm",
		},
		["SCROLLOFSTRENGTHV"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Scroll of Strength V",
			["icon"] = 132918,
			["id"] = 58488,
		},
		["BERSERKERRAGE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Berserker Rage",
			["icon"] = 136009,
			["id"] = 18499,
		},
		["AQUATICFORM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Aquatic Form",
			["icon"] = 132112,
			["id"] = 1066,
		},
		["BROWNLINENROBE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Brown Linen Robe",
			["icon"] = 136243,
			["id"] = 7623,
		},
		["WINDFURYWEAPON"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 8232,
			["icon"] = 136018,
			["name"] = "Windfury Weapon",
		},
		["GNOMISHLIGHTNINGGENERATOR"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 55039,
			["icon"] = 136050,
			["name"] = "Gnomish Lightning Generator",
		},
		["EXPERTSCRIBE"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Expert Scribe",
			["icon"] = 237171,
			["id"] = 45377,
		},
		["GOBLINDRAGONGUN"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 12908,
			["icon"] = 136243,
			["name"] = "Goblin Dragon Gun",
		},
		["FELIRONBOMB"] = {
			["maxRange"] = 15,
			["minRange"] = 0,
			["castTime"] = 941,
			["id"] = 30216,
			["icon"] = 135826,
			["name"] = "Fel Iron Bomb",
		},
		["CLAW"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Claw",
			["icon"] = 132140,
			["id"] = 1082,
		},
		["DESTRUCTION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 28508,
			["icon"] = 134729,
			["name"] = "Destruction",
		},
		["SENTRYTOTEM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 6495,
			["icon"] = 136082,
			["name"] = "Sentry Totem",
		},
		["GOLDENDRAGONRING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Golden Dragon Ring",
			["icon"] = 136243,
			["id"] = 25613,
		},
		["FLETCHERSGLOVES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Fletcher's Gloves",
			["icon"] = 136247,
			["id"] = 9145,
		},
		["GLYPHOFFLAMESHOCK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Flame Shock",
			["icon"] = 136243,
			["id"] = 55447,
		},
		["RITUALOFREFRESHMENT"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 43987,
			["icon"] = 135739,
			["name"] = "Ritual of Refreshment",
		},
		["SIMPLEBLACKDRESS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Simple Black Dress",
			["icon"] = 132149,
			["id"] = 12077,
		},
		["SMELTADAMANTITE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 1500,
			["id"] = 29358,
			["icon"] = 136243,
			["name"] = "Smelt Adamantite",
		},
		["DUALWIELD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 674,
			["icon"] = 132147,
			["name"] = "Dual Wield",
		},
		["FIREBLAST"] = {
			["maxRange"] = 20,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 2136,
			["icon"] = 135807,
			["name"] = "Fire Blast",
		},
		["MANASHIELD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 1463,
			["icon"] = 136153,
			["name"] = "Mana Shield",
		},
		["MASTERENGINEER"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 30351,
			["icon"] = 136243,
			["name"] = "Master Engineer",
		},
		["SUNROCKRING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 56194,
			["icon"] = 134071,
			["name"] = "Sun Rock Ring",
		},
		["PERCEPTION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Perception",
			["icon"] = 136090,
			["id"] = 20600,
		},
		["LEGPLATESOFCONQUEST"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55187,
			["icon"] = 136241,
			["name"] = "Legplates of Conquest",
		},
		["DARKARCTICCHESTPIECE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 51570,
			["icon"] = 136247,
			["name"] = "Dark Arctic Chestpiece",
		},
		["MISTYDARKJADE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53922,
			["icon"] = 134071,
			["name"] = "Misty Dark Jade",
		},
		["NIGHINVULNERABILITYBELT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["id"] = 30570,
			["icon"] = 136243,
			["name"] = "Nigh-Invulnerability Belt",
		},
		["FINELEATHERBELT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Fine Leather Belt",
			["icon"] = 136247,
			["id"] = 3763,
		},
		["WHIRRINGBRONZEGIZMO"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Whirring Bronze Gizmo",
			["icon"] = 136243,
			["id"] = 3942,
		},
		["FROSTRESISTANCE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 4077,
			["icon"] = 135843,
			["name"] = "Frost Resistance",
		},
		["SUPERSAPPERCHARGE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 30486,
			["icon"] = 135826,
			["name"] = "Super Sapper Charge",
		},
		["RETRIBUTIONAURA"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Retribution Aura",
			["icon"] = 135873,
			["id"] = 7294,
		},
		["DEATHPACT"] = {
			["maxRange"] = 50000,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 17471,
			["icon"] = 136146,
			["name"] = "Death Pact",
		},
		["GREATERARCANEELIXIR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Greater Arcane Elixir",
			["icon"] = 134827,
			["id"] = 16889,
		},
		["ICESCALELEGARMOR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 50967,
			["icon"] = 136247,
			["name"] = "Icescale Leg Armor",
		},
		["GUNSPECIALIZATION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 20595,
			["icon"] = 134537,
			["name"] = "Gun Specialization",
		},
		["FIRE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 44146,
			["icon"] = 134424,
			["name"] = "Fire",
		},
		["RUNEOFTHEFALLENCRUSADER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 53344,
			["icon"] = 135957,
			["name"] = "Rune of the Fallen Crusader",
		},
		["MASTERCOOK"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 18816,
			["id"] = 33361,
			["icon"] = 133971,
			["name"] = "Master Cook",
		},
		["SWORDSPECIALIZATION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Sword Specialization",
			["icon"] = 132223,
			["id"] = 4352,
		},
		["REGALSHADOWCRYSTAL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53868,
			["icon"] = 134071,
			["name"] = "Regal Shadow Crystal",
		},
		["RUNECLOTHTUNIC"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Runecloth Tunic",
			["icon"] = 132149,
			["id"] = 18407,
		},
		["SCROLLOFINTELLECTVI"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Scroll of Intellect VI",
			["icon"] = 132918,
			["id"] = 50602,
		},
		["GUARDIANGLOVES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Guardian Gloves",
			["icon"] = 136247,
			["id"] = 7156,
		},
		["GLYPHOFFROSTNOVA"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Frost Nova",
			["icon"] = 136243,
			["id"] = 56376,
		},
		["STONEFORM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 7020,
			["icon"] = 132275,
			["name"] = "Stoneform",
		},
		["ORNATESARONITELEGPLATES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 56554,
			["icon"] = 136241,
			["name"] = "Ornate Saronite Legplates",
		},
		["WHITELINENROBE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "White Linen Robe",
			["icon"] = 136243,
			["id"] = 7624,
		},
		["SHADOWYTAROT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Shadowy Tarot",
			["icon"] = 132918,
			["id"] = 59491,
		},
		["DARKGLOWEMBROIDERY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 55768,
			["icon"] = 136037,
			["name"] = "Darkglow Embroidery",
		},
		["GLYPHOFOVERPOWER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Glyph of Overpower",
			["icon"] = 132918,
			["id"] = 57161,
		},
		["EBONWEAVE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 56002,
			["icon"] = 132149,
			["name"] = "Ebonweave",
		},
		["ORNATESARONITEHAUBERK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 56555,
			["icon"] = 136241,
			["name"] = "Ornate Saronite Hauberk",
		},
		["AURORASLIPPERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 56023,
			["icon"] = 136249,
			["name"] = "Aurora Slippers",
		},
		["HUNTERSMARK"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 1130,
			["icon"] = 132212,
			["name"] = "Hunter's Mark",
		},
		["SPIKEDCOBALTCHESTPIECE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 54944,
			["icon"] = 136241,
			["name"] = "Spiked Cobalt Chestpiece",
		},
		["HEAVYBLASTINGPOWDER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Heavy Blasting Powder",
			["icon"] = 136243,
			["id"] = 3945,
		},
		["SCALEDICEWALKERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 60630,
			["icon"] = 136247,
			["name"] = "Scaled Icewalkers",
		},
		["CHARGER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 1500,
			["name"] = "Charger",
			["icon"] = 132226,
			["id"] = 23214,
		},
		["SENSEUNDEAD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Sense Undead",
			["icon"] = 135974,
			["id"] = 5502,
		},
		["HEAVYBOREANLEATHER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 1500,
			["id"] = 50936,
			["icon"] = 136247,
			["name"] = "Heavy Borean Leather",
		},
		["EXPLOSIVEDECOY"] = {
			["maxRange"] = 20,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 54359,
			["icon"] = 136172,
			["name"] = "Explosive Decoy",
		},
		["WICKEDLEATHERHEADBAND"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Wicked Leather Headband",
			["icon"] = 136243,
			["id"] = 19071,
		},
		["ELIXIROFGREATERWATERBREATHING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Elixir of Greater Water Breathing",
			["icon"] = 136243,
			["id"] = 22808,
		},
		["COBALTTRIANGLESHIELD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 54550,
			["icon"] = 136241,
			["name"] = "Cobalt Triangle Shield",
		},
		["PURIFIEDSHADOWCRYSTAL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53863,
			["icon"] = 134071,
			["name"] = "Purified Shadow Crystal",
		},
		["BRONZEWARHAMMER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Bronze Warhammer",
			["icon"] = 133055, --136192,
			["id"] = 9985,
		},
		["HEROICTHROW"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Heroic Throw",
			["icon"] = 132453,
			["id"] = 57755,
		},
		["ARCANESHOT"] = {
			["maxRange"] = 35,
			["minRange"] = 5,
			["castTime"] = -999500,
			["id"] = 3044,
			["icon"] = 132218,
			["name"] = "Arcane Shot",
		},
		["JOURNEYMANENGINEER"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Journeyman Engineer",
			["icon"] = 136243,
			["id"] = 4040,
		},
		["HEALINGPOTION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Healing Potion",
			["icon"] = 136243,
			["id"] = 439,
		},
		["HEAVYWEIGHTSTONE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Heavy Weightstone",
			["icon"] = 135257, --136192,
			["id"] = 3117,
		},
		["GREATERBLESSINGOFSANCTUARY"] = {
			["maxRange"] = 40,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Greater Blessing of Sanctuary",
			["icon"] = 135911,
			["id"] = 25899,
		},
		["DENSESTONESTATUE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Dense Stone Statue",
			["icon"] = 134230,
			["id"] = 32805,
		},
		["TELEPORTTHERAMORE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 9408,
			["id"] = 49359,
			["icon"] = 135764,
			["name"] = "Teleport: Theramore",
		},
		["COPPERTUBE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Copper Tube",
			["icon"] = 136243,
			["id"] = 3924,
		},
		["DUSKYBRACERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Dusky Bracers",
			["icon"] = 136247,
			["id"] = 9201,
		},
		["COARSEWEIGHTSTONE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Coarse Weightstone",
			["icon"] = 135256, --136192,
			["id"] = 3116,
		},
		["GLYPHOFRAPIDCHARGE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Glyph of Rapid Charge",
			["icon"] = 132918,
			["id"] = 57162,
		},
		["BOREANARMORKIT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2822,
			["id"] = 50906,
			["icon"] = 136247,
			["name"] = "Borean Armor Kit",
		},
		["MOONGLOWINK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Moonglow Ink",
			["icon"] = 132918,
			["id"] = 52843,
		},
		["SWORDGUARDEMBROIDERY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 55775,
			["icon"] = 236282,
			["name"] = "Swordguard Embroidery",
		},
		["MASTERBLACKSMITH"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 29845,
			["icon"] = 136241,
			["name"] = "Master Blacksmith",
		},
		["RUNECLOTHCLOAK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Runecloth Cloak",
			["icon"] = 132149,
			["id"] = 18409,
		},
		["SOLIDAZUREMOONSTONE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 28950,
			["icon"] = 136243,
			["name"] = "Solid Azure Moonstone",
		},
		["ENRAGEDREGENERATION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Enraged Regeneration",
			["icon"] = 132345,
			["id"] = 55694,
		},
		["ARTISANLEATHERWORKER"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Artisan Leatherworker",
			["icon"] = 133611,
			["id"] = 10663,
		},
		["ORNATESARONITEWALKERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 56552,
			["icon"] = 136241,
			["name"] = "Ornate Saronite Walkers",
		},
		["RHINODOGS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 45553,
			["icon"] = 133971,
			["name"] = "Rhino Dogs",
		},
		["CROWNOFTHESEAWITCH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["id"] = 41418,
			["icon"] = 134071,
			["name"] = "Crown of the Sea Witch",
		},
		["CITRINERINGOFRAPIDHEALING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Citrine Ring of Rapid Healing",
			["icon"] = 136243,
			["id"] = 25621,
		},
		["ICEBORNESHOULDERPADS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 50940,
			["icon"] = 136247,
			["name"] = "Iceborne Shoulderpads",
		},
		["JOURNEYMANFIRSTAID"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Journeyman First Aid",
			["icon"] = 135966,
			["id"] = 3280,
		},
		["TITANSTEELSHANKER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 56234,
			["icon"] = 136241,
			["name"] = "Titansteel Shanker",
		},
		["SPIDERSAUSAGE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Spider Sausage",
			["icon"] = 134022, --136192,
			["id"] = 21175,
		},
		["GLYPHOFHEALINGSTREAMTOTEM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Healing Stream Totem",
			["icon"] = 136243,
			["id"] = 55456,
		},
		["GLYPHOFWHIRLWIND"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 57172,
			["icon"] = 132918,
			["name"] = "Glyph of Whirlwind",
		},
		["HOLYWRATH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Holy Wrath",
			["icon"] = 135902,
			["id"] = 2812,
		},
		["SPIKEDCOBALTBRACERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 54948,
			["icon"] = 136241,
			["name"] = "Spiked Cobalt Bracers",
		},
		["GLYPHOFFROSTSTRIKE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Glyph of Frost Strike",
			["icon"] = 132918,
			["id"] = 57216,
		},
		["GOBLINBEAMWELDER"] = {
			["maxRange"] = 20,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 67324,
			["icon"] = 136028,
			["name"] = "Goblin Beam Welder",
		},
		["SINISTERSTRIKE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 1752,
			["icon"] = 136189,
			["name"] = "Sinister Strike",
		},
		["BAKEDMANTARAY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 45569,
			["icon"] = 133971,
			["name"] = "Baked Manta Ray",
		},
		["GLYPHOFINSECTSWARM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Insect Swarm",
			["icon"] = 136243,
			["id"] = 54830,
		},
		["SPICEBREAD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Spice Bread",
			["icon"] = 134051, --136192,
			["id"] = 37836,
		},
		["MAGEWEAVEBANDAGE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 1500,
			["name"] = "Mageweave Bandage",
			["icon"] = 133689, --136192,
			["id"] = 10840,
		},
		["ENGINEERINGSPECIALIZATION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 20593,
			["icon"] = 134063,
			["name"] = "Engineering Specialization",
		},
		["DURABLENERUBHIDECAPE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 60640,
			["icon"] = 136247,
			["name"] = "Durable Nerubhide Cape",
		},
		["FEROCIOUSBITE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Ferocious Bite",
			["icon"] = 132127,
			["id"] = 22568,
		},
		["DREAMWEAVEGLOVES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["name"] = "Dreamweave Gloves",
			["icon"] = 132149,
			["id"] = 12067,
		},
		["SAVAGEDEFENSE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Savage Defense",
			["icon"] = 132278,
			["id"] = 62600,
		},
		["FLYINGMACHINE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 1500,
			["id"] = 44153,
			["icon"] = 132240,
			["name"] = "Flying Machine",
		},
		["ICYMANAPOTION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53839,
			["icon"] = 136240,
			["name"] = "Icy Mana Potion",
		},
		["BUGSQUASHERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 60620,
			["icon"] = 136247,
			["name"] = "Bugsquashers",
		},
		["GLYPHOFMENDING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Mending",
			["icon"] = 136243,
			["id"] = 56833,
		},
		["BRONZEAXE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Bronze Axe",
			["icon"] = 132408, --136192,
			["id"] = 2741,
		},
		["GREENIRONLEGGINGS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Green Iron Leggings",
			["icon"] = 134585, --136192,
			["id"] = 3506,
		},
		["HANDOFPROTECTION"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Hand of Protection",
			["icon"] = 135964,
			["id"] = 1022,
		},
		["MARKSBOOMSTICK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["id"] = 54353,
			["icon"] = 136243,
			["name"] = "Mark \"S\" Boomstick",
		},
		["REND"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Rend",
			["icon"] = 132155,
			["id"] = 772,
		},
		["SPELLWEAVEROBE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 56028,
			["icon"] = 136249,
			["name"] = "Spellweave Robe",
		},
		["BOLTOFRUNECLOTH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Bolt of Runecloth",
			["icon"] = 132149,
			["id"] = 18401,
		},
		["ARTISANTAILOR"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Artisan Tailor",
			["icon"] = 136249,
			["id"] = 12181,
		},
		["WINDFURYTOTEM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 8512,
			["icon"] = 136114,
			["name"] = "Windfury Totem",
		},
		["BLACKDUSKWEAVELEGGINGS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 55925,
			["icon"] = 136249,
			["name"] = "Black Duskweave Leggings",
		},
		["FELIRONCHAINGLOVES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 29552,
			["icon"] = 132945, --136192,
			["name"] = "Fel Iron Chain Gloves",
		},
		["ICEBORNEHELM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 60608,
			["icon"] = 136247,
			["name"] = "Iceborne Helm",
		},
		["GLYPHOFSTARFIRE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Starfire",
			["icon"] = 136243,
			["id"] = 54845,
		},
		["BRONZEMACE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Bronze Mace",
			["icon"] = 133483, --136192,
			["id"] = 2740,
		},
		["CRYSTALCHALCEDONYAMULET"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 58142,
			["icon"] = 134071,
			["name"] = "Crystal Chalcedony Amulet",
		},
		["STONECLAWTOTEM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 5730,
			["icon"] = 136097,
			["name"] = "Stoneclaw Totem",
		},
		["COBALTCHESTPIECE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 52570,
			["icon"] = 136241,
			["name"] = "Cobalt Chestpiece",
		},
		["APPRENTICEJEWELCRAFTER"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Apprentice Jewelcrafter",
			["icon"] = 134071,
			["id"] = 25245,
		},
		["PORTALTHERAMORE"] = {
			["maxRange"] = 10,
			["minRange"] = 0,
			["castTime"] = 9408,
			["id"] = 49360,
			["icon"] = 135749,
			["name"] = "Portal: Theramore",
		},
		["DENSEWEIGHTSTONE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Dense Weightstone",
			["icon"] = 135259, --136192,
			["id"] = 16640,
		},
		["MASTERFIRSTAID"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 18816,
			["id"] = 27029,
			["icon"] = 135966,
			["name"] = "Master First Aid",
		},
		["BLAZEGUARD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 50000,
			["id"] = 34537,
			["icon"] = 136241,
			["name"] = "Blazeguard",
		},
		["SILVEREDBRONZEBOOTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Silvered Bronze Boots",
			["icon"] = 132535, --136192,
			["id"] = 3331,
		},
		["HEAVYGRINDINGSTONE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Heavy Grinding Stone",
			["icon"] = 135245, --136192,
			["id"] = 3337,
		},
		["HATOFWINTRYDOOM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 56018,
			["icon"] = 136249,
			["name"] = "Hat of Wintry Doom",
		},
		["ELIXIROFMIGHTYTHOUGHTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 60367,
			["icon"] = 136240,
			["name"] = "Elixir of Mighty Thoughts",
		},
		["APPRENTICELEATHERWORKER"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Apprentice Leatherworker",
			["icon"] = 133611,
			["id"] = 2155,
		},
		["COBALTBELT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 52568,
			["icon"] = 136241,
			["name"] = "Cobalt Belt",
		},
		["UNHOLYPRESENCE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 48265,
			["icon"] = 135775,
			["name"] = "Unholy Presence",
		},
		["FROSTSCALEGLOVES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 50953,
			["icon"] = 136247,
			["name"] = "Frostscale Gloves",
		},
		["NERUBIANBRACERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 60622,
			["icon"] = 136247,
			["name"] = "Nerubian Bracers",
		},
		["AZUREMOONSTONERING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["id"] = 31050,
			["icon"] = 136243,
			["name"] = "Azure Moonstone Ring",
		},
		["GLYPHOFFEAR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Fear",
			["icon"] = 136243,
			["id"] = 56244,
		},
		["TITANIUMWEAPONCHAIN"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2822,
			["id"] = 55836,
			["icon"] = 132507,
			["name"] = "Titanium Weapon Chain",
		},
		["IRONSTRUT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Iron Strut",
			["icon"] = 136243,
			["id"] = 3958,
		},
		["EARTHELEMENTALTOTEM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 2062,
			["icon"] = 136024,
			["name"] = "Earth Elemental Totem",
		},
		["MITHRILTUBE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Mithril Tube",
			["icon"] = 136243,
			["id"] = 12589,
		},
		["GLYPHOFBONESHIELD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Glyph of Bone Shield",
			["icon"] = 132918,
			["id"] = 57210,
		},
		["TITANSTEELSHIELDWALL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 56400,
			["icon"] = 136241,
			["name"] = "Titansteel Shield Wall",
		},
		["FINELEATHERTUNIC"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Fine Leather Tunic",
			["icon"] = 136247,
			["id"] = 3761,
		},
		["FIRENOVA"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 1535,
			["icon"] = 135824,
			["name"] = "Fire Nova",
		},
		["APPRENTICEALCHEMIST"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Apprentice Alchemist",
			["icon"] = 136240,
			["id"] = 2275,
		},
		["LESSERFLASKOFRESISTANCE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 62213,
			["icon"] = 136240,
			["name"] = "Lesser Flask of Resistance",
		},
		["ARCANEELIXIR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Arcane Elixir",
			["icon"] = 134810,
			["id"] = 11390,
		},
		["TWILIGHTSERPENT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 56184,
			["icon"] = 237242,
			["name"] = "Twilight Serpent",
		},
		["GLYPHOFWATERMASTERY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Water Mastery",
			["icon"] = 136243,
			["id"] = 55436,
		},
		["DEEPTHUNDER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 50000,
			["id"] = 34548,
			["icon"] = 136241,
			["name"] = "Deep Thunder",
		},
		["FISHING"] = {
			["maxRange"] = 20,
			["minRange"] = 10,
			["castTime"] = 0,
			["id"] = 7620,
			["icon"] = 136245,
			["name"] = "Fishing",
		},
		["INFUSEDSHADOWCRYSTAL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53867,
			["icon"] = 134071,
			["name"] = "Infused Shadow Crystal",
		},
		["SHIELDBASH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Shield Bash",
			["icon"] = 132357,
			["id"] = 72,
		},
		["GRANDMASTERTAILOR"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 51308,
			["icon"] = 136249,
			["name"] = "Grand Master Tailor",
		},
		["SMELTTITANSTEEL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 1500,
			["id"] = 55208,
			["icon"] = 135811,
			["name"] = "Smelt Titansteel",
		},
		["SCROLLOFRECALLII"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Scroll of Recall II",
			["icon"] = 132918,
			["id"] = 60336,
		},
		["HANDOFSACRIFICE"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Hand of Sacrifice",
			["icon"] = 135966,
			["id"] = 6940,
		},
		["FELCLOTHROBE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Felcloth Robe",
			["icon"] = 132149,
			["id"] = 18451,
		},
		["SHADOWMIGHTRING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 58146,
			["icon"] = 134071,
			["name"] = "Shadowmight Ring",
		},
		["JADERINGOFSLAYING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 58144,
			["icon"] = 134071,
			["name"] = "Jade Ring of Slaying",
		},
		["DARKICEBORNECHESTGUARD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 60613,
			["icon"] = 136247,
			["name"] = "Dark Iceborne Chestguard",
		},
		["GLACIALROBE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 60993,
			["icon"] = 136249,
			["name"] = "Glacial Robe",
		},
		["CORRODEDSARONITEEDGE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55183,
			["icon"] = 136241,
			["name"] = "Corroded Saronite Edge",
		},
		["RUNEOFRAZORICE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 53343,
			["icon"] = 135842,
			["name"] = "Rune of Razorice",
		},
		["TWOHANDEDMACES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Two-Handed Maces",
			["icon"] = 133479,
			["id"] = 199,
		},
		["CONCENTRATIONAURA"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Concentration Aura",
			["icon"] = 135933,
			["id"] = 19746,
		},
		["GLYPHOFSLICEANDDICE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Slice and Dice",
			["icon"] = 136243,
			["id"] = 56810,
		},
		["BRONZEGREATSWORD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Bronze Greatsword",
			["icon"] = 135321, --136192,
			["id"] = 9986,
		},
		["THORIUMTUBE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Thorium Tube",
			["icon"] = 136243,
			["id"] = 19795,
		},
		["COBALTSHOULDERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 52572,
			["icon"] = 136241,
			["name"] = "Cobalt Shoulders",
		},
		["AQUAMARINESIGNET"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Aquamarine Signet",
			["icon"] = 136243,
			["id"] = 26874,
		},
		["APPRENTICEENGINEER"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Apprentice Engineer",
			["icon"] = 136243,
			["id"] = 4039,
		},
		["THORIUMSETTING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Thorium Setting",
			["icon"] = 136243,
			["id"] = 26880,
		},
		["NETHERWEAVEBAG"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["id"] = 26746,
			["icon"] = 132149,
			["name"] = "Netherweave Bag",
		},
		["TELEPORTSTORMWIND"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 9408,
			["id"] = 3561,
			["icon"] = 135763,
			["name"] = "Teleport: Stormwind",
		},
		["GLYPHOFREJUVENATION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Rejuvenation",
			["icon"] = 136243,
			["id"] = 54754,
		},
		["THORIUMBOOTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Thorium Boots",
			["icon"] = 132589, --136192,
			["id"] = 16652,
		},
		["WICKEDLEATHERBRACERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Wicked Leather Bracers",
			["icon"] = 136243,
			["id"] = 19052,
		},
		["THICKLEATHERAMMOPOUCH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Thick Leather Ammo Pouch",
			["icon"] = 136247,
			["id"] = 14932,
		},
		["SUNSCOPE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2822,
			["id"] = 55076,
			["icon"] = 136243,
			["name"] = "Sun Scope",
		},
		["FROSTPRESENCE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 48263,
			["icon"] = 135773,
			["name"] = "Frost Presence",
		},
		["BRIGHTBLOODSTONE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53835,
			["icon"] = 134071,
			["name"] = "Bright Bloodstone",
		},
		["DEATHSTRIKE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 45463,
			["icon"] = 237517,
			["name"] = "Death Strike",
		},
		["ARMYOFTHEDEAD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 42650,
			["icon"] = 237511,
			["name"] = "Army of the Dead",
		},
		["INNERVATE"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Innervate",
			["icon"] = 136048,
			["id"] = 29166,
		},
		["GLYPHOFSAP"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Sap",
			["icon"] = 136243,
			["id"] = 56798,
		},
		["GRAYWOOLENSHIRT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Gray Woolen Shirt",
			["icon"] = 132149,
			["id"] = 2406,
		},
		["REVENGE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Revenge",
			["icon"] = 132353,
			["id"] = 6572,
		},
		["JAGGEDDEEPPERIDOT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 28917,
			["icon"] = 136243,
			["name"] = "Jagged Deep Peridot",
		},
		["AZURESPELLTHREAD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3763,
			["id"] = 55632,
			["icon"] = 136243,
			["name"] = "Azure Spellthread",
		},
		["DISARM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Disarm",
			["icon"] = 132343,
			["id"] = 676,
		},
		["SOLIDCHALCEDONY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53934,
			["icon"] = 134071,
			["name"] = "Solid Chalcedony",
		},
		["THORIUMRIFLE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["name"] = "Thorium Rifle",
			["icon"] = 136243,
			["id"] = 19792,
		},
		["SEARINGTOTEM"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 2075,
			["icon"] = 135825,
			["name"] = "Searing Totem",
		},
		["CRAFTEDSOLIDSHOT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Crafted Solid Shot",
			["icon"] = 136243,
			["id"] = 3947,
		},
		["SOCKETGLOVES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 55641,
			["icon"] = 132984,
			["name"] = "Socket Gloves",
		},
		["NIGHTSCAPEBOOTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Nightscape Boots",
			["icon"] = 136247,
			["id"] = 10558,
		},
		["LIGHTEARTHFORGEDBLADE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["id"] = 36125,
			["icon"] = 135411, --136192,
			["name"] = "Light Earthforged Blade",
		},
		["TITANSTEELDESTROYER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55369,
			["icon"] = 136241,
			["name"] = "Titansteel Destroyer",
		},
		["NIGHTSHOCKHOOD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 60655,
			["icon"] = 136247,
			["name"] = "Nightshock Hood",
		},
		["ICEBORNEGLOVES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 50941,
			["icon"] = 136247,
			["name"] = "Iceborne Gloves",
		},
		["CONJUREFOOD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2822,
			["id"] = 587,
			["icon"] = 133952,
			["name"] = "Conjure Food",
		},
		["SIMPLELINENBOOTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Simple Linen Boots",
			["icon"] = 132149,
			["id"] = 12045,
		},
		["FROSTSAVAGEROBE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 59587,
			["icon"] = 136249,
			["name"] = "Frostsavage Robe",
		},
		["TRUESILVERSKELETONKEY"] = {
			["maxRange"] = 5,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Truesilver Skeleton Key",
			["icon"] = 136243,
			["id"] = 19651,
		},
		["GLYPHOFSTORMSTRIKE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 55446,
			["icon"] = 136243,
			["name"] = "Glyph of Stormstrike",
		},
		["STYLISHREDSHIRT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Stylish Red Shirt",
			["icon"] = 132149,
			["id"] = 3866,
		},
		["REGROWTH"] = {
			["maxRange"] = 40,
			["minRange"] = 0,
			["castTime"] = 1882,
			["name"] = "Regrowth",
			["icon"] = 136085,
			["id"] = 8936,
		},
		["SAVAGESARONITESKULLSHIELD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55312,
			["icon"] = 136241,
			["name"] = "Savage Saronite Skullshield",
		},
		["MITHRILFRAGBOMB"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 941,
			["name"] = "Mithril Frag Bomb",
			["icon"] = 135826,
			["id"] = 12421,
		},
		["SEAFOAMGAUNTLETS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 60665,
			["icon"] = 136247,
			["name"] = "Seafoam Gauntlets",
		},
		["MIGHTYALCHEMISTSTONE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 50000,
			["id"] = 60405,
			["icon"] = 136240,
			["name"] = "Mighty Alchemist Stone",
		},
		["GOBLINROCKETLAUNCHER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 30563,
			["icon"] = 136243,
			["name"] = "Goblin Rocket Launcher",
		},
		["GLYPHOFSHADOWBURN"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Shadowburn",
			["icon"] = 136243,
			["id"] = 56229,
		},
		["SPIKEDTITANSTEELHELM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55372,
			["icon"] = 136241,
			["name"] = "Spiked Titansteel Helm",
		},
		["CELESTIALINK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Celestial Ink",
			["icon"] = 132918,
			["id"] = 57709,
		},
		["RUNECLOTHPANTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Runecloth Pants",
			["icon"] = 132149,
			["id"] = 18438,
		},
		["ICEBORNEWRISTGUARDS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 60607,
			["icon"] = 136247,
			["name"] = "Iceborne Wristguards",
		},
		["HOWLINGBLAST"] = {
			["maxRange"] = 20,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 49184,
			["icon"] = 135833,
			["name"] = "Howling Blast",
		},
		["GLYPHOFHOLYLIGHT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Holy Light",
			["icon"] = 136243,
			["id"] = 54937,
		},
		["FISTWEAPONS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Fist Weapons",
			["icon"] = 132938,
			["id"] = 15590,
		},
		["GLYPHOFSPRINT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Sprint",
			["icon"] = 136243,
			["id"] = 56811,
		},
		["HANDSTITCHEDLEATHERPANTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Handstitched Leather Pants",
			["icon"] = 136247,
			["id"] = 2153,
		},
		["ICEBANECHESTGUARD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 61008,
			["icon"] = 136241,
			["name"] = "Icebane Chestguard",
		},
		["RUNEOFSWORDSHATTERING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 53323,
			["icon"] = 132269,
			["name"] = "Rune of Swordshattering",
		},
		["DAZZLINGFORESTEMERALD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 54007,
			["icon"] = 134071,
			["name"] = "Dazzling Forest Emerald",
		},
		["RUNICLEATHERPANTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Runic Leather Pants",
			["icon"] = 136243,
			["id"] = 19091,
		},
		["FROSTWOVENSHOULDERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 55902,
			["icon"] = 136249,
			["name"] = "Frostwoven Shoulders",
		},
		["TRUESILVERHEALINGRING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Truesilver Healing Ring",
			["icon"] = 136243,
			["id"] = 26885,
		},
		["RUNEDCOPPERBRACERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Runed Copper Bracers",
			["icon"] = 132602, --136192,
			["id"] = 2664,
		},
		["BOLTOFWOOLENCLOTH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Bolt of Woolen Cloth",
			["icon"] = 136249,
			["id"] = 2964,
		},
		["FINDMINERALS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 2580,
			["icon"] = 136025,
			["name"] = "Find Minerals",
		},
		["GLYPHOFGOUGE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Gouge",
			["icon"] = 136243,
			["id"] = 56809,
		},
		["RADIANTDEEPPERIDOT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 28916,
			["icon"] = 136243,
			["name"] = "Radiant Deep Peridot",
		},
		["SMELTMITHRIL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 1500,
			["name"] = "Smelt Mithril",
			["icon"] = 136243,
			["id"] = 10097,
		},
		["BROWNLINENPANTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Brown Linen Pants",
			["icon"] = 136243,
			["id"] = 3914,
		},
		["GRANDMASTERSCRIBE"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 45380,
			["icon"] = 237171,
			["name"] = "Grand Master Scribe",
		},
		["ADAMANTITEGRENADE"] = {
			["maxRange"] = 45,
			["minRange"] = 0,
			["castTime"] = 941,
			["id"] = 30217,
			["icon"] = 135826,
			["name"] = "Adamantite Grenade",
		},
		["WICKEDLEATHERBELT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Wicked Leather Belt",
			["icon"] = 136243,
			["id"] = 19092,
		},
		["GEMCUTTING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 28875,
			["icon"] = 135998,
			["name"] = "Gemcutting",
		},
		["CLOAKOFCRIMSONSNOW"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 64730,
			["icon"] = 136249,
			["name"] = "Cloak of Crimson Snow",
		},
		["STURDYCOBALTQUICKBLADE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55200,
			["icon"] = 136241,
			["name"] = "Sturdy Cobalt Quickblade",
		},
		["WOOLENBAG"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Woolen Bag",
			["icon"] = 136249,
			["id"] = 3757,
		},
		["TEMPEREDTITANSTEELTREADS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55376,
			["icon"] = 136241,
			["name"] = "Tempered Titansteel Treads",
		},
		["LUNARCRESCENT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 50000,
			["id"] = 34543,
			["icon"] = 136241,
			["name"] = "Lunar Crescent",
		},
		["ELIXIROFFIREPOWER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Elixir of Firepower",
			["icon"] = 136243,
			["id"] = 7845,
		},
		["NECKLACEOFTHEDEEP"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["id"] = 40514,
			["icon"] = 134071,
			["name"] = "Necklace of the Deep",
		},
		["NETHERWEAVEBELT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 26765,
			["icon"] = 132149,
			["name"] = "Netherweave Belt",
		},
		["NOISEMACHINE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["id"] = 56467,
			["icon"] = 136243,
			["name"] = "Noise Machine",
		},
		["GLYPHOFMAUL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Maul",
			["icon"] = 136243,
			["id"] = 54811,
		},
		["BLACKMAGEWEAVEGLOVES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Black Mageweave Gloves",
			["icon"] = 132149,
			["id"] = 12053,
		},
		["BOILEDCLAMS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Boiled Clams",
			["icon"] = 134433,
			["id"] = 6499,
		},
		["GRANDMASTERFISHING"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 51293,
			["icon"] = 136245,
			["name"] = "Grand Master Fishing",
		},
		["QUICKSUNCRYSTAL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53856,
			["icon"] = 134071,
			["name"] = "Quick Sun Crystal",
		},
		["GUARDIANPANTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Guardian Pants",
			["icon"] = 136247,
			["id"] = 7147,
		},
		["EMBOSSEDLEATHERCLOAK"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Embossed Leather Cloak",
			["icon"] = 136247,
			["id"] = 2162,
		},
		["GOBLINDEVILEDCLAMS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Goblin Deviled Clams",
			["icon"] = 134433,
			["id"] = 6500,
		},
		["LIGHTLEATHERPANTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Light Leather Pants",
			["icon"] = 136247,
			["id"] = 9068,
		},
		["ICEBORNECHESTGUARD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 50938,
			["icon"] = 136247,
			["name"] = "Iceborne Chestguard",
		},
		["BIGIRONBOMB"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Big Iron Bomb",
			["icon"] = 136243,
			["id"] = 3967,
		},
		["EXPERTFIRSTAID"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 18816,
			["name"] = "Expert First Aid",
			["icon"] = 135966,
			["id"] = 19903,
		},
		["BLAZEFURY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 50000,
			["id"] = 36258,
			["icon"] = 136241,
			["name"] = "Blazefury",
		},
		["COPPERMACE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Copper Mace",
			["icon"] = 133476, --136192,
			["id"] = 2737,
		},
		["BLACKMAGEWEAVEBOOTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Black Mageweave Boots",
			["icon"] = 132149,
			["id"] = 12073,
		},
		["REVIVE"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Revive",
			["icon"] = 135955,
			["id"] = 24341,
		},
		["COPPERDAGGER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Copper Dagger",
			["icon"] = 135650, --136192,
			["id"] = 8880,
		},
		["WOOLBANDAGE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 1500,
			["name"] = "Wool Bandage",
			["icon"] = 133684, --136192,
			["id"] = 3277,
		},
		["BRILLIANTSARONITEGAUNTLETS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55056,
			["icon"] = 136241,
			["name"] = "Brilliant Saronite Gauntlets",
		},
		["EARTHBINDTOTEM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 2484,
			["icon"] = 136102,
			["name"] = "Earthbind Totem",
		},
		["DENSEGRINDINGSTONE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Dense Grinding Stone",
			["icon"] = 135247, --136192,
			["id"] = 16639,
		},
		["THORIUMBRACERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Thorium Bracers",
			["icon"] = 132612, --136192,
			["id"] = 16644,
		},
		["ARTISANJEWELCRAFTER"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Artisan Jewelcrafter",
			["icon"] = 134073,
			["id"] = 28899,
		},
		["CRIMSONSILKBELT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Crimson Silk Belt",
			["icon"] = 132149,
			["id"] = 8772,
		},
		["SCROLLOFINTELLECTIV"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Scroll of Intellect IV",
			["icon"] = 132918,
			["id"] = 50600,
		},
		["SHIELDOFRIGHTEOUSNESS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Shield of Righteousness",
			["icon"] = 236265,
			["id"] = 53600,
		},
		["GLYPHOFHEALTHSTONE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Healthstone",
			["icon"] = 136243,
			["id"] = 56224,
		},
		["REDRINGOFDESTRUCTION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Red Ring of Destruction",
			["icon"] = 136243,
			["id"] = 36525,
		},
		["CLOAKOFHARSHWINDS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 60631,
			["icon"] = 136247,
			["name"] = "Cloak of Harsh Winds",
		},
		["TELEPORTDARNASSUS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 9408,
			["id"] = 3565,
			["icon"] = 135755,
			["name"] = "Teleport: Darnassus",
		},
		["PURIFIEDJAGGALPEARL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 41420,
			["icon"] = 134071,
			["name"] = "Purified Jaggal Pearl",
		},
		["GLOWINGTHORIUMBAND"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Glowing Thorium Band",
			["icon"] = 134072,
			["id"] = 34960,
		},
		["SOVEREIGNSHADOWDRAENITE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 28936,
			["icon"] = 136243,
			["name"] = "Sovereign Shadow Draenite",
		},
		["ARCANEINTELLECT"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 1459,
			["icon"] = 135932,
			["name"] = "Arcane Intellect",
		},
		["GLYPHOFDARKDEATH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Dark Death",
			["icon"] = 136243,
			["id"] = 63333,
		},
		["HEAVYQUIVER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Heavy Quiver",
			["icon"] = 136247,
			["id"] = 9193,
		},
		["ARCTICWRISTGUARDS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 51571,
			["icon"] = 136247,
			["name"] = "Arctic Wristguards",
		},
		["FLAMESHOCK"] = {
			["maxRange"] = 25,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 8050,
			["icon"] = 135813,
			["name"] = "Flame Shock",
		},
		["STRANGETAROT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Strange Tarot",
			["icon"] = 132918,
			["id"] = 59480,
		},
		["GOLDENSCALEBRACERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Golden Scale Bracers",
			["icon"] = 132609, --136192,
			["id"] = 7223,
		},
		["SPELLSTEAL"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 30449,
			["icon"] = 135729,
			["name"] = "Spellsteal",
		},
		["DESTRUCTIONHOLOGOGS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 50000,
			["id"] = 41320,
			["icon"] = 136243,
			["name"] = "Destruction Holo-gogs",
		},
		["FROSTSCALEBOOTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 50954,
			["icon"] = 136247,
			["name"] = "Frostscale Boots",
		},
		["OBLITERATE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 49020,
			["icon"] = 135771,
			["name"] = "Obliterate",
		},
		["WRATHELIXIR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 53746,
			["icon"] = 134737,
			["name"] = "Wrath Elixir",
		},
		["GREENWOOLENVEST"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Green Woolen Vest",
			["icon"] = 136249,
			["id"] = 2399,
		},
		["RUNEOFSPELLSHATTERING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 53342,
			["icon"] = 136120,
			["name"] = "Rune of Spellshattering",
		},
		["POWERWORDSHIELD"] = {
			["maxRange"] = 40,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 17,
			["icon"] = 135940,
			["name"] = "Power Word: Shield",
		},
		["AQUADYNAMICFISHATTRACTOR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["name"] = "Aquadynamic Fish Attractor",
			["icon"] = 134335,
			["id"] = 8089,
		},
		["MOONSOULCROWN"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Moonsoul Crown",
			["icon"] = 136243,
			["id"] = 25321,
		},
		["SPIKEDCOBALTGAUNTLETS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 54945,
			["icon"] = 136241,
			["name"] = "Spiked Cobalt Gauntlets",
		},
		["RUNICLEATHERBRACERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Runic Leather Bracers",
			["icon"] = 136243,
			["id"] = 19065,
		},
		["SCROLLOFAGILITYVIII"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 58483,
			["icon"] = 132918,
			["name"] = "Scroll of Agility VIII",
		},
		["SMELTBRONZE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 1500,
			["name"] = "Smelt Bronze",
			["icon"] = 136243,
			["id"] = 2659,
		},
		["HEAVYWOOLENGLOVES"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Heavy Woolen Gloves",
			["icon"] = 132149,
			["id"] = 3843,
		},
		["ARCTICBELT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 50949,
			["icon"] = 136247,
			["name"] = "Arctic Belt",
		},
		["SMITE"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 1411,
			["id"] = 585,
			["icon"] = 135924,
			["name"] = "Smite",
		},
		["GLYPHOFSHADOWBOLT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Glyph of Shadow Bolt",
			["icon"] = 136243,
			["id"] = 56240,
		},
		["ACCURATEHUGECITRINE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53892,
			["icon"] = 134071,
			["name"] = "Accurate Huge Citrine",
		},
		["DEATHANDDECAY"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 1882,
			["id"] = 37788,
			["icon"] = 136144,
			["name"] = "Death and Decay",
		},
		["ORNATESPYGLASS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Ornate Spyglass",
			["icon"] = 136243,
			["id"] = 6458,
		},
		["SARONITEDEFENDER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 54557,
			["icon"] = 136241,
			["name"] = "Saronite Defender",
		},
		["HEAVYMAGEWEAVEBANDAGE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 1500,
			["name"] = "Heavy Mageweave Bandage",
			["icon"] = 133690, --136192,
			["id"] = 10841,
		},
		["GLYPHOFFOCUS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 62080,
			["icon"] = 136243,
			["name"] = "Glyph of Focus",
		},
		["GLYPHOFSTRANGULATE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 57225,
			["icon"] = 132918,
			["name"] = "Glyph of Strangulate",
		},
		["BOOKOFCLEVERTRICKS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 59496,
			["icon"] = 132918,
			["name"] = "Book of Clever Tricks",
		},
		["FROSTRESISTANCETOTEM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 8181,
			["icon"] = 135866,
			["name"] = "Frost Resistance Totem",
		},
		["LIONHEARTEXECUTIONER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 50000,
			["id"] = 36259,
			["icon"] = 136241,
			["name"] = "Lionheart Executioner",
		},
		["SUMMONWATERELEMENTAL"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 31687,
			["icon"] = 135862,
			["name"] = "Summon Water Elemental",
		},
		["SAVAGESARONITEBRACERS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55305,
			["icon"] = 136241,
			["name"] = "Savage Saronite Bracers",
		},
		["ARCANEMISSILES"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 5143,
			["icon"] = 136096,
			["name"] = "Arcane Missiles",
		},
		["ELIXIROFDEFENSE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Elixir of Defense",
			["icon"] = 136243,
			["id"] = 3177,
		},
		["FINDHERBS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 2383,
			["icon"] = 133939,
			["name"] = "Find Herbs",
		},
		["BRIGHTSCARLETRUBY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 53947,
			["icon"] = 134071,
			["name"] = "Bright Scarlet Ruby",
		},
		["GNOMISHMINDCONTROLCAP"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 12907,
			["icon"] = 136243,
			["name"] = "Gnomish Mind Control Cap",
		},
		["ELIXIROFGREATERINTELLECT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Elixir of Greater Intellect",
			["icon"] = 136243,
			["id"] = 11465,
		},
		["BASICCAMPFIRE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 818,
			["icon"] = 135805,
			["name"] = "Basic Campfire",
		},
		["BLOODTAP"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 7122,
			["icon"] = 132278,
			["name"] = "Blood Tap",
		},
		["TRANSMUTETITANIUM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 60350,
			["icon"] = 237045,
			["name"] = "Transmute: Titanium",
		},
		["GLACIALWAISTBAND"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 60990,
			["icon"] = 136249,
			["name"] = "Glacial Waistband",
		},
		["MASTERSINSCRIPTIONOFTHEAXE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 4000,
			["id"] = 61117,
			["icon"] = 237171,
			["name"] = "Master's Inscription of the Axe",
		},
		["HURRICANE"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Hurricane",
			["icon"] = 136018,
			["id"] = 16914,
		},
		["ICEBARRIER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 11426,
			["icon"] = 135988,
			["name"] = "Ice Barrier",
		},
		["NORTHRENDALCHEMYRESEARCH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 10000,
			["id"] = 60893,
			["icon"] = 136240,
			["name"] = "Northrend Alchemy Research",
		},
		["ARCANEBARRAGE"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 44425,
			["icon"] = 236205,
			["name"] = "Arcane Barrage",
		},
		["PORTALDALARAN"] = {
			["maxRange"] = 10,
			["minRange"] = 0,
			["castTime"] = 9408,
			["id"] = 53142,
			["icon"] = 237508,
			["name"] = "Portal: Dalaran",
		},
		["COLDWEATHERFLYING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 54197,
			["icon"] = 135833,
			["name"] = "Cold Weather Flying",
		},
		["GLYPHOFICELANCE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 56377,
			["icon"] = 136243,
			["name"] = "Glyph of Ice Lance",
		},
		["DEADLYBLUNDERBUSS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Deadly Blunderbuss",
			["icon"] = 136243,
			["id"] = 3936,
		},
		["FORGEDCOBALTCLAYMORE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["id"] = 55203,
			["icon"] = 136241,
			["name"] = "Forged Cobalt Claymore",
		},
		["BLOODRAGE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Bloodrage",
			["icon"] = 132277,
			["id"] = 2687,
		},
		["DREAMLESSSLEEPPOTION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Dreamless Sleep Potion",
			["icon"] = 136243,
			["id"] = 15833,
		},
		["FROST"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 18977,
			["icon"] = 135849,
			["name"] = "Frost",
		},
		["TRUESILVERTRANSFORMER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["name"] = "Truesilver Transformer",
			["icon"] = 136243,
			["id"] = 23071,
		},
		["CHARGE"] = {
			["maxRange"] = 25,
			["minRange"] = 8,
			["castTime"] = 0,
			["name"] = "Charge",
			["icon"] = 132337,
			["id"] = 100,
		},
		["WRATHOFAIRTOTEM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 2895,
			["icon"] = 136092,
			["name"] = "Wrath of Air Totem",
		},
		["NIGHTSHOCKGIRDLE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["id"] = 60658,
			["icon"] = 136247,
			["name"] = "Nightshock Girdle",
		},
		["NITROBOOSTS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 54861,
			["icon"] = 135788,
			["name"] = "Nitro Boosts",
		},
		["MITHRILBLUNDERBUSS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["name"] = "Mithril Blunderbuss",
			["icon"] = 136243,
			["id"] = 12595,
		},
		["INTRICATEDARKJADE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 53925,
			["icon"] = 134071,
			["name"] = "Intricate Dark Jade",
		},
		["COMMANDINGSHOUT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Commanding Shout",
			["icon"] = 132351,
			["id"] = 469,
		},
		["FELIRONMUSKET"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["id"] = 30312,
			["icon"] = 136243,
			["name"] = "Fel Iron Musket",
		},
		["AZURESILKHOOD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 12500,
			["name"] = "Azure Silk Hood",
			["icon"] = 132149,
			["id"] = 8760,
		},
		["PHANTOMBLADE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 45000,
			["id"] = 10007,
			["icon"] = 135350, --136192,
			["name"] = "Phantom Blade",
		},
		["NORTHRENDINSCRIPTIONRESEARCH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 3000,
			["id"] = 61177,
			["icon"] = 237171,
			["name"] = "Northrend Inscription Research",
		},
		["JOURNEYMANCOOK"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Journeyman Cook",
			["icon"] = 133971,
			["id"] = 3412,
		},
		["BLESSINGOFMIGHT"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Blessing of Might",
			["icon"] = 135906,
			["id"] = 19740,
		},
		["CINDERCLOTHROBE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 25000,
			["name"] = "Cindercloth Robe",
			["icon"] = 132149,
			["id"] = 12069,
		},
		["WILDGROWTH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Wild Growth",
			["icon"] = 136224,
			["id"] = 34161,
		},
		["REBIRTH"] = {
			["maxRange"] = 30,
			["minRange"] = 0,
			["castTime"] = 1882,
			["name"] = "Rebirth",
			["icon"] = 136080,
			["id"] = 20484,
		},
		["ENCHANTGLOVESCRUSHER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Enchant Gloves - Crusher",
			["icon"] = 136244,
			["id"] = 60668,
		},
		["ENCHANTGLOVESMINORHASTE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 13948,
			["icon"] = 135913,
			["name"] = "Enchant Gloves - Minor Haste",
		},
		["ENCHANTCHESTSUPERIORMANA"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 13917,
			["icon"] = 135913,
			["name"] = "Enchant Chest - Superior Mana",
		},
		["ENCHANTSHIELDMINORSTAMINA"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 13378,
			["icon"] = 135913,
			["name"] = "Enchant Shield - Minor Stamina",
		},
		["ENCHANTBRACERSTAMINA"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 13648,
			["icon"] = 135913,
			["name"] = "Enchant Bracer - Stamina",
		},
		["ENCHANTBRACERSEXCEPTIONALINTELLECT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Enchant Bracers - Exceptional Intellect",
			["icon"] = 136244,
			["id"] = 44555,
		},
		["ENCHANTWEAPONEXCEPTIONALAGILITY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Enchant Weapon - Exceptional Agility",
			["icon"] = 136244,
			["id"] = 44633,
		},
		["ENCHANTBOOTSLESSERAGILITY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 13637,
			["icon"] = 135913,
			["name"] = "Enchant Boots - Lesser Agility",
		},
		["ENCHANTGLOVESEXCEPTIONALSPELLPOWER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Enchant Gloves - Exceptional Spellpower",
			["icon"] = 136244,
			["id"] = 44592,
		},
		["ENCHANTCLOAKGREATERAGILITY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 34004,
			["icon"] = 135913,
			["name"] = "Enchant Cloak - Greater Agility",
		},
		["ENCHANTGLOVESMAJORSTRENGTH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 33995,
			["icon"] = 135913,
			["name"] = "Enchant Gloves - Major Strength",
		},
		["ENCHANTCHESTGREATERMANARESTORATION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Enchant Chest - Greater Mana Restoration",
			["icon"] = 136244,
			["id"] = 44509,
		},
		["ENCHANTBRACERSGREATERSTATS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Enchant Bracers - Greater Stats",
			["icon"] = 136244,
			["id"] = 44616,
		},
		["ENCHANTCLOAKMAJORAGILITY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Enchant Cloak - Major Agility",
			["icon"] = 136244,
			["id"] = 60663,
		},
		["ENCHANTWEAPONGREATERSTRIKING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 13943,
			["icon"] = 135913,
			["name"] = "Enchant Weapon - Greater Striking",
		},
		["ENCHANTWEAPONGREATERPOTENCY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Enchant Weapon - Greater Potency",
			["icon"] = 135913,
			["id"] = 60621,
		},
		["ENCHANTGLOVESBLASTING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 33993,
			["icon"] = 135913,
			["name"] = "Enchant Gloves - Blasting",
		},
		["ENCHANTBOOTSMINORSPEED"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 13890,
			["icon"] = 135913,
			["name"] = "Enchant Boots - Minor Speed",
		},
		["ENCHANTCHESTMANA"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 13607,
			["icon"] = 135913,
			["name"] = "Enchant Chest - Mana",
		},
		["ENCHANTCLOAKDEFENSE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 13635,
			["icon"] = 135913,
			["name"] = "Enchant Cloak - Defense",
		},
		["ENCHANTWEAPONMINORSTRIKING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 7788,
			["icon"] = 135913,
			["name"] = "Enchant Weapon - Minor Striking",
		},
		["ENCHANTCHESTLESSERABSORPTION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 13538,
			["icon"] = 135913,
			["name"] = "Enchant Chest - Lesser Absorption",
		},
		["ENCHANTCHESTSUPERSTATS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Enchant Chest - Super Stats",
			["icon"] = 136244,
			["id"] = 44623,
		},
		["RUNEDTRUESILVERROD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 30000,
			["id"] = 13702,
			["icon"] = 135148,
			["name"] = "Runed Truesilver Rod",
		},
		["ENCHANTBRACERSSTRIKING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Enchant Bracers - Striking",
			["icon"] = 136244,
			["id"] = 60616,
		},
		["ENCHANTWEAPONEXCEPTIONALSPELLPOWER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Enchant Weapon - Exceptional Spellpower",
			["icon"] = 136244,
			["id"] = 44629,
		},
		["APPRENTICEENCHANTER"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 7414,
			["icon"] = 136244,
			["name"] = "Apprentice Enchanter",
		},
		["ENCHANTGLOVESAGILITY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 13815,
			["icon"] = 135913,
			["name"] = "Enchant Gloves - Agility",
		},
		["ENCHANTCHESTRESTOREMANAPRIME"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 33991,
			["icon"] = 135913,
			["name"] = "Enchant Chest - Restore Mana Prime",
		},
		["ENCHANTCHESTSUPERHEALTH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Enchant Chest - Super Health",
			["icon"] = 136244,
			["id"] = 47900,
		},
		["ENCHANTCHESTSTATS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 13941,
			["icon"] = 135913,
			["name"] = "Enchant Chest - Stats",
		},
		["ENCHANTBRACERSPIRIT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 13642,
			["icon"] = 135913,
			["name"] = "Enchant Bracer - Spirit",
		},
		["GREATERMAGICWAND"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 10000,
			["id"] = 14807,
			["icon"] = 135144,
			["name"] = "Greater Magic Wand",
		},
		["ENCHANTBOOTSGREATERSPIRIT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Enchant Boots - Greater Spirit",
			["icon"] = 136244,
			["id"] = 44508,
		},
		["RUNEDETERNIUMROD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 30000,
			["name"] = "Runed Eternium Rod",
			["icon"] = 134923,
			["id"] = 32667,
		},
		["ENCHANTCLOAKSPELLPIERCING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Enchant Cloak - Spell Piercing",
			["icon"] = 136244,
			["id"] = 44582,
		},
		["ENCHANTSHIELDDEFENSE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Enchant Shield - Defense",
			["icon"] = 136244,
			["id"] = 44489,
		},
		["ENCHANTCHESTMAJORMANA"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 20028,
			["icon"] = 135913,
			["name"] = "Enchant Chest - Major Mana",
		},
		["ENCHANTCHESTGREATERHEALTH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 13640,
			["icon"] = 135913,
			["name"] = "Enchant Chest - Greater Health",
		},
		["ENCHANT2HWEAPONMINORIMPACT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 7745,
			["icon"] = 135913,
			["name"] = "Enchant 2H Weapon - Minor Impact",
		},
		["ENCHANTBOOTSLESSERACCURACY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 63746,
			["icon"] = 135913,
			["name"] = "Enchant Boots - Lesser Accuracy",
		},
		["ENCHANTCLOAKRESISTANCE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 13794,
			["icon"] = 135913,
			["name"] = "Enchant Cloak - Resistance",
		},
		["GREATERMYSTICWAND"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 10000,
			["id"] = 14810,
			["icon"] = 135469,
			["name"] = "Greater Mystic Wand",
		},
		["ENCHANTCHESTLESSERSTATS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 13700,
			["icon"] = 135913,
			["name"] = "Enchant Chest - Lesser Stats",
		},
		["ENCHANTWEAPONEXCEPTIONALSPIRIT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Enchant Weapon - Exceptional Spirit",
			["icon"] = 136244,
			["id"] = 44510,
		},
		["ENCHANTBRACERSTATS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 27905,
			["icon"] = 135913,
			["name"] = "Enchant Bracer - Stats",
		},
		["ENCHANTGLOVESASSAULT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 33996,
			["icon"] = 135913,
			["name"] = "Enchant Gloves - Assault",
		},
		["ENCHANTBOOTSGREATERVITALITY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Enchant Boots - Greater Vitality",
			["icon"] = 136244,
			["id"] = 44584,
		},
		["ENCHANTRINGGREATERSPELLPOWER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Enchant Ring - Greater Spellpower",
			["icon"] = 136244,
			["id"] = 44636,
		},
		["ENCHANTSHIELDTOUGHSHIELD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 27944,
			["icon"] = 135913,
			["name"] = "Enchant Shield - Tough Shield",
		},
		["ENCHANTSHIELDLESSERSPIRIT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 13485,
			["icon"] = 135913,
			["name"] = "Enchant Shield - Lesser Spirit",
		},
		["ENCHANTCHESTGREATERDEFENSE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Enchant Chest - Greater Defense",
			["icon"] = 135913,
			["id"] = 47766,
		},
		["ENCHANTBRACERSEXPERTISE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Enchant Bracers - Expertise",
			["icon"] = 136244,
			["id"] = 44598,
		},
		["NEXUSTRANSFORMATION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 42613,
			["icon"] = 132882,
			["name"] = "Nexus Transformation",
		},
		["ENCHANTSHIELDSPIRIT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 13659,
			["icon"] = 135913,
			["name"] = "Enchant Shield - Spirit",
		},
		["ENCHANTBOOTSASSAULT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Enchant Boots - Assault",
			["icon"] = 136244,
			["id"] = 60606,
		},
		["ENCHANTCHESTMINORABSORPTION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 7426,
			["icon"] = 135913,
			["name"] = "Enchant Chest - Minor Absorption",
		},
		["ENCHANTBRACERMAJORINTELLECT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 34001,
			["icon"] = 135913,
			["name"] = "Enchant Bracer - Major Intellect",
		},
		["ENCHANTWEAPONSTRIKING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 13693,
			["icon"] = 135913,
			["name"] = "Enchant Weapon - Striking",
		},
		["ENCHANTCHESTGREATERMANA"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 13663,
			["icon"] = 135913,
			["name"] = "Enchant Chest - Greater Mana",
		},
		["ENCHANTEDLEATHER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 17181,
			["icon"] = 134418,
			["name"] = "Enchanted Leather",
		},
		["ENCHANTCLOAKGREATERRESISTANCE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 20014,
			["icon"] = 135913,
			["name"] = "Enchant Cloak - Greater Resistance",
		},
		["ENCHANTBRACERLESSERSTAMINA"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 13501,
			["icon"] = 135913,
			["name"] = "Enchant Bracer - Lesser Stamina",
		},
		["ENCHANTCLOAKFIRERESISTANCE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 13657,
			["icon"] = 135913,
			["name"] = "Enchant Cloak - Fire Resistance",
		},
		["RUNEDSILVERROD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 30000,
			["id"] = 7795,
			["icon"] = 135138,
			["name"] = "Runed Silver Rod",
		},
		["ENCHANT2HWEAPONGREATERIMPACT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 13937,
			["icon"] = 135913,
			["name"] = "Enchant 2H Weapon - Greater Impact",
		},
		["ENCHANTCLOAKLESSERFIRERESISTANCE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 7861,
			["icon"] = 135913,
			["name"] = "Enchant Cloak - Lesser Fire Resistance",
		},
		["ENCHANTCHESTEXCEPTIONALHEALTH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 27957,
			["icon"] = 135913,
			["name"] = "Enchant Chest - Exceptional Health",
		},
		["ENCHANTBOOTSSTAMINA"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 13836,
			["icon"] = 135913,
			["name"] = "Enchant Boots - Stamina",
		},
		["ENCHANTCHESTMIGHTYHEALTH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Enchant Chest - Mighty Health",
			["icon"] = 136244,
			["id"] = 44492,
		},
		["ENCHANTCLOAKGREATERDEFENSE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 13746,
			["icon"] = 135913,
			["name"] = "Enchant Cloak - Greater Defense",
		},
		["ENCHANTGLOVESPRECISION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Enchant Gloves - Precision",
			["icon"] = 135913,
			["id"] = 44488,
		},
		["LESSERMYSTICWAND"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 10000,
			["id"] = 14809,
			["icon"] = 135139,
			["name"] = "Lesser Mystic Wand",
		},
		["ENCHANTBRACERLESSERINTELLECT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 13622,
			["icon"] = 135913,
			["name"] = "Enchant Bracer - Lesser Intellect",
		},
		["ENCHANTCHESTMINORSTATS"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 13626,
			["icon"] = 135913,
			["name"] = "Enchant Chest - Minor Stats",
		},
		["ENCHANTBOOTSMINORSTAMINA"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 7863,
			["icon"] = 135913,
			["name"] = "Enchant Boots - Minor Stamina",
		},
		["ENCHANT2HWEAPONLESSERIMPACT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 13529,
			["icon"] = 135913,
			["name"] = "Enchant 2H Weapon - Lesser Impact",
		},
		["ENCHANTCLOAKLESSERPROTECTION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 13421,
			["icon"] = 135913,
			["name"] = "Enchant Cloak - Lesser Protection",
		},
		["ENCHANTBRACERGREATERINTELLECT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 20008,
			["icon"] = 135913,
			["name"] = "Enchant Bracer - Greater Intellect",
		},
		["EXPERTENCHANTER"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 7416,
			["icon"] = 136244,
			["name"] = "Expert Enchanter",
		},
		["JOURNEYMANENCHANTER"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 7415,
			["icon"] = 136244,
			["name"] = "Journeyman Enchanter",
		},
		["ENCHANTBRACERSTRENGTH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 13661,
			["icon"] = 135913,
			["name"] = "Enchant Bracer - Strength",
		},
		["ENCHANTWEAPONLESSERSTRIKING"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 13503,
			["icon"] = 135913,
			["name"] = "Enchant Weapon - Lesser Striking",
		},
		["ENCHANTGLOVESGATHERER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Enchant Gloves - Gatherer",
			["icon"] = 136244,
			["id"] = 44506,
		},
		["ENCHANTRINGSTAMINA"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Enchant Ring - Stamina",
			["icon"] = 136244,
			["id"] = 59636,
		},
		["ENCHANTBOOTSGREATERAGILITY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 20023,
			["icon"] = 135913,
			["name"] = "Enchant Boots - Greater Agility",
		},
		["ENCHANTBRACERSGREATERSPELLPOWER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Enchant Bracers - Greater Spellpower",
			["icon"] = 136244,
			["id"] = 44635,
		},
		["ENCHANTSHIELDGREATERINTELLECT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Enchant Shield - Greater Intellect",
			["icon"] = 136244,
			["id"] = 60653,
		},
		["ENCHANTGLOVESMAJORAGILITY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Enchant Gloves - Major Agility",
			["icon"] = 136244,
			["id"] = 44529,
		},
		["ARTISANENCHANTER"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 13921,
			["icon"] = 136244,
			["name"] = "Artisan Enchanter",
		},
		["ENCHANTCLOAKSPEED"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Enchant Cloak - Speed",
			["icon"] = 136244,
			["id"] = 60609,
		},
		["ENCHANTGLOVESGREATERSTRENGTH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 20013,
			["icon"] = 135913,
			["name"] = "Enchant Gloves - Greater Strength",
		},
		["ENCHANT2HWEAPONIMPACT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 13695,
			["icon"] = 135913,
			["name"] = "Enchant 2H Weapon - Impact",
		},
		["ENCHANTGLOVESGREATERASSAULT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Enchant Gloves - Greater Assault",
			["icon"] = 136244,
			["id"] = 44513,
		},
		["VOIDSPHERE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 28028,
			["icon"] = 132886,
			["name"] = "Void Sphere",
		},
		["ENCHANTCLOAKMINORRESISTANCE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 7454,
			["icon"] = 135913,
			["name"] = "Enchant Cloak - Minor Resistance",
		},
		["ENCHANTSHIELDVITALITY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 20016,
			["icon"] = 135913,
			["name"] = "Enchant Shield - Vitality",
		},
		["ENCHANTSHIELDRESILIENCE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 44383,
			["icon"] = 135913,
			["name"] = "Enchant Shield - Resilience",
		},
		["ENCHANT2HWEAPONGREATERSAVAGERY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Enchant 2H Weapon - Greater Savagery",
			["icon"] = 136244,
			["id"] = 44630,
		},
		["ENCHANTCHESTMAJORSPIRIT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 33990,
			["icon"] = 135913,
			["name"] = "Enchant Chest - Major Spirit",
		},
		["RUNEDGOLDENROD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 30000,
			["id"] = 13628,
			["icon"] = 135147,
			["name"] = "Runed Golden Rod",
		},
		["ENCHANTEDTHORIUM"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 17180,
			["icon"] = 133229,
			["name"] = "Enchanted Thorium",
		},
		["ENCHANTCLOAKMINORPROTECTION"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 7771,
			["icon"] = 135913,
			["name"] = "Enchant Cloak - Minor Protection",
		},
		["ENCHANTBOOTSGREATERFORTITUDE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Enchant Boots - Greater Fortitude",
			["icon"] = 136244,
			["id"] = 44528,
		},
		["ENCHANTGLOVESSTRENGTH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 13887,
			["icon"] = 135913,
			["name"] = "Enchant Gloves - Strength",
		},
		["ENCHANTSHIELDLESSERSTAMINA"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 13631,
			["icon"] = 135913,
			["name"] = "Enchant Shield - Lesser Stamina",
		},
		["ENCHANTCHESTEXCEPTIONALMANA"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Enchant Chest - Exceptional Mana",
			["icon"] = 135913,
			["id"] = 27958,
		},
		["ENCHANTCLOAKMAJORARMOR"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 27961,
			["icon"] = 135913,
			["name"] = "Enchant Cloak - Major Armor",
		},
		["ENCHANTCHESTLESSERHEALTH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 7748,
			["icon"] = 135913,
			["name"] = "Enchant Chest - Lesser Health",
		},
		["ENCHANTGLOVESGREATERAGILITY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 20012,
			["icon"] = 135913,
			["name"] = "Enchant Gloves - Greater Agility",
		},
		["ENCHANTBRACERSMAJORSPIRIT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Enchant Bracers - Major Spirit",
			["icon"] = 136244,
			["id"] = 44593,
		},
		["ENCHANTBRACERGREATERSTRENGTH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 13939,
			["icon"] = 135913,
			["name"] = "Enchant Bracer - Greater Strength",
		},
		["ENCHANTBRACERBRAWN"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 27899,
			["icon"] = 135913,
			["name"] = "Enchant Bracer - Brawn",
		},
		["LESSERMAGICWAND"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 10000,
			["id"] = 14293,
			["icon"] = 135139,
			["name"] = "Lesser Magic Wand",
		},
		["RUNEDFELIRONROD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 30000,
			["id"] = 32664,
			["icon"] = 134924,
			["name"] = "Runed Fel Iron Rod",
		},
		["ENCHANTCHESTHEALTH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 7857,
			["icon"] = 135913,
			["name"] = "Enchant Chest - Health",
		},
		["ENCHANTBOOTSICEWALKER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Enchant Boots - Icewalker",
			["icon"] = 136244,
			["id"] = 60623,
		},
		["GRANDMASTERENCHANTER"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["name"] = "Grand Master Enchanter",
			["icon"] = 136244,
			["id"] = 51312,
		},
		["ENCHANTCHESTMINORHEALTH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 7420,
			["icon"] = 135913,
			["name"] = "Enchant Chest - Minor Health",
		},
		["ENCHANTBRACERASSAULT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 34002,
			["icon"] = 135913,
			["name"] = "Enchant Bracer - Assault",
		},
		["ENCHANTSTAFFSPELLPOWER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Enchant Staff - Spellpower",
			["icon"] = 135913,
			["id"] = 62959,
		},
		["MASTERENCHANTER"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 28030,
			["icon"] = 136244,
			["name"] = "Master Enchanter",
		},
		["ENCHANTBOOTSAGILITY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 13935,
			["icon"] = 135913,
			["name"] = "Enchant Boots - Agility",
		},
		["ENCHANTBOOTSSUPERIORAGILITY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Enchant Boots - Superior Agility",
			["icon"] = 136244,
			["id"] = 44589,
		},
		["ENCHANTCLOAKSUPERIORAGILITY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Enchant Cloak - Superior Agility",
			["icon"] = 136244,
			["id"] = 44500,
		},
		["RUNEDTITANIUMROD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 30000,
			["name"] = "Runed Titanium Rod",
			["icon"] = 134923,
			["id"] = 60619,
		},
		["ENCHANTGLOVESEXPERTISE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Enchant Gloves - Expertise",
			["icon"] = 136244,
			["id"] = 44484,
		},
		["PRISMATICSPHERE"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 28027,
			["icon"] = 132872,
			["name"] = "Prismatic Sphere",
		},
		["ENCHANTRINGASSAULT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["name"] = "Enchant Ring - Assault",
			["icon"] = 136244,
			["id"] = 44645,
		},
		["ENCHANTBOOTSLESSERSTAMINA"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 13644,
			["icon"] = 135913,
			["name"] = "Enchant Boots - Lesser Stamina",
		},
		["ENCHANTBRACERMINORAGILITY"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 7779,
			["icon"] = 135913,
			["name"] = "Enchant Bracer - Minor Agility",
		},
		["SMALLPRISMATICSHARD"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 42615,
			["icon"] = 132882,
			["name"] = "Small Prismatic Shard",
		},
		["ENCHANTBRACERINTELLECT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 13822,
			["icon"] = 135913,
			["name"] = "Enchant Bracer - Intellect",
		},
		["ENCHANTCHESTSUPERIORHEALTH"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 13858,
			["icon"] = 135913,
			["name"] = "Enchant Chest - Superior Health",
		},
		["ENCHANTSHIELDGREATERSPIRIT"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 13905,
			["icon"] = 135913,
			["name"] = "Enchant Shield - Greater Spirit",
		},
		["ENCHANTBRACERMINORSTAMINA"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 5000,
			["id"] = 7457,
			["icon"] = 135913,
			["name"] = "Enchant Bracer - Minor Stamina",
		},
		["MASTERALCHEMIST"] = {
			["maxRange"] = 100,
			["minRange"] = 0,
			["castTime"] = 0,
			["id"] = 28597,
			["icon"] = 136240,
			["name"] = "Master Alchemist",
		},
		["ROUGHBLASTINGPOWDER"] = {
			["maxRange"] = 0,
			["minRange"] = 0,
			["castTime"] = 2000,
			["id"] = 3918,
			["icon"] = 133848,
			["name"] = "Rough Blasting Powder",
		},
	}
